/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "R5900.h"
#include "InterTables.h"


void COP0() {
	Int_COP0PrintTable[_Rs_]();
}

void COP0_BC0() {
#ifdef COP0_LOG
	COP0_LOG("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc));
#endif
    Int_COP0BC0PrintTable[(cpuRegs.code >> 16) & 0x03]();
}

void COP0_Func() {
#ifdef COP0_LOG
	COP0_LOG("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc));
#endif
    Int_COP0C0PrintTable[_Funct_]();
}

void COP0_Unknown() {
#ifdef CPU_LOG
	CPU_LOG("COP0 Unknown opcode called\n");
#endif
}

void UpdateCP0Status() {
	u32 value = cpuRegs.CP0.n.Status.val;

//	if (Log) SysPrintf("WriteCP0Status: %x\n", value);
	if (value & 0x06 ||
		(value & 0x18) == 0) { // Kernel Mode (KSU = 0 | EXL = 1 | ERL = 1)
//		if (value & 0x4) SysPrintf("*PCSX2*: ERL Set\n");
		memSetKernelMode();
	} else { // User Mode
		memSetUserMode();
	}
	cpuTestHwInts();
}

void WriteCP0Status(u32 value) {
	cpuRegs.CP0.n.Status.val = value;
	UpdateCP0Status();
}

void MFC0() {
	if (!_Rt_) return;
#ifdef COP0_LOG
	if (_Rd_ != 9) { COP0_LOG("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc)); }
#endif
//	if (_Rd_ != 9 && _Rd_ != 12 && _Rd_ != 13 && _Rd_ != 14 && _Rd_ != 28) { SysPrintf("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc)); }
	cpuRegs.GPR.r[_Rt_].UD[0] = (s32)cpuRegs.CP0.r[_Rd_];
}

void MTC0() {
#ifdef COP0_LOG
	COP0_LOG("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc));
#endif
//	if (_Rd_ != 12 && _Rd_ != 14) { SysPrintf("%s\n", disR5900F(cpuRegs.code, cpuRegs.pc)); }
	switch (_Rd_) {
		case 12: WriteCP0Status(cpuRegs.GPR.r[_Rt_].UL[0]); break;
		default: cpuRegs.CP0.r[_Rd_] = cpuRegs.GPR.r[_Rt_].UL[0]; break;
	}
}
/*
#define CPCOND0 ( \
	(!(psHu32(DMAC_PCR) & 0x0001) || (psHu32(DMAC_STAT) & 0x0001)) && \
	(!(psHu32(DMAC_PCR) & 0x0002) || (psHu32(DMAC_STAT) & 0x0002)) && \
	(!(psHu32(DMAC_PCR) & 0x0004) || (psHu32(DMAC_STAT) & 0x0004)) && \
	(!(psHu32(DMAC_PCR) & 0x0008) || (psHu32(DMAC_STAT) & 0x0008)) && \
	(!(psHu32(DMAC_PCR) & 0x0010) || (psHu32(DMAC_STAT) & 0x0010)) && \
	(!(psHu32(DMAC_PCR) & 0x0020) || (psHu32(DMAC_STAT) & 0x0020)) && \
	(!(psHu32(DMAC_PCR) & 0x0040) || (psHu32(DMAC_STAT) & 0x0040)) && \
	(!(psHu32(DMAC_PCR) & 0x0080) || (psHu32(DMAC_STAT) & 0x0080)) && \
	(!(psHu32(DMAC_PCR) & 0x0100) || (psHu32(DMAC_STAT) & 0x0100)) && \
	(!(psHu32(DMAC_PCR) & 0x0200) || (psHu32(DMAC_STAT) & 0x0200)) \
)*/
#define CPCOND0	1

#define BC0(cond) \
	if (CPCOND0 cond) { \
		intDoBranch(_BranchTarget_); \
	}

void BC0F() {
//	SysPrintf("BC0F\n");
	BC0(== 0);
}

void BC0T() {
//	SysPrintf("BC0T\n");
	BC0(== 1);
}

#define BC0L(cond) \
	if (CPCOND0 cond) { \
		intDoBranch(_BranchTarget_); \
	} else cpuRegs.pc+= 4;

void BC0FL() {
//	SysPrintf("BC0FL\n");
	BC0L(== 0);
}

void BC0TL() {
//	SysPrintf("BC0TL\n");
	BC0L(== 1);
}

void TLBR() {
#ifdef CPU_LOG
/*	CPU_LOG("COP0_TLBR %d:%x,%x,%x,%x\n",
			cpuRegs.CP0.n.Random,   cpuRegs.CP0.n.PageMask, cpuRegs.CP0.n.EntryHi,
			cpuRegs.CP0.n.EntryLo0, cpuRegs.CP0.n.EntryLo1);*/
#endif
	SysPrintf("COP0_TLBR\n");
}

void ClearTLB(int i) {
	u32 mask, addr;
	u32 saddr, eaddr;

/*	SysPrintf("ClearTLB %d:%x,%x,%x,%x\n",
			i, cpuRegs.CP0.n.PageMask, cpuRegs.CP0.n.EntryHi,
			cpuRegs.CP0.n.EntryLo0, cpuRegs.CP0.n.EntryLo1);*/

	if (tlb[i].EntryLo0 & 0x2) {
		mask  = ((~tlb[i].Mask) << 1) & 0xfffff;
		saddr = tlb[i].VPN2 >> 12;
		eaddr = saddr + tlb[i].Mask + 1;

//		SysPrintf("tlb clear %x -> %x\n", saddr<<12, eaddr<<12);
		for (addr=saddr; addr<eaddr; addr++) {
			if ((addr & mask) == ((tlb[i].VPN2 >> 12) & mask)) { //match
				memClearPageAddr(addr << 12);
			}
		}
	}

	if (tlb[i].EntryLo1 & 0x2) {
		mask  = ((~tlb[i].Mask) << 1) & 0xfffff;
		saddr = (tlb[i].VPN2 >> 12) + tlb[i].Mask + 1;
		eaddr = saddr + tlb[i].Mask + 1;

//		SysPrintf("tlb clear %x -> %x\n", saddr<<12, eaddr<<12);
		for (addr=saddr; addr<eaddr; addr++) {
			if ((addr & mask) == ((tlb[i].VPN2 >> 12) & mask)) { //match
				memClearPageAddr(addr << 12);
			}
		}
	}
}

void WriteTLB(int i) {
	u32 mask, addr;
	u32 saddr, eaddr;

/*	SysPrintf("WriteTLB %d:%x,%x,%x,%x\n",
			i, cpuRegs.CP0.n.PageMask, cpuRegs.CP0.n.EntryHi,
			cpuRegs.CP0.n.EntryLo0, cpuRegs.CP0.n.EntryLo1);*/

	tlb[i].PageMask = cpuRegs.CP0.n.PageMask;
	tlb[i].EntryHi = cpuRegs.CP0.n.EntryHi;
	tlb[i].EntryLo0 = cpuRegs.CP0.n.EntryLo0;
	tlb[i].EntryLo1 = cpuRegs.CP0.n.EntryLo1;

	tlb[i].Mask = (cpuRegs.CP0.n.PageMask >> 13) & 0xfff;
	tlb[i].nMask = (~tlb[i].Mask) & 0xfff;
	tlb[i].VPN2 = ((cpuRegs.CP0.n.EntryHi >> 13) & (~tlb[i].Mask)) << 13;
	tlb[i].ASID = cpuRegs.CP0.n.EntryHi & 0xfff;
	tlb[i].G = cpuRegs.CP0.n.EntryLo0 & cpuRegs.CP0.n.EntryLo1 & 0x1;
	tlb[i].PFN0 = (((cpuRegs.CP0.n.EntryLo0 >> 6) & 0xFFFFF) & (~tlb[i].Mask)) << 12;
	tlb[i].PFN0|= (0x80000000);
	tlb[i].PFN1 = (((cpuRegs.CP0.n.EntryLo1 >> 6) & 0xFFFFF) & (~tlb[i].Mask)) << 12;
	tlb[i].PFN1|= (0x80000000);
//	SysPrintf("VPN2=%8.8x, PFN0=%8.8x, PFN1=%8.8x (Mask=%x, nMask=%x)\n", tlb[i].VPN2, tlb[i].PFN0, tlb[i].PFN1, tlb[i].Mask, tlb[i].nMask);
	if (tlb[i].VPN2 == 0x70000000) return;

	if (tlb[i].EntryLo0 & 0x2) {
		mask  = ((~tlb[i].Mask) << 1) & 0xfffff;
		saddr = tlb[i].VPN2 >> 12;
		eaddr = saddr + tlb[i].Mask + 1;

//		SysPrintf("tlb set %x - %x\n", saddr<<12, eaddr<<12);
		for (addr=saddr; addr<eaddr; addr++) {
			if ((addr & mask) == ((tlb[i].VPN2 >> 12) & mask)) { //match
				memSetPageAddr(addr << 12, tlb[i].PFN0 + ((addr - saddr) << 12));
			}
		}
	}

	if (tlb[i].EntryLo1 & 0x2) {
		mask  = ((~tlb[i].Mask) << 1) & 0xfffff;
		saddr = (tlb[i].VPN2 >> 12) + tlb[i].Mask + 1;
		eaddr = saddr + tlb[i].Mask + 1;

//		SysPrintf("tlb set %x -> %x\n", saddr<<12, eaddr<<12);
		for (addr=saddr; addr<eaddr; addr++) {
			if ((addr & mask) == ((tlb[i].VPN2 >> 12) & mask)) { //match
				memSetPageAddr(addr << 12, tlb[i].PFN1 + ((addr - saddr) << 12));
			}
		}
	}
}

void TLBWI() { 
	int j = cpuRegs.CP0.n.Index & 0x3f;

	if (j > 48) return;

#ifdef CPU_LOG
/*	CPU_LOG("COP0_TLBWI %d:%x,%x,%x,%x\n",
			cpuRegs.CP0.n.Index,    cpuRegs.CP0.n.PageMask, cpuRegs.CP0.n.EntryHi,
			cpuRegs.CP0.n.EntryLo0, cpuRegs.CP0.n.EntryLo1);*/
#endif

	ClearTLB(j);
	WriteTLB(j);
}

void TLBWR() { 
	int j = cpuRegs.CP0.n.Random & 0x3f;

	if (j > 48) return;

#ifdef CPU_LOG
/*	CPU_LOG("COP0_TLBWR %d:%x,%x,%x,%x\n",
			cpuRegs.CP0.n.Random,   cpuRegs.CP0.n.PageMask, cpuRegs.CP0.n.EntryHi,
			cpuRegs.CP0.n.EntryLo0, cpuRegs.CP0.n.EntryLo1);*/
#endif
	ClearTLB(j);
	WriteTLB(j);
}

void TLBP() {
	static int hack;
	int i;

	hack++;
	if (hack >= 48) hack = 0;
	cpuRegs.CP0.n.Index = hack;//0x80000001;
	return;
	for (i=0; i < 48; i++) {
		int mask = (tlb[i].nMask) << 13;

		SysPrintf("Probe[%d] %x, %x (%x:%x)\n", i, cpuRegs.CP0.n.EntryHi, tlb[i].EntryHi, tlb[i].PageMask, mask);
		if (((cpuRegs.CP0.n.EntryHi & mask) == (tlb[i].EntryHi & mask)/* &&
			(cpuRegs.CP0.n.EntryHi & 0xff) == (tlb[i].Hi)*/) ||
			tlb[i].VPN2 == 0/* & 0x80000000*/) {
			//match
/*			if (cpuRegs.CP0.n.EntryHi & page) {//odd page
				if (tlb[i].Lo1 & 0x2) cpuRegs.CP0.n.Index = i;
			} else {//even page
				if (tlb[i].Lo0 & 0x2) cpuRegs.CP0.n.Index = i;
			}*/
			cpuRegs.CP0.n.Index = i; break;
		}
	}
	if (cpuRegs.CP0.n.Index == 0x23) { Log=1; varLog|= 0x40000000; }
	SysPrintf("COP0_TLBP: %x\n", cpuRegs.CP0.n.Index);
}

void ERET() {
	if (cpuRegs.CP0.n.Status.b.ERL) {
		cpuRegs.pc = cpuRegs.CP0.n.ErrorEPC;
		cpuRegs.CP0.n.Status.b.ERL = 0;
	} else {
		cpuRegs.pc = cpuRegs.CP0.n.EPC;
		cpuRegs.CP0.n.Status.b.EXL = 0;
	}
	UpdateCP0Status();
	intSetBranch();
}

void DI() {
	if (cpuRegs.CP0.n.Status.b.EDI || cpuRegs.CP0.n.Status.b.EXL ||
		cpuRegs.CP0.n.Status.b.ERL || (cpuRegs.CP0.n.Status.b.KSU == 0)) {
		cpuRegs.CP0.n.Status.b.EIE = 0;
		UpdateCP0Status();
	}
}

void EI() {
	if (cpuRegs.CP0.n.Status.b.EDI || cpuRegs.CP0.n.Status.b.EXL ||
		cpuRegs.CP0.n.Status.b.ERL || (cpuRegs.CP0.n.Status.b.KSU == 0)) {
		cpuRegs.CP0.n.Status.b.EIE = 1;
		UpdateCP0Status();
	}
}

