/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2005  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/stat.h>

#include "Common.h"

typedef struct {
	u32 tag;
	u64 data[2];
} _cache;
_cache pCache[256];

void freeCache(int i) {
	u64 *p;
	u32 mem = pCache[i].tag << 4;

	pCache[i].tag = 0;
	p = (u64 *)((u8*)(memLUTW[mem >> 12]) + (mem & 0xff0));
	p[0] = pCache[i].data[0];
	p[1] = pCache[i].data[1];
}

int getFreeCache() {
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) {
			break;
		}
	}

	if (i == 256) {
		i = 0;
		freeCache(i);
	}

	return i;
}

void writeCache8(u32 mem, u8 value) {
	u64 *p;
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("writeCache8 %8.8x found at %d\n", mem, i);
			*(u8*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
			return;
		}
	}

	i = getFreeCache();
//	CPU_LOG("writeCache8 %8.8x adding to %d\n", mem, i);

	pCache[i].tag = 0x40000000 | (mem >> 4);
	p = (u64 *)((u8*)(memLUTW[mem >> 12]) + (mem & 0xff0));
	pCache[i].data[0] = p[0];
	pCache[i].data[1] = p[1];
//	CPU_LOG("%I64x_%I64x\n", p[1], p[0]);
	*(u8*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
}

void writeCache16(u32 mem, u16 value) {
	u64 *p;
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("writeCache16 %8.8x found at %d\n", mem, i);
			*(u16*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
			return;
		}
	}

	i = getFreeCache();
//	CPU_LOG("writeCache16 %8.8x adding to %d\n", mem, i);

	pCache[i].tag = 0x40000000 | (mem >> 4);
	p = (u64 *)((u8*)(memLUTW[mem >> 12]) + (mem & 0xff0));
	pCache[i].data[0] = p[0];
	pCache[i].data[1] = p[1];
//	CPU_LOG("%I64x_%I64x\n", p[1], p[0]);
	*(u16*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
}

void writeCache32(u32 mem, u32 value) {
	u64 *p;
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("writeCache32 %8.8x found at %d\n", mem, i);
			*(u32*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
			return;
		}
	}

	i = getFreeCache();
//	CPU_LOG("writeCache32 %8.8x adding to %d\n", mem, i);

	pCache[i].tag = 0x40000000 | (mem >> 4);
	p = (u64 *)((u8*)(memLUTW[mem >> 12]) + (mem & 0xff0));
	pCache[i].data[0] = p[0];
	pCache[i].data[1] = p[1];
//	CPU_LOG("%I64x_%I64x\n", p[1], p[0]);
	*(u32*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
}

void writeCache64(u32 mem, u64 value) {
	u64 *p;
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("writeCache64 %8.8x found at %d\n", mem, i);
			*(u64*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
			return;
		}
	}

	i = getFreeCache();
//	CPU_LOG("writeCache64 %8.8x adding to %d\n", mem, i);

	pCache[i].tag = 0x40000000 | (mem >> 4);
	p = (u64 *)((u8*)(memLUTW[mem >> 12]) + (mem & 0xff0));
	pCache[i].data[0] = p[0];
	pCache[i].data[1] = p[1];
//	CPU_LOG("%I64x_%I64x\n", p[1], p[0]);
	*(u64*)(((u8*)pCache[i].data)+(mem&0xf)) = value;
}

void writeCache128(u32 mem, u64 *value) {
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("writeCache128 %8.8x found at %d\n", mem, i);
			pCache[i].data[0] = value[0];
			pCache[i].data[1] = value[1];
			return;
		}
	}

	i = getFreeCache();
//	CPU_LOG("writeCache128 %8.8x adding to %d\n", mem, i);

	pCache[i].tag = 0x40000000 | (mem >> 4);
	pCache[i].data[0] = value[0];
	pCache[i].data[1] = value[1];
//	CPU_LOG("%I64x_%I64x\n", value[1], value[0]);
}

u8 *readCache(u32 mem) {
	u64 *p;
	int i;

	for (i=0; i<256; i++) {
		if ((pCache[i].tag & 0x40000000) == 0) continue;
		if ((pCache[i].tag & 0xfffffff) == (mem >> 4)) {
//			CPU_LOG("readCache %8.8x found at %d\n", mem, i);
			return (u8*)pCache[i].data;
		}
	}

	i = getFreeCache();
//	CPU_LOG("readCache %8.8x adding to %d\n", mem, i);

	i=0;
	pCache[i].tag = 0x40000000 | (mem >> 4);
	p = (u64 *)((u8*)(memLUTR[mem >> 12]) + (mem & 0xff0));
	pCache[i].data[0] = p[0];
	pCache[i].data[1] = p[1];
//	CPU_LOG("%I64x_%I64x\n", p[1], p[0]);
	return (u8*)pCache[i].data;
}


void CACHE() {
    u32 addr;

    addr = cpuRegs.GPR.r[_Rs_].UL[0] + _Imm_;
#ifdef CPU_LOG
//	CPU_LOG("CACHE op %x, addr %x\n", _Rt_, addr);
#endif

/*	switch (_Rt_) {
		case 0x07: // IXIN
			break;

		case 0x00: // IXLTG
			break;

		case 0x04: // IXSTG
			break;

		case 0x0b: // IHIN
			break;

		case 0x0e: // IFL
			break;

		case 0x01: // IXLDT
			break;

		case 0x05: // IXSDT
			break;

		case 0x10: // BTAC Probe
			break;
	}*/
}

