/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __COMMON_H__
#define __COMMON_H__

#include <zlib.h>
#if defined(__WIN32__)

#include <windows.h>

typedef struct {
	HWND hWnd;           // Main window handle
	HINSTANCE hInstance; // Application instance
	HMENU hMenu;         // Main window menu
	HANDLE hConsole;
} AppData;

extern AppData gApp;

#elif defined (__LINUX__)

#include <sys/types.h>

#define BOOL int

#undef TRUE
#define TRUE  1
#undef FALSE
#define FALSE 0

#define max(a,b)            (((a) > (b)) ? (a) : (b))
#define min(a,b)            (((a) < (b)) ? (a) : (b))

#define strnicmp strncasecmp
#define stricmp strcasecmp

#endif

#ifdef ENABLE_NLS

#ifdef __MSCW32__
#include "libintlmsc.h"
#else
#include "libintl.h"
#endif

#undef _
#define _(String) dgettext (PACKAGE, String)
#ifdef gettext_noop
#  define N_(String) gettext_noop (String)
#else
#  define N_(String) (String)
#endif

#else

#define _(msgid) msgid
#define N_(msgid) msgid

#endif

#include "PS2Etypes.h"

typedef struct {
  char Bios[256];
  char GS[256];
  char PAD1[256];
  char PAD2[256];
  char SPU2[256];
  char CDVD[256];
  char DEV9[256];
  char USB[256];
  char FW[256];
  char Mcd1[256];
  char Mcd2[256];
  char PluginsDir[256];
  char BiosDir[256];
  char Lang[256];
  int PsxOut;
  int Cpu;
  int PsxType;
  int Cdda;
  int Mdec;
  int Patch;
  int Regcaching;
  int VUrec;
  int PadHack;
  int ThPriority;
} PcsxConfig;

PcsxConfig Config;

#define PS2CLK 294912000	/* 294.912 mhz */
//#define PSXCLK	33868800	/* 33.8688 Mhz */
#define PSXCLK	36864000	/* 36.864 Mhz */

#define BIAS 2

/* Config.PsxType == 1: PAL:
	 VBlank interlaced		50.00 Hz
	 VBlank non-interlaced	49.76 Hz
	 HBlank					15.625 KHz
   Config.PsxType == 0: NSTC
	 VBlank interlaced		59.94 Hz
	 VBlank non-interlaced	59.82 Hz
	 HBlank					15.73426573 KHz */
#define PS2VBLANK_NTSC	((int)(PS2CLK / 59.94))
#define PS2VBLANK_PAL	((int)(PS2CLK / 50))
#define PS2VBLANK		(Config.PsxType ? PS2VBLANK_PAL : PS2VBLANK_NTSC)
#define PS2HBLANK_NTSC	((int)(PS2CLK / 15734.26573))
#define PS2HBLANK_PAL	((int)(PS2CLK / 15625))
#define PS2HBLANK		(Config.PsxType ? PS2HBLANK_PAL : PS2HBLANK_NTSC)

#define PSXVBLANK_NTSC	((int)(PSXCLK / 59.94))
#define PSXVBLANK_PAL	((int)(PSXCLK / 50))
#define PSXVBLANK		(Config.PsxType ? PSXVBLANK_PAL : PSXVBLANK_NTSC)
#define PSXHBLANK_NTSC	((int)(PSXCLK / 15734.26573))
#define PSXHBLANK_PAL	((int)(PSXCLK / 15625))
#define PSXHBLANK		(Config.PsxType ? PSXHBLANK_PAL : PSXHBLANK_NTSC)


#include "R5900.h"
#include "Debug.h"
#include "System.h"
#include "Memory.h"
#include "Elfheader.h"
#include "Hw.h"
#include "GS.h"
#include "Vif.h"
#include "SPR.h"
#include "Sif.h"
#include "Plugins.h"
#include "PS2Edefs.h"
#include "Counters.h"
#include "IPU.h"
#include "Misc.h"
#include "Patch.h"
#include "Stats.h"
#include "COP0.h"
#include "VifDma.h"
#if (defined(__i386__) || defined(__x86_64__))
#include "ix86/ix86.h"
#endif

#define gzfreeze(ptr, size) \
	if (Mode == 1) gzwrite(f, ptr, size); \
	if (Mode == 0) gzread(f, ptr, size);

#define gzfreezel(ptr) gzfreeze(ptr, sizeof(ptr))

int cdCaseopen;

int Log;

extern void __Log(char *fmt, ...);
extern u16 logProtocol;
extern u8  logSource;

#define PCSX2_VERSION "0.8.1"

#endif /* __COMMON_H__ */
