/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <time.h>

#include "Common.h"
#include "PsxCommon.h"


int cnts = 5;

void rcntUpdTarget(int index) {
//	if (index == 1) SysPrintf("rcntUpdTarget %x\n", index);
	counters[index].sCycleT = cpuRegs.cycle;
	counters[index].CycleT  = ((counters[index].target - rcntRcount(index)) * counters[index].rate) / BIAS;
}

void rcntUpd(int index) {
	counters[index].sCycle = cpuRegs.cycle;
	counters[index].Cycle  = ((0xffff - (counters[index].count&0xffff)) * counters[index].rate) / BIAS;
	rcntUpdTarget(index);
}

void rcntReset(int index) {
	counters[index].count = 0;
	counters[index].mode&= ~0x00400000;
	rcntUpd(index);
}

void rcntSet() {
	u32 c;
	int i;

	nextCounter = 0xffffffff;
	nextsCounter = cpuRegs.cycle;

	for (i = 0; i < cnts; i++) {
		if ((counters[i].mode & 0x80) == 0) continue; // Stopped

		c = counters[i].Cycle - (cpuRegs.cycle - counters[i].sCycle);
		if (c < nextCounter) {
			nextCounter = c;
		}
		c = counters[i].CycleT - (cpuRegs.cycle - counters[i].sCycleT);
		if (c < nextCounter) {
			nextCounter = c;
		}
	}
}

void rcntInit() {
	int i;

	memset(counters, 0, sizeof(counters));

	for (i=0; i<4; i++) {
		counters[i].rate = 2;
	}
	counters[0].interrupt =  9;
	counters[1].interrupt = 10;
	counters[2].interrupt = 11;
	counters[3].interrupt = 12;

	counters[4].mode = 0x1c0; // The VSync counter mode
	counters[4].count = 0;
	UpdateVSyncRate();

	for (i=0; i<cnts; i++) rcntUpd(i);
	rcntSet();
}

void UpdateVSyncRate() {
	if (Config.PsxType) {
		counters[4].rate = PS2HBLANK_PAL / BIAS;
		counters[4].target = 286;
	} else {
		counters[4].rate = PS2HBLANK_NTSC / BIAS;
		counters[4].target = 240;
	}
}

void UpdateVSyncRateEnd() {
	if (Config.PsxType) {
		counters[4].target = 26;
	} else {
		counters[4].target = 22;
	}
}

extern void PadReadStatus(int pad);

void VSync() {
	if (counters[4].mode & 0x10000) { // VSync End (22 hsyncs)
		counters[4].mode&= ~0x10000;
		UpdateVSyncRate();
		if (!(IMR & 0x800)) hwIntcIrq(3);
//		if (!(IMR & 0x800)) hwIntcIrq(0); /* really? Aura uses this instead of 0x8 (linuz) */
		CSRr|= 0x8;
		psxVSyncEnd();
#ifndef NOSTATS
		statsVSync();
#endif
	} else { // VSync Start (240 hsyncs) 
		UpdateVSyncRateEnd();
		GSvsync();
		SysUpdate();
		counters[4].mode|= 0x10000;
		hwIntcIrq(2);
		psxVSyncStart();
		applypatch(1);
//		SysPrintf("VSyncStart: %x (MASK=%x, Status=%x)\n", cpuRegs.cycle, psHu32(INTC_MASK), cpuRegs.CP0.n.Status.val);
	}
}

void rcntUpdate() {
	int i;

	for (i=0; i<4; i++) {
		if ((counters[i].mode & 0x80) == 0) continue; // Stopped

		if ((cpuRegs.cycle - counters[i].sCycleT) >= counters[i].CycleT &&
			(counters[i].mode & 0x00400000) == 0) {
			if (counters[i].mode & 0x40) { // Reset on target
				rcntReset(i);
			} else {
				counters[i].mode|= 0x00400000;
			}
			if (counters[i].mode & 0x0100) { // Target interrupt
				counters[i].mode|= 0x0400; // Target flag
				hwIntcIrq(counters[i].interrupt);
//				SysPrintf("counter[%d] target interrupt (%x)\n", i, cpuRegs.cycle);
			}
		}
		if ((cpuRegs.cycle - counters[i].sCycle) >= counters[i].Cycle) {
			if (counters[i].mode & 0x0200) { // Overflow interrupt
				counters[i].mode|= 0x0800; // Overflow flag
				hwIntcIrq(counters[i].interrupt);
//				SysPrintf("counter[%d] overflow interrupt (%x)\n", i, cpuRegs.cycle);
			}

			rcntReset(i);
		}
	}

	if ((cpuRegs.cycle - counters[4].sCycleT) >= counters[4].CycleT) {
		counters[4].sCycleT = cpuRegs.cycle;
		counters[4].CycleT  = counters[4].rate;// / BIAS;
		counters[4].count++;
//		if (IMR & 0x400) hwIntcIrq(2);
		CSRr|= 0x4;

		if (counters[4].count >= counters[4].target) {
			counters[4].count = 0;
			VSync();
		}
	}

	rcntSet();
}

void rcntWcount(int index, u32 value) {
//	SysPrintf ("writeCcount[%d] = %x\n", index, value);
	counters[index].count = value;
	rcntUpd(index);
	rcntSet();
}

void rcntWmode(int index, u32 value)  {
//	SysPrintf ("writeCmode[%ld] = %lx\n", index, value);
	if (value & 0xc00) {
		counters[index].mode&= ~(value & 0xc00);
	}
	if ((value & 0x3ff) == (counters[index].mode & 0x3ff)) {
		return;
	}
	counters[index].mode = value & 0x3ff;

	switch (value & 0x3) {
		case 0: counters[index].rate = 2; break;
		case 1: counters[index].rate = 2 * 16; break;
		case 2: counters[index].rate = 2 * 256; break;
		case 3: counters[index].rate = PS2HBLANK; break;
	}

	// Need to set a rate and target
	rcntUpd(index);
	rcntSet();
}

void rcntWtarget(int index, u32 value) {
//	SysPrintf ("writeCtarget[%ld] = %lx\n", index, value);
	counters[index].target = value;
	rcntUpdTarget(index);
	rcntSet();
}

void rcntWhold(int index, u32 value) {
//	SysPrintf ("writeChold[%ld] = %lx\n", index, value);
	counters[index].hold = value;
	rcntUpd(index);
	rcntSet();
}

u32 rcntRcount(int index) {
	u16 ret;
	ret = (u16)(counters[index].count + (BIAS * (cpuRegs.cycle - counters[index].sCycle)) / counters[index].rate);
//	SysPrintf("rcntRcount[%d] %x\n", index, ret);
	return ret;
}

int rcntFreeze(gzFile f, int Mode) {
	gzfreezel(counters);
	gzfreeze(&nextCounter, sizeof(nextCounter));
	gzfreeze(&nextsCounter, sizeof(nextsCounter));

	return 0;
}

