/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Common.h"

DMACh *gif;

void gsInit() {
	gif = (DMACh*)&psH[0xA000];
}

void gsReset() {
	CSRr = 0xc;
	IMR = 0xff00;
}

void CSRwrite(u32 value) {
//	SysPrintf("GS_CSR: %x (CSRr=%x)\n", value, CSRr);
	CSRw = value;
	if (value & 0x200) { // resetGS
		CSRr = 0xc;
		IMR = 0xff00;
	} else {
		if (value & 0x8) CSRr&= ~0x8;
		if (value & 0x4) CSRr&= ~0x4;
	}
}

void IMRwrite(u32 value) {
//	SysPrintf("GS_IMR: %x\n", value);
	IMR = value;
	if (IMR & 0x8000) IMR = 0;
}

void gsWrite8(u32 mem, u8 value) {
	GSwrite8(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite((CSRw & ~0x000000ff) | value); break;
		case 0x12001001: // GS_CSR
			CSRwrite((CSRw & ~0x0000ff00) | (value <<  8)); break;
		case 0x12001002: // GS_CSR
			CSRwrite((CSRw & ~0x00ff0000) | (value << 16)); break;
		case 0x12001003: // GS_CSR
			CSRwrite((CSRw & ~0xff000000) | (value << 24)); break;

/*		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;*/
	}

#ifdef HW_LOG
	HW_LOG("GS write 8 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite16(u32 mem, u16 value) {
	GSwrite16(mem, value);

/*	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite(value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;
	}*/

#ifdef HW_LOG
	HW_LOG("GS write 16 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite32(u32 mem, u32 value) {
	GSwrite32(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite(value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite(value);
			break;
	}

#ifdef HW_LOG
	HW_LOG("GS write 32 at %8.8lx with data %8.8lx\n", mem, value);
#endif
}

void gsWrite64(u32 mem, u64 value) {
	GSwrite64(mem, value);

	switch (mem) {
		case 0x12001000: // GS_CSR
			CSRwrite((u32)value);
			break;

		case 0x12001010: // GS_IMR
			IMRwrite((u32)value);
			break;
	}

#ifdef HW_LOG
	HW_LOG("GS write 64 at %8.8lx with data %8.8lx_%8.8lx\n", mem, ((u32*)&value)[1], (u32)value);
#endif
}

u8 gsRead8(u32 mem) {
	u8 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)CSRr; break;
		case 0x12001001: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >>  8); break;
		case 0x12001002: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >> 16); break;
		case 0x12001003: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread8(mem);
			ret|= (u8)(CSRr >> 24); break;
		default:
			ret = GSread8(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 8 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u16 gsRead16(u32 mem) {
	u16 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread16(mem);
			ret|= (u16)CSRr; break;
		case 0x12001002: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread16(mem);
			ret|= (u16)(CSRr >> 16); break;

		default:
			ret = GSread16(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 16 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u32 gsRead32(u32 mem) {
	u32 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread32(mem);
			ret|= CSRr; break;

		default:
			ret = GSread32(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 32 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

u64 gsRead64(u32 mem) {
	u64 ret=0;

	switch (mem) {
		case 0x12001000: // GS_CSR
			if (GSsetCSR == NULL) ret = GSread64(mem);
//			ret|= 0x00032000/* | 0x4000*/;
			ret|= CSRr;
			break;

		default:
			ret = GSread64(mem);
			break;
	}

#ifdef HW_LOG
//	HW_LOG("GS read 64 %8.8lx, at %8.8lx\n", ret, mem);
#endif
	return ret;
}

void gsIrq() {
	hwIntcIrq(0);
}

int  gsInterrupt() {
#ifdef GS_LOG 
	GS_LOG("gsInterrupt: %8.8x\n", cpuRegs.cycle);
#endif

	gif->chcr &=~ 0x100;
	hwDmacIrq(DMAC_GIF);
	psHu32(GIF_STAT)&= ~0xE00; // OPH=0 | APATH=0
	psHu32(GIF_STAT)&= ~0x1F000000; // QFC=0

	return 1;
}

int  _GIFchain() {
	u32 qwc = gif->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;

	pMem = dmaGetAddr(gif->madr);
	if (pMem == NULL) return -1;

	GSgifTransfer3(pMem, qwc);
	gif->madr+= gif->qwc*16;
	gif->qwc = 0;
	return qwc;
}

#define GIFchain() \
	if (gif->qwc) { \
		if (_GIFchain() == -1) break; \
		cycles+= gif->qwc / BIAS; /* guessing */ \
	}

void dmaGIF() {
	u32 *ptag;
	u32 temp,finalAddress;
	int done=0;
	int cycles=0;
	int id;

#ifdef GIF_LOG
	GIF_LOG("dmaGIF chcr = %lx, madr = %lx, qwc  = %lx\n"
			"        tadr = %lx, asr0 = %lx, asr1 = %lx\n",
			gif->chcr, gif->madr, gif->qwc,
			gif->tadr, gif->asr0, gif->asr1);
#endif

	if (psHu32(GIF_MODE) & 0x4) {
	} else
	if (vif1Regs->mskpath3 || psHu32(GIF_MODE) & 0x1) {
		gif->chcr &=~ 0x100;
		hwDmacIrq(2);
		return;
	}
	if ((psHu32(DMAC_CTRL) & 0xC0) == 0x80) { // STD == GIF
		gif->chcr &=~ 0x100;
		hwDmacIrq(2);
		return;
	}
	if ((psHu32(DMAC_CTRL) & 0xC) == 0xC/* &&
		(gif->tadr == psHu32(0xe050))*/) { // GIF MFIFO
		return;
	}

	psHu32(GIF_STAT)|= 0xE00; // OPH=1 | APATH=3
	psHu32(GIF_STAT)|= 0x1F000000; // FQC=31, hack ;)

	// Transfer Dn_QWC from Dn_MADR to GIF
	while (1) { GIFchain(); break; }

	if ((gif->chcr & 0xc) == 0) { // Normal Mode
		INT(2, cycles);
		return;
	}

	// Chain Mode
	while (done == 0) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)dmaGetAddr(gif->tadr);
		if (ptag == NULL) {
			psHu32(DMAC_STAT)|= 1<<15;
			break;
		}
	
		// Transfer dma tag if tte is set
		if (gif->chcr & 0x40) {
#ifdef GIF_LOG
			GIF_LOG("GIF TTE: %x_%x\n", ptag[3], ptag[2]);
#endif

//			Log=1;
		}
//		if (gif->chcr & 0x40) GSgifTransfer3(ptag+2, 4);

		gif->chcr = ( gif->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

		id        = (ptag[0] >> 28) & 0x7;
		gif->qwc  = (u16)ptag[0];
		gif->madr = ptag[1];

#ifdef GIF_LOG
		GIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], gif->qwc, id, gif->madr);
#endif
		
		switch (id) {
			case 0: // refe
				gif->tadr += 16;
				done = 1;
				break;

			case 1: // cnt
				gif->madr = gif->tadr + 16;
				// Set the taddr to the next tag
				gif->tadr += 16 + (gif->qwc * 16);
				break;

			case 2: // next
				temp = gif->madr;
				gif->madr = gif->tadr + 16;
				gif->tadr = temp;
				break;

			case 3: // ref
			case 4: // refs
				gif->tadr += 16;
				break;

			case 5: // call
				temp = gif->madr;
				finalAddress = gif->tadr + 16 + (gif->qwc * 16);

				if (gif->asr0 == 0) 
					gif->asr0 = finalAddress;
				else
					gif->asr1 = finalAddress;

				gif->madr = gif->tadr + 16;
				gif->tadr = temp;
				break;

			case 6: // ret
				gif->madr = gif->tadr + 16;
				if (gif->asr1 != 0) {
					gif->tadr = gif->asr1;
					gif->asr1 = 0;
				} else {
					if(gif->asr0 != 0) {
						gif->tadr = gif->asr0;
						gif->asr0 = 0;
					} else {
						gif->tadr = 0;
						done = 1;
					}
				}
				break;

			case 7: // end
				gif->madr = gif->tadr + 16;
				done = 1;
				break;
		}
		GIFchain();

/*		if (gif->chcr & 0x80 && ptag[0] >> 31) {
			_dmaGetAddr(ptag, gif->tadr, 2);
//			gif->chcr = ( gif->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

#ifdef GIF_LOG
			GIF_LOG("dmaIrq Set\n");
#endif
//			SysPrintf("GSdmaIrq Set\n");
			break;
		}*/
	}
	INT(2, cycles);
}

#undef GIFchain
#define GIFchain() \
	if (dma->qwc) { \
		_dmaGetAddr(pMem, dma->madr, 8); \
		GSgifTransfer3(pMem, dma->qwc); \
	}

void MFIFO_GIFtransfer(DMACh *dma, u32 sadr, int qwc) {
	u32 *pMem,*ptag;
	int id;
	int cycles = 0;
 
#ifdef GIF_LOG
	GIF_LOG("MFIFO_GIFtransfer %x\n",qwc);
#endif
 
	if((dma->chcr & 0x100) != 0x100) return;

	while (qwc > 0) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)&psS[sadr & 0x3fff]; qwc--;
 
		id        = (ptag[0] >> 28) & 0x7;
		dma->qwc  = (u16)ptag[0];
		dma->madr = ptag[1];
 
#ifdef GIF_LOG
		GIF_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], dma->qwc, id, dma->madr);
#endif
 
		switch (id) {
			case 0: // refe
				GIFchain();
				sadr += 16;
				dma->madr += dma->qwc * 16;
				dma->tadr += 16;
				dma->qwc = 0;
				qwc = 0;
				dma->chcr &=~ 0x100;
				hwDmacIrq(2);
//				INT(2,cycles);
				break;
 
			case 1: // cnt
				dma->madr = sadr + 16;
				GIFchain();
				// Set the taddr to the next tag
				sadr += 16 + (dma->qwc * 16);
				dma->tadr += 16 + (dma->qwc * 16);
				qwc-= dma->qwc;
				dma->qwc = 0;
				break;
 
			case 3: // ref
			case 4: // refs
				GIFchain();
				sadr += 16;
				dma->tadr += 16;
				dma->qwc = 0;
				break;
 
			case 7: // end
				dma->madr = sadr + 16;
				GIFchain();
				dma->qwc = 0;
				qwc = 0;
				dma->chcr &=~ 0x100;
				hwDmacIrq(2);
//				INT(2,cycles);
				break;
		}
	}
}

int gsFreeze(gzFile f, int Mode) {
	gzfreeze(&CSRw, sizeof(CSRw));
	gzfreeze(&CSRr, sizeof(CSRr));
	gzfreeze(&IMR, sizeof(IMR));

	return 0;
}

