/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "VU0.h"
#include "VUmicro.h"
#include "PsxMem.h"
#include "IPU.h"

int hwInit() {
	psH = (u8*)malloc(0x00010000);
	if (psH == NULL) {
		SysMessage(_("Error allocating memory")); return -1;
	}
	vif0Regs = (VIFregisters*)&psH[0x3800];
	vif1Regs = (VIFregisters*)&psH[0x3c00];

	gsInit();
	vif0Init();
	vif1Init();
	vifDmaInit();
	sifInit();
	sprInit();

	return 0;
}

void hwShutdown() {
	if (psH == NULL) return;
	free(psH); psH = NULL;
}

void hwReset() {
	memset(psH, 0, 0x00010000);
	// i guess this is kinda a version, it's used by some bioses
	psHu32(0xf590) = 0x1201;
	gsReset();
}

u8  hwRead8(u32 mem) {
	u8 ret;

	switch (mem) {
		default:
			if (mem < 0x10010000)
				ret = psHu8(mem);
			else ret = 0;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware Read 8 at %x\n",mem);
#endif
			break;
	}

	return ret;
}

u16 hwRead16(u32 mem) {
	u16 ret;

	switch (mem) {
		case 0x10000000: ret = (u16)rcntRcount(0); break;
		case 0x10000010: ret = (u16)counters[0].mode; break;
		case 0x10000020: ret = (u16)counters[0].target; break;
		case 0x10000030: ret = (u16)counters[0].hold; break;

		case 0x10000800: ret = (u16)rcntRcount(1); break;
		case 0x10000810: ret = (u16)counters[1].mode; break;
		case 0x10000820: ret = (u16)counters[1].target; break;
		case 0x10000830: ret = (u16)counters[1].hold; break;

		case 0x10001000: ret = (u16)rcntRcount(2); break;
		case 0x10001010: ret = (u16)counters[2].mode; break;
		case 0x10001020: ret = (u16)counters[2].target; break;

		case 0x10001800: ret = (u16)rcntRcount(3); break;
		case 0x10001810: ret = (u16)counters[3].mode; break;
		case 0x10001820: ret = (u16)counters[3].target; break;

		default:
			if (mem < 0x10010000)
				ret = psHu16(mem);
			else ret = 0;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware Read 16 at %x\n",mem);
#endif
			break;
	}

	return ret;
}

static int b440;

u32 hwRead32(u32 mem) {
	u32 ret;

	//IPU regs
	if ((mem>=0x10002000) && (mem<=0x10002030))
		return IPU_read32(mem);

	switch (mem) {
		case 0x10000000: ret = (u16)rcntRcount(0); break;
		case 0x10000010: ret = (u16)counters[0].mode; break;
		case 0x10000020: ret = (u16)counters[0].target; break;
		case 0x10000030: ret = (u16)counters[0].hold; break;

		case 0x10000800: ret = (u16)rcntRcount(1); break;
		case 0x10000810: ret = (u16)counters[1].mode; break;
		case 0x10000820: ret = (u16)counters[1].target; break;
		case 0x10000830: ret = (u16)counters[1].hold; break;

		case 0x10001000: ret = (u16)rcntRcount(2); break;
		case 0x10001010: ret = (u16)counters[2].mode; break;
		case 0x10001020: ret = (u16)counters[2].target; break;

		case 0x10001800: ret = (u16)rcntRcount(3); break;
		case 0x10001810: ret = (u16)counters[3].mode; break;
		case 0x10001820: ret = (u16)counters[3].target; break;

#ifdef HW_LOG
		case 0x1000A000:
			ret = psHu32(mem);//dma2 chcr
			HW_LOG("Hardware read DMA2_CHCR 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A010:
			ret = psHu32(mem);//dma2 madr
			HW_LOG("Hardware read DMA2_MADR 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A020:
			ret = psHu32(mem);//dma2 qwc
			HW_LOG("Hardware readDMA2_QWC 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A030:
			ret = psHu32(mem);//dma2 taddr
			HW_LOG("Hardware read DMA2_TADDR 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A040:
			ret = psHu32(mem);//dma2 asr0
			HW_LOG("Hardware read DMA2_ASR0 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A050:
			ret = psHu32(mem);//dma2 asr1
			HW_LOG("Hardware read DMA2_ASR1 32bit at %lx, ret %lx\n", mem, ret);
			break;
		case 0x1000A080:
			ret = psHu32(mem);//dma2 saddr
			HW_LOG("Hardware read DMA2_SADDR 32 at %lx, ret %lx\n", mem, ret);
			break;
#endif

#ifdef HW_LOG
		case 0x1000e010: // DMAC_STAT
			HW_LOG("DMAC_STAT Read  32bit %x\n", psHu32(0xe010));
			return psHu32(0xe010);
#endif

#ifdef HW_LOG
		case 0x1000f000: // INTC_STAT
//			HW_LOG("INTC_STAT Read  32bit %x\n", psHu32(0xf000));
			return psHu32(0xf000);
		case 0x1000f010: // INTC_MASK
			HW_LOG("INTC_MASK Read  32bit %x\n", psHu32(0xf010));
			return psHu32(0xf010);
#endif

		case 0x1000f130:
			ret = 0;
			break;

		case 0x1000f410:
			ret = 0;
			break;

		case 0x1000f430:
			ret = 0;
			break;

		case 0x1000f440:
			b440++;
			switch (b440) {
				case 3: case 6: case 9: case 10:
				case 13: case 14:
				case 17: case 18:
				case 21: case 22:
				case 25: case 26:
				case 29: case 30:
					ret = 0; break;
				default:
					ret = 0x1f; break;
			}
			break;

		case 0x1000f520: // DMAC_ENABLER
#ifdef HW_LOG
			HW_LOG("DMAC_ENABLER Read 32bit %lx\n", psHu32(0xf590));
#endif
			return psHu32(0xf590);

		default:
			if ((mem & 0xffffff0f) == 0x1000f200) {
				ret = psxMemRead32(0xbd000000 | (mem & 0xf0)); // seems so
				//SysPrintf("[EE pc=%08X]", cpuRegs.pc);	//see also PsxMem.c
#ifdef HW_LOG
//				HW_LOG("sif %x Read 32bit %x\n", mem, ret);
#endif
				break;
			}

			if (mem < 0x10010000)
				ret = psHu32(mem);
			else ret = 0;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware Read 32 at %lx, ret %lx\n", mem, ret);
#endif
			break;
	}

	return ret;
}

u64 hwRead64(u32 mem) {
	u64 ret;

	//IPU regs
	if ((mem>=0x10002000) && (mem<=0x10002030))
		return IPU_read64(mem);

	switch (mem) {
		default:
			if (mem < 0x10010000)
				ret = psHu64(mem);
			else ret = 0;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware Read 64 at %x\n",mem);
#endif
			break;
	}
	return ret;
}

void hwRead128(u32 mem, u64 *out) {
	if (mem >= 0x10004000 && mem < 0x10008000) {
		ReadFIFO(mem, out); return;
	}

	switch (mem) {
		default:
			if (mem < 0x10010000) {
				out[0] = psHu64(mem);
				out[1] = psHu64(mem);
			}
#ifdef HW_LOG
			HW_LOG("Unknown Hardware Read 128 at %x\n",mem);
#endif
			break;
	}
}


#define DmaExec8(name, num) { \
	psHu8(mem) = (u8)value; \
	if (value & 0x1/* && psHu32(DMAC_CTRL) & 0x1*/) { \
		dma##name(); \
	} \
}

char sio_buffer[1024];
static int sio_count;

void hwWrite8(u32 mem, u8 value) {
	switch (mem) {
		case 0x1000f180:
			if (value == '\n') {
				sio_buffer[sio_count] = 0;
				SysPrintf("%s\n", sio_buffer);
				sio_count = 0;
			} else {
				if (sio_count < 1023) {
					sio_buffer[sio_count++] = value;
				}
			}
//			SysPrintf("%c", value);
			break;

		case 0x10008001: // dma0 - vif0
#ifdef DMA_LOG
			DMA_LOG("VIF0dma %lx\n", value);
#endif
			DmaExec8(VIF0, 0);
			break;

		case 0x10009001: // dma1 - vif1
#ifdef DMA_LOG
			DMA_LOG("VIF1dma %lx\n", value);
#endif
			DmaExec8(VIF1, 1);
			break;

		case 0x1000a001: // dma2 - gif
#ifdef DMA_LOG
			DMA_LOG("GSdma %lx\n", value);
#endif
			DmaExec8(GIF, 2);
			break;

		case 0x1000b001: // dma3 - fromIPU
#ifdef DMA_LOG
			DMA_LOG("IPU0dma %lx\n", value);
#endif
			DmaExec8(IPU0, 3);
			break;

		case 0x1000b401: // dma4 - toIPU
#ifdef DMA_LOG
			DMA_LOG("IPU1dma %lx\n", value);
#endif
			DmaExec8(IPU1, 4);
			break;

		case 0x1000c001: // dma5 - sif0
#ifdef DMA_LOG
			DMA_LOG("SIF0dma %lx\n", value);
#endif
//			if (value == 0) psxSu32(0x30) = 0x40000;
			DmaExec8(SIF0, 5);
			break;

		case 0x1000c401: // dma6 - sif1
#ifdef DMA_LOG
			DMA_LOG("SIF1dma %lx\n", value);
#endif
			DmaExec8(SIF1, 6);
			break;

		case 0x1000c801: // dma7 - sif2
#ifdef DMA_LOG
			DMA_LOG("SIF2dma %lx\n", value);
#endif
			DmaExec8(SIF2, 7);
			break;

		case 0x1000d001: // dma8 - fromSPR
#ifdef DMA_LOG
			DMA_LOG("fromSPRdma8 %lx\n", value);
#endif
			DmaExec8(SPR0, 8);
			break;

		case 0x1000d401: // dma9 - toSPR
#ifdef DMA_LOG
			DMA_LOG("toSPRdma8 %lx\n", value);
#endif
			DmaExec8(SPR1, 9);
			break;

		default:
			if (mem < 0x10010000)
				psHu8(mem) = value;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware write 8 at %x with value %x\n", mem, value);
#endif
			break;
	}
}

void hwWrite16(u32 mem, u16 value) {
	switch (mem) {
		default:
			if (mem < 0x10010000)
				psHu16(mem) = value;
#ifdef HW_LOG
		    HW_LOG("Unknown Hardware write 16 at %x with value %x\n",mem,value);
#endif
			break;
	}
}

#define DmaExec(name, num) { \
	psHu32(mem) = (u32)value; \
	if (psHu32(mem) & 0x100/* && (psHu32(DMAC_CTRL) & 0x1)*/) { \
		dma##name(); \
	} \
}

void hwWrite32(u32 mem, u32 value) {
	int i;

	//IPU regs
	if ((mem>=0x10002000) && (mem<=0x10002030)) {
    	psHu32(mem) = value;
		IPU_write32(mem,value);
		return;
	}

	if ((mem>=0x10003800) && (mem<0x10003c00)) {
		vif0Write32(mem, value); return;
	}
	if ((mem>=0x10003c00) && (mem<0x10004000)) {
		vif1Write32(mem, value); return;
	}

	switch (mem) {
		case 0x10000000: rcntWcount(0, value); break;
		case 0x10000010: rcntWmode(0, value); break;
		case 0x10000020: rcntWtarget(0, value); break;
		case 0x10000030: rcntWhold(0, value); break;

		case 0x10000800: rcntWcount(1, value); break;
		case 0x10000810: rcntWmode(1, value); break;
		case 0x10000820: rcntWtarget(1, value); break;
		case 0x10000830: rcntWhold(1, value); break;

		case 0x10001000: rcntWcount(2, value); break;
		case 0x10001010: rcntWmode(2, value); break;
		case 0x10001020: rcntWtarget(2, value); break;

		case 0x10001800: rcntWcount(3, value); break;
		case 0x10001810: rcntWmode(3, value); break;
		case 0x10001820: rcntWtarget(3, value); break;

		case GIF_STAT: // stat is readonly
			return;

		case 0x10003830: // VIF0_MARK
		case 0x10003c30: // VIF1_MARK

			break;

		case 0x10008000: // dma0 - vif0
#ifdef DMA_LOG
			DMA_LOG("VIF0dma %lx\n", value);
#endif
			DmaExec(VIF0, 0);
			break;

// Latest Fix for Florin by asadr (VIF1)
		case 0x10009000: // dma1 - vif1 - chcr
#ifdef DMA_LOG
			DMA_LOG("VIF1dma CHCR %lx\n", value);
#endif
			DmaExec(VIF1, 1);
			break;
#ifdef HW_LOG
		case 0x10009010: // dma1 - vif1 - madr
			HW_LOG("VIF1dma Madr %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 madr
			break;
		case 0x10009020: // dma1 - vif1 - qwc
			HW_LOG("VIF1dma QWC %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 qwc
			break;
		case 0x10009030: // dma1 - vif1 - tadr
			HW_LOG("VIF1dma TADR %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 tadr
			break;
		case 0x10009040: // dma1 - vif1 - asr0
			HW_LOG("VIF1dma ASR0 %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 asr0
			break;
		case 0x10009050: // dma1 - vif1 - asr1
			HW_LOG("VIF1dma ASR1 %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 asr1
			break;
		case 0x10009080: // dma1 - vif1 - sadr
			HW_LOG("VIF1dma SADR %lx\n", value);
			psHu32(mem & 0xffff) = value;//dma1 sadr
			break;
#endif
// ---------------------------------------------------

		case 0x1000a000: // dma2 - gif
#ifdef DMA_LOG
			DMA_LOG("GSdma %lx\n", value);
#endif
			DmaExec(GIF, 2);
			break;
#ifdef HW_LOG
	    case 0x1000a010:
		    psHu32(mem) = value;//dma2 madr
			HW_LOG("Hardware write DMA2_MADR 32bit at %x with value %x\n",mem,value);
		    break;
	    case 0x1000a020:
            psHu32(mem) = value;//dma2 qwc
		    HW_LOG("Hardware write DMA2_QWC 32bit at %x with value %x\n",mem,value);
		    break;
	    case 0x1000a030:
            psHu32(mem) = value;//dma2 taddr
		    HW_LOG("Hardware write DMA2_TADDR 32bit at %x with value %x\n",mem,value);
		    break;
	    case 0x1000a040:
            psHu32(mem) = value;//dma2 asr0
		    HW_LOG("Hardware write DMA2_ASR0 32bit at %x with value %x\n",mem,value);
		    break;
	    case 0x1000a050:
            psHu32(mem) = value;//dma2 asr1
		    HW_LOG("Hardware write DMA2_ASR1 32bit at %x with value %x\n",mem,value);
		    break;
	    case 0x1000a080:
            psHu32(mem) = value;//dma2 saddr
		    HW_LOG("Hardware write DMA2_SADDR 32bit at %x with value %x\n",mem,value);
		    break;
#endif
		case 0x1000b000: // dma3 - fromIPU
#ifdef DMA_LOG
			DMA_LOG("IPU0dma %lx\n", value);
#endif
			DmaExec(IPU0, 3);
			break;
#ifdef HW_LOG
		case 0x1000b010:
	   		psHu32(mem) = value;//dma2 madr
			HW_LOG("Hardware write IPU0DMA_MADR 32bit at %x with value %x\n",mem,value);
			break;
		case 0x1000b020:
    		psHu32(mem) = value;//dma2 madr
			HW_LOG("Hardware write IPU0DMA_QWC 32bit at %x with value %x\n",mem,value);
       		break;
		case 0x1000b030:
			psHu32(mem) = value;//dma2 tadr
			HW_LOG("Hardware write IPU0DMA_TADR 32bit at %x with value %x\n",mem,value);
			break;
		case 0x1000b080:
			psHu32(mem) = value;//dma2 saddr
			HW_LOG("Hardware write IPU0DMA_SADDR 32bit at %x with value %x\n",mem,value);
			break;
#endif
		case 0x1000b400: // dma4 - toIPU
#ifdef DMA_LOG
			DMA_LOG("IPU1dma %lx\n", value);
#endif
			DmaExec(IPU1, 4);
			break;
#ifdef HW_LOG
		case 0x1000b410:
    		psHu32(mem) = value;//dma2 madr
			HW_LOG("Hardware write IPU1DMA_MADR 32bit at %x with value %x\n",mem,value);
       		break;
		case 0x1000b420:
    		psHu32(mem) = value;//dma2 madr
			HW_LOG("Hardware write IPU1DMA_QWC 32bit at %x with value %x\n",mem,value);
       		break;
		case 0x1000b430:
			psHu32(mem) = value;//dma2 tadr
			HW_LOG("Hardware write IPU1DMA_TADR 32bit at %x with value %x\n",mem,value);
			break;
		case 0x1000b480:
			psHu32(mem) = value;//dma2 saddr
			HW_LOG("Hardware write IPU1DMA_SADDR 32bit at %x with value %x\n",mem,value);
			break;
#endif

		case 0x1000c000: // dma5 - sif0
#ifdef DMA_LOG
			DMA_LOG("SIF0dma %lx\n", value);
#endif
//			if (value == 0) psxSu32(0x30) = 0x40000;
			DmaExec(SIF0, 5);
			break;

		case 0x1000c400: // dma6 - sif1
#ifdef DMA_LOG
			DMA_LOG("SIF1dma %lx\n", value);
#endif
			DmaExec(SIF1, 6);
			break;

#ifdef HW_LOG
		case 0x1000c420: // dma6 - sif1 - qwc
			HW_LOG("SIF1dma QWC = %lx\n", value);
			psHu32(mem) = value;
			break;
#endif

#ifdef HW_LOG
		case 0x1000c430: // dma6 - sif1 - tadr
			HW_LOG("SIF1dma TADR = %lx\n", value);
			psHu32(mem) = value;
			break;
#endif

		case 0x1000c800: // dma7 - sif2
#ifdef DMA_LOG
			DMA_LOG("SIF2dma %lx\n", value);
#endif
			DmaExec(SIF2, 7);
			break;

		case 0x1000d000: // dma8 - fromSPR
#ifdef DMA_LOG
			DMA_LOG("fromSPRdma %lx\n", value);
#endif
			DmaExec(SPR0, 8);
			break;

		case 0x1000d400: // dma9 - toSPR
#ifdef DMA_LOG
			DMA_LOG("toSPRdma %lx\n", value);
#endif
			DmaExec(SPR1, 9);
			break;

#ifdef HW_LOG
		case 0x1000e000: // DMAC_CTRL
			HW_LOG("DMAC_CTRL Write 32bit %x\n", value);
			psHu32(mem) = value;
			break;
#endif

		case 0x1000e010: // DMAC_STAT
#ifdef HW_LOG
			HW_LOG("DMAC_STAT Write 32bit %x\n", value);
#endif
			psHu16(0xe010)&= ~(value & 0xffff); // clear on 1
			value = value >> 16;
			for (i=0; i<16; i++) { // reverse on 1
				if (value & (1<<i)) {
					if (psHu16(0xe012) & (1<<i)) psHu16(0xe012)&= ~(1<<i);
					else psHu16(0xe012)|= 1<<i;
				}
			}
			cpuTestDMACInts();
			break;

		case 0x1000f000: // INTC_STAT
#ifdef HW_LOG
			HW_LOG("INTC_STAT Write 32bit %x\n", value);
#endif
			psHu32(0xf000)&=~value;			
			break;

		case 0x1000f010: // INTC_MASK
#ifdef HW_LOG
			HW_LOG("INTC_MASK Write 32bit %x\n", value);
#endif
			for (i=0; i<16; i++) { // reverse on 1
				if (value & (1<<i)) {
					if (psHu32(0xf010) & (1<<i)) psHu32(0xf010)&= ~(1<<i);
					else psHu32(0xf010)|= 1<<i;
				}
			}
			cpuTestINTCInts();
			break;

#ifdef HW_LOG
		case 0x1000f590: // DMAC_ENABLEW
			HW_LOG("DMAC_ENABLEW Write 32bit %lx\n", value);
			psHu32(mem) = value;
			return;
#endif

		default:
			if ((mem & 0xffffff0f) == 0x1000f200) {
				psxMemWrite32(0xbd000001 | (mem & 0xf0), value); // seems so
//				SysPrintf("[EE pc=%08X]", cpuRegs.pc);	//see also PsxMem.c
#ifdef HW_LOG
				HW_LOG("sif %x Write 32bit %x\n", mem, value);
#endif
				break;
			}

			if (mem < 0x10010000)
				psHu32(mem) = value;
#ifdef HW_LOG
			HW_LOG("Unknown Hardware write 32 at %x with value %x (%x)\n", mem, value, cpuRegs.CP0.n.Status);
#endif
			break;
	}
}

void hwWrite64(u32 mem, u64 value) {
	u32 val32;
	int i;

	if ((mem>=0x10002000) && (mem<=0x10002030)) {
		IPU_write64(mem, value);
		return;
	}
	if ((mem>=0x10003800) && (mem<0x10003c00)) {
		vif0Write32(mem, value); return;
	}
	if ((mem>=0x10003c00) && (mem<0x10004000)) {
		vif1Write32(mem, value); return;
	}

	switch (mem) {
		case GIF_MODE:
			if (value & 0x1) psHu32(GIF_STAT)|= 0x1;
			else psHu32(GIF_STAT)&= ~0x1;
			if (value & 0x4) psHu32(GIF_STAT)|= 0x4;
			else psHu32(GIF_STAT)&= ~0x4;
			break;

		case GIF_STAT: // stat is readonly
			return;

		case 0x1000a000: // dma2 - gif
#ifdef DMA_LOG
			DMA_LOG("GSdma %lx\n", value);
#endif
			DmaExec(GIF, 2);
			break;

#ifdef HW_LOG
		case 0x1000e000: // DMAC_CTRL
			HW_LOG("DMAC_CTRL Write 64bit %x\n", value);
			psHu64(mem) = value;
			break;
#endif

		case 0x1000e010: // DMAC_STAT
#ifdef HW_LOG
			HW_LOG("DMAC_STAT Write 64bit %x\n", value);
#endif
			val32 = (u32)value;
			psHu16(0xe010)&= ~(val32 & 0xffff); // clear on 1
			val32 = val32 >> 16;
			for (i=0; i<16; i++) { // reverse on 1
				if (val32 & (1<<i)) {
					if (psHu16(0xe012) & (1<<i)) psHu16(0xe012)&= ~(1<<i);
					else psHu16(0xe012)|= 1<<i;
				}
			}
			cpuTestDMACInts();
			break;

		case 0x1000f000: // INTC_STAT
#ifdef HW_LOG
			HW_LOG("INTC_STAT Write 64bit %x\n", value);
#endif
			psHu32(0xf000)&=~value;			
			break;

		case 0x1000f010: // INTC_MASK
#ifdef HW_LOG
			HW_LOG("INTC_MASK Write 32bit %x\n", value);
#endif
			for (i=0; i<16; i++) { // reverse on 1
				if (value & (1<<i)) {
					if (psHu32(0xf010) & (1<<i)) psHu32(0xf010)&= ~(1<<i);
					else psHu32(0xf010)|= 1<<i;
				}
			}
			cpuTestINTCInts();
			break;

		default:
			if (mem < 0x10010000)
				psHu64(mem) = value;
#ifdef HW_LOG
		    HW_LOG("Unknown Hardware write 64 at %x with value %x (status=%x)\n",mem,value, cpuRegs.CP0.n.Status);
#endif
			break;
	}
}

void hwWrite128(u32 mem, u64 *value) {
	if (mem >= 0x10004000 && mem < 0x10008000) {
		WriteFIFO(mem, value); return;
	}

	switch (mem) {
		default:
			if (mem < 0x10010000) {
				psHu64(mem  ) = value[0];
				psHu64(mem+8) = value[1];
			}
#ifdef HW_LOG
		    HW_LOG("Unknown Hardware write 128 at %x with value %x_%x (status=%x)\n", mem, value[1], value[0], cpuRegs.CP0.n.Status);
#endif
			break;
	}
}

int  intcInterrupt() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) != 0x10001) return 0;

	if ((psHu32(INTC_STAT) & psHu32(INTC_MASK)) == 0) {
		SysPrintf("*PCSX2*: intcInterrupt already cleared\n");
		return 1;
	}
#ifdef HW_LOG
	HW_LOG("intcInterrupt %x\n", psHu32(INTC_STAT) & psHu32(INTC_MASK));
#endif
	cpuException(0x800400, 0);

	return 1;
}

int  dmacInterrupt() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) != 0x10001) return 0;

	if ((psHu16(0xe012) & psHu16(0xe010) || 
		 psHu16(0xe010) & 0x8000) == 0) return 1;
#ifdef HW_LOG
	HW_LOG("dmacInterrupt %x\n", (psHu16(0xe012) & psHu16(0xe010) || 
								  psHu16(0xe010) & 0x8000));
#endif
	cpuException(0x800800, 0);

	return 1;
}

void *dmaGetAddr(u32 addr) {
	u8 *ptr;

/*#ifdef DMA_LOG
	if (addr & 0xf) { DMA_LOG("*PCSX2*: DMA address not 128bit aligned: %8.8x\n", addr); }
#endif*/
	if (addr & 0xf) { SysPrintf("*PCSX2*: DMA address not 128bit aligned: %8.8x\n", addr); }

	if (addr & 0x80000000) {
		return (void*)&psS[addr & 0x3ff0];
	}

	ptr = (u8*)memLUTR[addr >> 12];
	if (ptr == NULL) {
		SysPrintf("*PCSX2*: DMA error: %8.8x\n", addr);
		return NULL;
	}
	return (void*)(ptr + (addr & 0xff0));
}

void hwIntcIrq(int n) {
/*	if (psHu32(INTC_STAT) & (1<<n)) {
		SysPrintf("*PCSX2*: psHu32(INTC_STAT) n=%d already set\n", n);
	}*/
	psHu32(INTC_STAT)|= 1<<n;
	cpuTestINTCInts();
}

void hwDmacIrq(int n) {
/*	if (psHu32(DMAC_STAT) & (1<<n)) {
		SysPrintf("*PCSX2*: psHu32(DMAC_STAT) n=%d already set\n", n);
	}*/
	psHu32(DMAC_STAT)|= 1<<n;
	cpuTestDMACInts();
}

