/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Common.h"
#include "IPU.h"
#include "Mpeg.h"
#include "yuv2rgb.h"

#ifdef __WIN32__
#define FASTCALL	__fastcall
#else
#define FASTCALL
#endif

int fifo_wread(void *value, int size);
int fifo_wread2(void *value, int size);
void fifo_wclear();
int fifo_cwsize = 0;
int readwpos = 0;
int writewpos = 0;
u32 fifo_input[32];

int fifo_rwrite(void* pMem, int size);
int fifo_rread(void *value, int size);
void fifo_rclear();
int fifo_crsize = 0;
int readrpos = 0;
int writerpos = 0;
u32 fifo_output[32];

u32 FASTCALL BigEndian(u32 a){
	return ((((a >> 24) & 0xFF) <<  0) + (((a >> 16) & 0xFF) <<  8) +
			(((a >>  8) & 0xFF) << 16) + (((a >>  0) & 0xFF) << 24));
}

// Color conversion stuff
char convert_data_buffer[0x1C];
convert_init_t	convert_init={convert_data_buffer, 0x1C};
convert_t *convert;

// Quantization matrix
u8 niq[64],			//non-intraquant matrix
	iq[64];			//intraquant matrix
u16 vqclut[16];			//clut conversion table
u8 th0=0, th1=0;		//thresholds for color conversions
int coded_block_pattern=0;
struct macroblock_8  mb8;
struct macroblock_16 mb16;
struct macroblock_rgb32 rgb32;
struct macroblock_rgb16 rgb16;
u32 data = 0;
u8 indx4[16*16/2];
u32	mpeg2_inited;		//mpeg2_idct_init() must be called only once
u8 readbits[60];	//local buffer with 128*2 bits
u8 PCT[]={'r', 'I', 'P', 'B', 'D', '-', '-', '-'};
static decoder_t decoder;						//static, only to place it in bss
decoder_t tempdec;

extern u8 mpeg2_scan_norm[64];
extern u8 mpeg2_scan_alt[64];

#define SATURATE_4BITS(val)		((val)>15 ? 15 : (val))

/////////////////////////////////////////
int GETMOREFROMFIFO()
{
	if ((int)ipuRegs.bp.BP>=128) {
		if(ipuRegs.bp.FP==2) {
			*(u32*)readbits=*((u32*)readbits+4);
			memcpy(readbits+4,readbits+20,16);
		}
		if(ipuRegs.bp.FP > 0) ipuRegs.bp.FP--;
		else ipuRegs.bp.FP = 0;
        ipuRegs.bp.BP-=128;
	}
	if (ipuRegs.bp.FP == 0){
		*(u32*)readbits=*((u32*)readbits+4);
		if(fifo_cwsize <= 0){
			if(((DMACh*)&psH[0xb400])->qwc > 0) dmaIPU1((DMACh*)&psH[0xb400]);
			else {
				SysPrintf("Out of Fifo's!!!!\n");
				return -1;
			}
		}
		fifo_wread(readbits+4, 1);
		ipuRegs.bp.FP = 1;
	}
	return 0;
}

int getBits(u8 *address, u32 size, u32 advance){
	register u32 howmuch, mask, shift=0, oldsize=size;
	if (ipuRegs.bp.BP & 7){//size <= 32 !!!
		address--;
		while (size){
			if (shift==0){
				*++address=0;
				shift=8;
			}
			if(GETMOREFROMFIFO() < 0) return -1;
			howmuch = min(min(8-(ipuRegs.bp.BP&7), 128-ipuRegs.bp.BP), min(size, shift));
			mask =((0xFF >> (ipuRegs.bp.BP&7)) << (8-howmuch-(ipuRegs.bp.BP&7))) & 0xFF;
			mask&=readbits[(ipuRegs.bp.BP+32)>>3];					//bp goes signed;)
			mask>>=8-howmuch-(ipuRegs.bp.BP&7);
			ipuRegs.bp.BP+=howmuch;
			size -= howmuch;
			shift -= howmuch;
			*address |= mask << shift;
		}
	}else
	{
		while (size){
			if(GETMOREFROMFIFO() < 0) return -1;
			howmuch=min(128-ipuRegs.bp.BP, size);
			memcpy(address, readbits+((ipuRegs.bp.BP+32)>>3), howmuch>>3);
			ipuRegs.bp.BP+=howmuch;
			size-=howmuch;
			address+=howmuch>>3;
		}
	}

	if((int)ipuRegs.bp.BP >= 128 && ipuRegs.bp.FP<2)
	{
        if(fifo_cwsize <= 0){
			if(((DMACh*)&psH[0xb400])->qwc > 0) dmaIPU1((DMACh*)&psH[0xb400]);
			else return -1;
		}
		ipuRegs.bp.FP += fifo_wread(readbits+20, 1);
		if(fifo_cwsize <= 0){
			if(((DMACh*)&psH[0xb400])->qwc > 0) dmaIPU1((DMACh*)&psH[0xb400]);
			else return -1;
		}
	}
	if (!advance) ipuRegs.bp.BP-=oldsize;
	return 0;
}
/////////////////////////////////////////////////////////
void ipuBCLR(u32 val) {
	fifo_wclear();
	ipuRegs.bp.BP = val & 0x7F;
	ipuRegs.bp.FP = 0;
	pReg.stopIN = 0;
#ifdef IPU_LOG
	IPU_LOG("Clear IPU input FIFO. Set Bit offset=0x%X\n", ipuRegs.bp.BP);
#endif

	ipuRegs.ctrl.BUSY = 0;
	ipuRegs.cmd.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void ipuIDEC(u32 val) {
	tIPU_CMD_IDEC idec={0, 0, 0, 0, 0, 0, 0, 0, 0};
	
	*(u32*)&idec=val;
#ifdef IPU_LOG
						IPU_LOG("IPU IDEC command.\n");
	if (idec.FB){		IPU_LOG(" Skip %d bits.",idec.FB);}
						IPU_LOG(" Quantizer step code=0x%X.",idec.QSC);
	if (idec.DTD==0){	IPU_LOG(" Does not decode DT.");
	}else{				IPU_LOG(" Decodes DT.");}
	if (idec.SGN==0){	IPU_LOG(" No bias.");
	}else{				IPU_LOG(" Bias=128.");}
	if (idec.DTE==1){	IPU_LOG(" Dither Enabled.");}
	if (idec.OFM==0){	IPU_LOG(" Output format is RGB32.");
	}else{				IPU_LOG(" Output format is RGB16.");}
						IPU_LOG("\n");
#endif
	if(pReg.stopOUT == 0)
	{
		ipuRegs.bp.BP+= idec.FB;//skip FB bits
		//from IPU_CTRL
		ipuRegs.ctrl.PCT = I_TYPE; //Intra DECoding;)
		decoder.coding_type		=ipuRegs.ctrl.PCT;
		decoder.mpeg1			=ipuRegs.ctrl.MP1;
		decoder.q_scale_type	=ipuRegs.ctrl.QST;
		decoder.intra_vlc_format=ipuRegs.ctrl.IVF;
		decoder.scan			=ipuRegs.ctrl.AS ? mpeg2_scan_alt: mpeg2_scan_norm;
		decoder.intra_dc_precision=ipuRegs.ctrl.IDP;
		//from IDEC value
		decoder.quantizer_scale	=idec.QSC;
		decoder.frame_pred_frame_dct=!idec.DTD;
		decoder.sgn				=idec.SGN;
		decoder.dte				=idec.DTE;
		decoder.ofm				=idec.OFM;
		//other stuff
		decoder.dcr				  =1;//resets DC prediction value
		decoder.intra_quantizer_matrix		=(u8*)iq;
		decoder.non_intra_quantizer_matrix	=(u8*)niq;
		decoder.picture_structure = FRAME_PICTURE;	//default: progressive...my guess:P
		decoder.mb8	=&mb8;
		decoder.mb16=&mb16;
		decoder.rgb32=&rgb32;
		decoder.rgb16=&rgb16;
		decoder.stride=16;
	}

	//TODO: 1. use/fix internal code for yuv2rgb conversion & use mmx for it also
	//TODO: 2. clean up the code
	ipuRegs.ctrl.SCD=mpeg2sliceIDEC(&decoder, readbits);
	// Complete transfer was not done and we ran out of space in FIFO
	if(pReg.stopOUT > 0) {
		ipuRegs.ctrl.BUSY = 1;
		return;
	}
	ipuRegs.ctrl.ECD=!ipuRegs.ctrl.SCD;

	coded_block_pattern=decoder.coded_block_pattern;

	if(ipuRegs.ctrl.SCD){
		int i;
		for (i=0; i<2; i++) {
			u8 byte;
			getBits(&byte, 8, 0);
			if (byte == 0) break;
			ipuRegs.bp.BP+= 8;
		}
	}
	ipuRegs.bp.BP-=32;//bitstream_init takes 32 bits

	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void ipuBDEC(u32 val) {
	tIPU_CMD_BDEC bdec={0, 0, 0, 0, 0, 0, 0, 0};
	u8 bit8=0;
	*(u32*)&bdec=val;

#ifdef IPU_LOG
							IPU_LOG("IPU BDEC(macroblock decode) command.\n");
	if (bdec.FB){			IPU_LOG(" Skip 0x%X bits.", bdec.FB);}
	if (bdec.MBI){			IPU_LOG(" Intra MB.");}
	else{					IPU_LOG(" Non-intra MB.");}
	if (bdec.DCR){			IPU_LOG(" Resets DC prediction value.");}
	else{					IPU_LOG(" Doesn't reset DC prediction value.");}
	if (bdec.DT){			IPU_LOG(" Use field DCT.");}
	else{					IPU_LOG(" Use frame DCT.");}
							IPU_LOG(" Quantiser step=0x%X\n",bdec.QSC);
#endif
	ipuRegs.bp.BP+= bdec.FB;//skip FB bits
	decoder.coding_type		= I_TYPE;
	decoder.mpeg1			=ipuRegs.ctrl.MP1;
	decoder.q_scale_type	=ipuRegs.ctrl.QST;
	decoder.intra_vlc_format=ipuRegs.ctrl.IVF;
	decoder.scan			=ipuRegs.ctrl.AS ? mpeg2_scan_alt: mpeg2_scan_norm;
	decoder.intra_dc_precision=ipuRegs.ctrl.IDP;
	//from BDEC value
	/* JayteeMaster: the quantizer (linear/non linear) depends on the q_scale_type */
	decoder.quantizer_scale  =decoder.q_scale_type?non_linear_quantizer_scale [bdec.QSC]:bdec.QSC<<1;
	decoder.macroblock_modes =bdec.DT ? DCT_TYPE_INTERLACED : 0;
	decoder.dcr				 =bdec.DCR;
	decoder.macroblock_modes|=bdec.MBI ? MACROBLOCK_INTRA : MACROBLOCK_PATTERN;
	//other stuff
	decoder.intra_quantizer_matrix		=(u8*)iq;
	decoder.non_intra_quantizer_matrix	=(u8*)niq;
	decoder.picture_structure = FRAME_PICTURE;	//default: progressive...my guess:P
	decoder.mb8	=&mb8;
	decoder.mb16=&mb16;
	decoder.rgb32=&rgb32;
	decoder.rgb16=&rgb16;
	decoder.stride=16;

	memset(&mb8, 0, sizeof(struct macroblock_8));
	memset(&mb16, 0, sizeof(struct macroblock_16));

	ipuRegs.ctrl.ECD=mpeg2_slice(&decoder, readbits,1);
	ipuRegs.ctrl.SCD=0;
	coded_block_pattern=decoder.coded_block_pattern;

	ipuRegs.bp.BP+=decoder.bitstream_bits-16;
	getBits((u8*)&bit8, 8, 0);  // Get next 8bits 
	if (bit8==0) ipuRegs.ctrl.SCD = 1;
	
	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void ipuVDEC(u32 val) {
	if (fifo_cwsize <= 0 && ((DMACh*)&psH[0xb400])->qwc == 0) {
			ipuRegs.ctrl.BUSY = 1;
			ipuRegs.cmd.BUSY = 1; 
			return;
	}
	ipuRegs.bp.BP+= val & 0x3F;
	getBits((u8*)&decoder.bitstream_buf, 32, 0);
	
	decoder.bitstream_bits = -16;	 
	decoder.bitstream_buf=BigEndian(decoder.bitstream_buf);
    	
	switch((val >> 26) & 3){
		case 0://Macroblock Address Increment
			decoder.mpeg1			=ipuRegs.ctrl.MP1;
			ipuRegs.cmd.DATA= get_macroblock_address_increment(&decoder);
			ipuRegs.bp.BP+=(decoder.bitstream_bits+16);
			ipuRegs.cmd.DATA = (ipuRegs.cmd.DATA & 0xFFFF) | ((decoder.bitstream_bits+16) << 16);
			ipuRegs.ctrl.ECD = (ipuRegs.cmd.DATA==0);
			break;
		case 1://Macroblock Type	//known issues: no error detected
			decoder.frame_pred_frame_dct=1;//prevent DCT_TYPE_INTERLACED
			decoder.coding_type		=ipuRegs.ctrl.PCT;
			ipuRegs.cmd.DATA=get_macroblock_modes(&decoder);
			ipuRegs.bp.BP+=(decoder.bitstream_bits+16);
			ipuRegs.cmd.DATA = (ipuRegs.cmd.DATA & 0xFFFF) | ((decoder.bitstream_bits+16) << 16);
			ipuRegs.ctrl.ECD = (ipuRegs.cmd.DATA==0);
			break;
		case 2://Motion Code		//known issues: no error detected
			ipuRegs.cmd.DATA=get_motion_delta(&decoder,0);
			ipuRegs.bp.BP+=(decoder.bitstream_bits+16);
			ipuRegs.cmd.DATA = (ipuRegs.cmd.DATA & 0xFFFF) | ((decoder.bitstream_bits+16) << 16);
			ipuRegs.ctrl.ECD = (ipuRegs.cmd.DATA==0);
			break;
		case 3://DMVector
			ipuRegs.cmd.DATA=get_dmv(&decoder);
			ipuRegs.bp.BP+=(decoder.bitstream_bits+16);
			ipuRegs.cmd.DATA = (ipuRegs.cmd.DATA & 0xFFFF) | ((decoder.bitstream_bits+16) << 16);
			break;
	}
	
#ifdef IPU_LOG
	IPU_LOG("IPU VDEC command data %x(%x). Skip 0x%X bits/Table=%d (%s)\n",
		ipuRegs.cmd.DATA&0xffff,(decoder.bitstream_bits+16) << 16,val & 0x3f, (val >> 26) & 3, (val >> 26) & 1 ?
		((val >> 26) & 2 ? "DMV" : "MBT") : (((val >> 26) & 2 ? "MC" : "MBAI")));
#endif
	
	ipuRegs.ctrl.BUSY = 0;
	ipuRegs.cmd.BUSY = 0; 
	hwIntcIrq(INTC_IPU);
}

void ipuFDEC(u32 val) {
#ifdef IPU_LOG
	IPU_LOG("IPU FDEC command. Skip 0x%X bits, FIFO 0x%X bytes, BP 0x%X, FP %d, CHCR 0x%x\n",
		val & 0x3f,fifo_cwsize,(int)ipuRegs.bp.BP,ipuRegs.bp.FP,((DMACh*)&psH[0xb400])->chcr);
#endif
	if (fifo_cwsize <= 0 && ((DMACh*)&psH[0xb400])->qwc == 0) {
			ipuRegs.ctrl.BUSY = 1;
			ipuRegs.cmd.BUSY = 1; 
			return;
	}

	ipuRegs.bp.BP+= val & 0x3F;
	getBits((u8*)&ipuRegs.cmd.DATA, 32, 0);
	ipuRegs.cmd.DATA = BigEndian(ipuRegs.cmd.DATA);
	ipuRegs.ctrl.BUSY = 0;
	ipuRegs.cmd.BUSY = 0; 
	hwIntcIrq(INTC_IPU);
}
void ipuSETIQ(u32 val) {
	int i;

#ifdef IPU_LOG
	IPU_LOG("IPU SETIQ command.\n");
#endif
#ifdef IPU_LOG
	if (val & 0x3f){
		IPU_LOG("Skip %d bits.\n", val & 0x3f);
	}
#endif
	ipuRegs.bp.BP+= val & 0x3F;
	if ((val >> 27) & 1){
		getBits((u8*)niq, 512, 1); // 8*8*8
#ifdef IPU_LOG
		IPU_LOG("Read non-intra quantisation matrix from IPU FIFO.\n");
		for (i=0; i<8; i++){
			IPU_LOG("%02X %02X %02X %02X %02X %02X %02X %02X\n",
				niq[i*8+0], niq[i*8+1], niq[i*8+2], niq[i*8+3],
				niq[i*8+4], niq[i*8+5], niq[i*8+6], niq[i*8+7]);
		}
#endif
	}else{
		getBits((u8*)iq, 512, 1);
#ifdef IPU_LOG
		IPU_LOG("Read intra quantisation matrix from IPU FIFO.\n");
		for (i=0; i<8; i++){
			IPU_LOG("%02X %02X %02X %02X %02X %02X %02X %02X\n",
				iq[i*8+0], iq[i*8+1], iq[i*8+2], iq[i*8+3],
				iq[i*8+4], iq[i*8+5], iq[i*8+6], iq[i*8+7]);
		}
#endif
	}
	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void ipuSETVQ(u32 val) {
	getBits((u8*)vqclut, 256, 1); // 16*2*8
#ifdef IPU_LOG
	IPU_LOG("IPU SETVQ command.\nRead VQCLUT table from IPU FIFO.\n");
	IPU_LOG(
		"%02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d "
		"%02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d\n"
		"%02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d "
		"%02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d %02d:%02d:%02d\n",
		vqclut[0] >> 10, (vqclut[0] >> 5) & 0x1F, vqclut[0] & 0x1F,
		vqclut[1] >> 10, (vqclut[1] >> 5) & 0x1F, vqclut[1] & 0x1F,
		vqclut[2] >> 10, (vqclut[2] >> 5) & 0x1F, vqclut[2] & 0x1F,
		vqclut[3] >> 10, (vqclut[3] >> 5) & 0x1F, vqclut[3] & 0x1F,
		vqclut[4] >> 10, (vqclut[4] >> 5) & 0x1F, vqclut[4] & 0x1F,
		vqclut[5] >> 10, (vqclut[5] >> 5) & 0x1F, vqclut[5] & 0x1F,
		vqclut[6] >> 10, (vqclut[6] >> 5) & 0x1F, vqclut[6] & 0x1F,
		vqclut[7] >> 10, (vqclut[7] >> 5) & 0x1F, vqclut[7] & 0x1F,
		vqclut[8] >> 10, (vqclut[8] >> 5) & 0x1F, vqclut[8] & 0x1F,
		vqclut[9] >> 10, (vqclut[9] >> 5) & 0x1F, vqclut[9] & 0x1F,
		vqclut[10] >> 10, (vqclut[10] >> 5) & 0x1F, vqclut[10] & 0x1F,
		vqclut[11] >> 10, (vqclut[11] >> 5) & 0x1F, vqclut[11] & 0x1F,
		vqclut[12] >> 10, (vqclut[12] >> 5) & 0x1F, vqclut[12] & 0x1F,
		vqclut[13] >> 10, (vqclut[13] >> 5) & 0x1F, vqclut[13] & 0x1F,
		vqclut[14] >> 10, (vqclut[14] >> 5) & 0x1F, vqclut[14] & 0x1F,
		vqclut[15] >> 10, (vqclut[15] >> 5) & 0x1F, vqclut[15] & 0x1F);
#endif
	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

// IPU Transfers are split into 8Qwords so we need to send ALL the data
void ipuCSC(u32 val) {
	tIPU_CMD_CSC  csc ={0, 0, 0, 0, 0};
	int i, size = 0;

	if(pReg.stopOUT == 0 && pReg.stopIN == 0) *(u32*)&csc=val;
#ifdef IPU_LOG
	IPU_LOG("IPU CSC(Colorspace conversion from YCbCr) command.\n");
	if (csc.OFM){	IPU_LOG("Output format is RGB16. ");}
	else{			IPU_LOG("Output format is RGB32. ");}
	if (csc.DTE){	IPU_LOG("Dithering enabled.");	}
#endif
	for (i=0; i<(int)csc.MBC; i++){
		if(pReg.stopOUT == 0 && pReg.stopIN == 0) {
			getBits((u8*)&mb8, 3072, 1);
			ipu_csc(&mb8, &rgb32, 0);
		}
		size = (csc.OFM==1) ? 32 : 64;
		if (csc.OFM){
			ipu_dither(&mb8, &rgb16, csc.DTE);
			pReg.size += fifo_rwrite(((u32*)&rgb16)+4*pReg.size, size - pReg.size);
		}else
			pReg.size += fifo_rwrite(((u32*)&rgb32)+4*pReg.size, size - pReg.size);

		if(pReg.size <= size)
		{
			if(((DMACh*)&psH[0xb000])->chcr & 0x100) dmaIPU0((DMACh*)&psH[0xb000]);
			else{
				csc.MBC -= i;
				ipuRegs.ctrl.BUSY = 1;
				pReg.stopOUT = 1;
				return;
			}
		}
		while(pReg.size < size)
		{
			if (csc.OFM)
				pReg.size += fifo_rwrite(((u32*)&rgb16)+4*pReg.size, size - pReg.size);
			else
				pReg.size += fifo_rwrite(((u32*)&rgb32)+4*pReg.size, size - pReg.size);
			
			if(pReg.size <= size)
			{
				if(((DMACh*)&psH[0xb000])->chcr & 0x100) dmaIPU0((DMACh*)&psH[0xb000]);
				else{
					csc.MBC -= i;
					ipuRegs.ctrl.BUSY = 1;
					pReg.stopOUT = 1;
					return;
				}
			}
		}
		pReg.size = 0;
	}
	pReg.size = 0;
	pReg.stopOUT = 0;
	pReg.stopIN = 0;
	ipuRegs.ctrl.BUSY = 0;
	
	hwIntcIrq(INTC_IPU);
}

// Todo - Need to add the same stop and start code as CSC
void ipuPACK(u32 val) {
	int i;
	tIPU_CMD_CSC  csc ={0, 0, 0, 0, 0};
 
	if(pReg.stopOUT == 0 && pReg.stopIN == 0) *(u32*)&csc=val;
#ifdef IPU_LOG
	IPU_LOG("IPU PACK (Colorspace conversion from RGB32) command.\n");
	if (csc.OFM){	IPU_LOG("Output format is RGB16. ");}
	else{			IPU_LOG("Output format is INDX4. ");}
	if (csc.DTE){	IPU_LOG("Dithering enabled.");	}
	IPU_LOG("Number of macroblocks to be converted: %d\n", csc.MBC);
#endif
	for (i=0; i<(int)csc.MBC; i++){
		fifo_wread(&rgb32, 64);
		//ipu_dither(&mb8, &rgb16, csc.DTE);
		if (csc.OFM)
			fifo_rwrite(&rgb16, 32);
		else{
		//	ipu_vq(&rgb16, indx4);
			fifo_rwrite(indx4,8);
		}
	}

	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void ipuSETTH(u32 val) {
	th0=val & 0xff;
	th1=(val>>16) & 0xff;
#ifdef IPU_LOG
	IPU_LOG("IPU SETTH (Set threshold value)command.\n"
		"Transparent threshold=%x/Translucent threshold=%x\n", th0, th1);
#endif
	ipuRegs.ctrl.BUSY = 0;
	hwIntcIrq(INTC_IPU);
}

void IPUCMD_WRITE(u32 val) {
#ifdef IPU_LOG
	IPU_LOG("=============================================================\n");
#endif

	ipuRegs.cmd.CMD = val;
	ipuRegs.ctrl.ECD = 0;
	ipuRegs.ctrl.SCD = 0; //clear ECD/SCD

	switch (val >> 28){
		case SCE_IPU_BCLR:			//==============[DONE]===============
			ipuBCLR(val);
	        break;

		case SCE_IPU_IDEC:			//==============[DONE]===============
			ipuIDEC(val);
			break;

		case SCE_IPU_BDEC:			//==============[DONE]===============
			ipuBDEC(val);
			break;

		case SCE_IPU_VDEC:			//==============[DONE]===============
			ipuVDEC(val);
			break;
	
		case SCE_IPU_FDEC:			//==============[DONE]===============
			ipuFDEC(val);
			break;
	
		case SCE_IPU_SETIQ:			//==============[DONE]===============
			ipuSETIQ(val);
			break;

		case SCE_IPU_SETVQ:			//==============[DONE]===============
			ipuSETVQ(val);
			break;
	
		case SCE_IPU_CSC:			//==============[DONE]===============
			ipuCSC(val);
			break;
	
		case SCE_IPU_PACK:			//==============[DONE]===============
			ipuPACK(val);
			break;

		case SCE_IPU_SETTH:			//==============[DONE]===============
			ipuSETTH(val);
			break;

		default:
			SysPrintf("Unknown IPU command: %x\n", val >> 28);
			break;
	}
#ifdef IPU_LOG
	IPU_LOG("=============================================================\n");
#endif
}

u32 IPU_read32(u32 mem) {
	u32 ret;

	switch (mem){
		case 0x10002000: // IPU_CMD
#ifdef IPU_LOG
			IPU_LOG("Ipu read32: IPU_CMD=0x%08X\n", ipuRegs.cmd.DATA);
#endif
			return ipuRegs.cmd.DATA;

		case 0x10002010: // IPU_CTRL
			ipuRegs.ctrl.IFC = fifo_cwsize;
			ipuRegs.ctrl.OFC = fifo_crsize;
			ipuRegs.ctrl.CBP = coded_block_pattern;
			ret = ipuRegs.ctrl.IFC;
			ret|= (ipuRegs.ctrl.OFC <<  4);
			ret|= (ipuRegs.ctrl.CBP <<  8);
			ret|= (ipuRegs.ctrl.ECD << 14);
			ret|= (ipuRegs.ctrl.SCD << 15);
			ret|= (ipuRegs.ctrl.IDP << 16);
			ret|= (ipuRegs.ctrl.AS  << 20);
			ret|= (ipuRegs.ctrl.IVF << 21);
			ret|= (ipuRegs.ctrl.QST << 22);
			ret|= (ipuRegs.ctrl.MP1 << 23);
			ret|= (ipuRegs.ctrl.PCT << 24);
			ret|= (ipuRegs.ctrl.BUSY << 31);

#ifdef IPU_LOG
			if(!ipuRegs.cmd.BUSY){
				IPU_LOG("Ipu read32: IPU_CTRL=0x%08X\n", ret);
			}
#endif
			return ret;

		case 0x10002020: // IPU_BP
			ipuRegs.bp.IFC = fifo_cwsize;

			ret = ipuRegs.bp.BP & 0x7f;
			ret|= (ipuRegs.bp.IFC <<  8);
			ret|= ((u32)ipuRegs.bp.FP + ((ipuRegs.bp.BP & 0x80000000) ? 1 : 0)) << 16;

#ifdef IPU_LOG
			IPU_LOG("Ipu read32: IPU_BP=0x%08X\n", ret);
#endif
			return ret;
		default:
#ifdef IPU_LOG
			IPU_LOG("Ipu read: IPU_TOP.BSTOP=%x\n", ipuRegs.top.BSTOP);
#endif
			break;
	}

	return 0;
}

u64 IPU_read64(u32 mem) {
	u64 ret;

	switch (mem){
		case 0x10002000: // IPU_CMD
#ifdef IPU_LOG
			if(!ipuRegs.cmd.BUSY){
				IPU_LOG("Ipu read64: IPU_CMD=BUSY=%x, DATA=%08X\n", ipuRegs.cmd.BUSY,ipuRegs.cmd.DATA);
			}
#endif
			ret = ipuRegs.cmd.DATA;
			ret|= ((u64)ipuRegs.cmd.BUSY << 63);

			return ret;
		case 0x10002030: // IPU_TOP
			if(ipuRegs.ctrl.BUSY) ipuRegs.top.BUSY  = 1;
			else ipuRegs.top.BUSY = 0;

			getBits((u8*)&ipuRegs.top.BSTOP, 32, 0);
#ifdef IPU_LOG
			IPU_LOG("Ipu read64: IPU_TOP=%x(%x),  bp = %d\n",BigEndian(ipuRegs.top.BSTOP),ipuRegs.top.BSTOP,ipuRegs.bp.BP);
#endif
			ipuRegs.top.BSTOP = BigEndian(ipuRegs.top.BSTOP);
			if ((ipuRegs.top.BSTOP & 0xFFFFFF00)==0x00000100)
				ipuRegs.ctrl.SCD=1;
			ret = ipuRegs.top.BSTOP;
			ret|= ((u64)ipuRegs.top.BUSY << 63);
			return ret;
		default:
#ifdef IPU_LOG
			IPU_LOG("Ipu read64: Unknown=%x\n", mem);
#endif
			break;

	}
	return 0;
}

void IPU_write32(u32 mem,u32 value) {
	switch (mem){
		case 0x10002000: // IPU_CMD
#ifdef IPU_LOG
	        IPU_LOG("Ipu write32: IPU_CMD=0x%08X\n",value);
#endif
			IPUCMD_WRITE(value);
			break;
		case 0x10002010: // IPU_CTRL
			ipuRegs.ctrl.IDP = (value >> 16) & 0x3;
			ipuRegs.ctrl.AS  = (value >> 20) & 0x1;
			ipuRegs.ctrl.IVF = (value >> 21) & 0x1;
			ipuRegs.ctrl.QST = (value >> 22) & 0x1;
			ipuRegs.ctrl.MP1 = (value >> 23) & 0x1;
			ipuRegs.ctrl.PCT = (value >> 24) & 0x7;
			if ((value >> 30) & 0x1) { // RESET
				if (!mpeg2_inited){
                    mpeg2_idct_init();
					convert=convert_rgb (CONVERT_RGB, 32);
					convert(16, 16, 0, NULL, &convert_init);
					memset(mb8.Y,0,sizeof(mb8.Y));
					memset(mb8.Cb,0,sizeof(mb8.Cb));
					memset(mb8.Cr,0,sizeof(mb8.Cr));
					memset(mb16.Y,0,sizeof(mb16.Y));
					memset(mb16.Cb,0,sizeof(mb16.Cb));
					memset(mb16.Cr,0,sizeof(mb16.Cr));
					mpeg2_inited=1;
				}
				fifo_wclear();
				fifo_rclear();
				coded_block_pattern = 0;

				ipuRegs.ctrl.BUSY = 0;
				ipuRegs.ctrl.ECD  = 0;
				ipuRegs.ctrl.SCD  = 0;
                ipuRegs.ctrl.IFC  = 0;
				ipuRegs.ctrl.OFC  = 0;
				ipuRegs.cmd.DATA  = 0;
				ipuRegs.cmd.BUSY  = 0;
				ipuRegs.bp.BP     = 0;
				ipuRegs.bp.IFC    = 0;
				ipuRegs.bp.FP     = 0;
				ipuRegs.top.BSTOP = 0;
				ipuRegs.top.BUSY  = 0;
			}
#ifdef IPU_LOG
	        IPU_LOG("Ipu write32: IPU_CTRL=0x%08X\n",value);
#endif
			break;
		default:
#ifdef IPU_LOG
			IPU_LOG("Ipu write32: Unknown=%x\n", mem);
#endif
			break;
	}
}

void IPU_write64(u32 mem, u64 value) {//never called
	switch (mem){
		case 0x10002000:
#ifdef IPU_LOG
	        IPU_LOG("Ipu write64: IPU_CMD=0x%08X\n",value);
#endif
			IPUCMD_WRITE((u32)value);
			break;

		default:
#ifdef IPU_LOG
			IPU_LOG("Ipu write64: Unknown=%x\n", mem);
#endif
			break;
	}
}

//////////////////////////////////////////////////////
///////////////////// CORE FUNCTIONS /////////////////
//////////////////////////////////////////////////////
void ipu_csc(struct macroblock_8 *mb8, struct macroblock_rgb32 *rgb32, int sgn){
	convert_init.start(convert_init.id, (u8*)rgb32, CONVERT_FRAME);
	convert_init.copy(convert_init.id, (u8*)mb8->Y, (u8*)mb8->Cr, (u8*)mb8->Cb, 0);
}

void ipu_dither(struct macroblock_8 *mb8, struct macroblock_rgb16 *rgb16, int dte){
	SysPrintf("IPU: Dither not implemented");
}

void ipu_vq(struct macroblock_rgb16 *rgb16, u8* indx4){
	SysPrintf("IPU: VQ not implemented");
}

void ipu_copy(struct macroblock_8 *mb8, struct macroblock_16 *mb16){
	unsigned char	*s=(unsigned char*)mb8;
	signed short	*d=(signed short*)mb16;
	int i;
	for (i=0; i< 256; i++)	*d++ = *s++;		//Y  bias	- 16
	for (i=0; i< 64; i++)	*d++ = *s++;		//Cr bias	- 128
	for (i=0; i< 64; i++)	*d++ = *s++;		//Cb bias	- 128
}


//////////////////////////////////////////////////////
///////////////////// IPU DMA ////////////////////////
//////////////////////////////////////////////////////
void fifo_rclear()
{
	memset(fifo_output,0,sizeof(fifo_output));
	fifo_crsize = 0;
	readrpos = 0;
	writerpos = 0;
}

int fifo_rread(void *value, int size)
{
	if(fifo_crsize < size)
	{
		int readsize = fifo_crsize;
		memcpy((u32*)value,(u32*)&fifo_output[readrpos],16*readsize);
		readrpos = (readrpos + 4*readsize) & 31;
		fifo_crsize-= readsize;
		return readsize;
	}

	memcpy((u32*)value,(u32*)&fifo_output[readrpos],16*size);
	readrpos = (readrpos + 4*size) & 31;
	fifo_crsize-=size;
	return size;
}

int fifo_rwrite(void* pMem, int size)
{

    if((8-fifo_crsize) < size)
	{
		int rem_qword = 8-fifo_crsize;

		memcpy((u32*)&fifo_output[writerpos],(u32*)pMem,16*rem_qword);
        writerpos = (writerpos + 4*rem_qword) & 31;
		fifo_crsize+=rem_qword;
		return rem_qword;
	}
    
	memcpy((u32*)&fifo_output[writerpos],(u32*)pMem,16*size);
    writerpos = (writerpos + 4*size) & 31;
	fifo_crsize+=size;
	return size;
}


void dmaIPU0() { // fromIPU
	DMACh *dma = (DMACh *)&psH[0xb000];
	u32 *pMem;

#ifdef IPU_LOG
	IPU_LOG("dmaIPU0 chcr = %lx, madr = %lx, qwc  = %lx, ipuout_fifo.count=%x\n",
			dma->chcr, dma->madr, dma->qwc, 8 - fifo_crsize);
#endif

	if ((dma->chcr & 0xC) != 0)
		SysMessage("IPU0 out: not a normal transfer!");


	if (dma->qwc){
		int readsize = 0;
		_dmaGetAddr(pMem, dma->madr, 3);
		readsize = fifo_rread(pMem,dma->qwc);
		dma->madr+= readsize*16;
		dma->qwc -= readsize;
		if(ipuRegs.ctrl.BUSY) // Was busy
		{
			ipuRegs.ctrl.BUSY = 0;
			IPUCMD_WRITE(ipuRegs.cmd.CMD);	// Execute the last command
		}

		if(dma->qwc > 0) 
			return;// If IDEC/BDEC is still running

		if ((psHu32(DMAC_CTRL) & 0x30) == 0x30) { // STS == fromIPU
			switch (psHu32(DMAC_CTRL) & 0xC0) {
				case 0x80: // GIF
					if (dma->madr > psHu32(DMAC_STADR)) {
						psHu32(DMAC_CTRL)&= ~0xC0;
						dmaGIF((DMACh*)&psH[0xa000]);
					}
					break;
			}
		}
	}

	dma->chcr&= ~0x100;
	hwDmacIrq(3);
}

void fifo_wclear()
{
	memset(fifo_input,0,sizeof(fifo_input));
	fifo_cwsize = 0;
	readwpos = 0;
	writewpos = 0;
}

int fifo_wread(void *value, int size)
{
	if(fifo_cwsize < size)
	{
		int rem_qword = fifo_cwsize;
		memcpy((u32*)value,(u32*)&fifo_input[readwpos],16*rem_qword);
		readwpos = (readwpos + 4*rem_qword) & 31;
        fifo_cwsize-=rem_qword;

		return rem_qword;
	}

	memcpy((u32*)value,(u32*)&fifo_input[readwpos],16*size);
	readwpos = (readwpos + 4*size) & 31;
    fifo_cwsize-=size;

	return size;
}

int fifo_wread2(void *value, int size)
{
	if(fifo_cwsize < size)
	{
		int rem_qword = fifo_cwsize;
		memcpy((u32*)value,(u32*)&fifo_input[readwpos],16*rem_qword);
		return rem_qword;
	}

	memcpy((u32*)value,(u32*)&fifo_input[readwpos],16*size);
	return size;
}

int fifo_wwrite(u32* pMem, int size)
{
	if(fifo_cwsize <= 0) memset(fifo_input,0,sizeof(fifo_input));
	if((8-fifo_cwsize) < size)
	{
		int rem_qword = 8-fifo_cwsize;
		memcpy((u32*)&fifo_input[writewpos],(u32*)pMem,16*rem_qword);
		writewpos = (writewpos + 4*rem_qword) & 31;
		fifo_cwsize+=rem_qword;
	    return rem_qword;
	}
    
	memcpy((u32*)&fifo_input[writewpos],(u32*)pMem,16*size);
	writewpos = (writewpos + 4*size) & 31;
	fifo_cwsize+=size;
	return size;
}

#define IPU1chain() \
	if (dma->qwc) { \
		int ret = 0; \
		_dmaGetAddr(pMem, dma->madr, 4);	\
		ret = fifo_wwrite(pMem,dma->qwc); \
		dma->madr+= ret*16; \
		dma->qwc -= ret; \
		cycles+= ret / BIAS;	\
		if(ipuRegs.ctrl.BUSY)	\
		{	\
			ipuRegs.cmd.BUSY = 0;	\
			IPUCMD_WRITE(ipuRegs.cmd.CMD);	\
		}	\
		if(dma->qwc > 0)	return;	\
}

void dmaIPU1() { // toIPU
	u32 *ptag, *pMem;
	DMACh *dma = (DMACh *)&psH[0xb400];
	u32 temp;
	int cycles=0;
	int id, loop=1;

#ifdef IPU_LOG
	IPU_LOG("dmaIPU1 chcr = %lx, madr = %lx, qwc  = %lx, tadr = %lx\n",
			dma->chcr, dma->madr, dma->qwc, dma->tadr);
#endif

	if ((dma->chcr & 0xc) == 0) { // Normal Mode
		IPU1chain();
		dma->chcr&= ~0x100;
        INT(4, cycles);
		return;
	}

	while (1) {	// Chain Mode
		IPU1chain();
		
		if(loop == 0)
			break;
		
		_dmaGetAddr(ptag, dma->tadr, 4);

		if (dma->chcr & 0x40) {
			SysPrintf("Warning! IPU1 TTE: %x_%x, Please report to PCSX2 Team\n", ptag[3], ptag[2]);
		}
		
        id        = (ptag[0] >> 28) & 0x7;
		dma->qwc  = (u16)(ptag[0] & 0xffff);
		dma->madr = ptag[1];

#ifdef IPU_LOG
		IPU_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx, fifosize=%x\n",
				ptag[1], ptag[0], dma->qwc, id, dma->madr, 8 - fifo_cwsize);
#endif

		dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

		switch (id) {
			case 0: // refe
                dma->tadr += 16;
				loop=0;
				break;

			case 1: // cnt
				dma->madr = dma->tadr + 16;
				// Set the taddr to the next tag
				dma->tadr += 16 + (dma->qwc * 16);
				break;

			case 2: // next
				temp = dma->madr;
				dma->madr = dma->tadr + 16;
				dma->tadr = temp;
				break;

			case 3: // ref
			case 4: // refs
				dma->tadr += 16;
				break;

			case 7: // end
				dma->madr = dma->tadr + 16;
				dma->tadr += 16;
				loop=0;
				break;

			default:
#ifdef IPU_LOG
				IPU_LOG("ERROR: different transfer mode!, Please report to PCSX2 Team\n");
#endif
				break;
		}
		
		/*if (dma->chcr & 0x80 && ptag[0] >> 31) {
			_dmaGetAddr(ptag, dma->tadr, 4);
			dma->chcr = ( dma->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );
#ifdef IPU_LOG
			IPU_LOG("dmaIrq Set\n"); 
#endif
			break;
		}*/
	}

	dma->chcr&= ~0x100;
	INT(4, cycles);
}