/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __IPU_H__
#define __IPU_H__

#include "Common.h"

#ifdef __WIN32__
#pragma pack(1)
#endif
#define IPU_CMD_DATA_M		(0xffffffff<< 0)
#define IPU_CMD_BUSY_M		(0x01L     <<63)

#define IPU_CMD_DATA_O		( 0)
#define IPU_CMD_BUSY_O		(63)

//
// Bitfield Structure
//
typedef struct {
	u32 DATA;	// VDEC decoded value
	u32 BUSY;	// VDEC command busy
	u32 CMD;	// last command
} tIPU_CMD;


#define IPU_TOP_BSTOP_M		(0xffffffffL<< 0)
#define IPU_TOP_BUSY_M		(0x1L       <<63)

#define IPU_TOP_BSTOP_O		( 0)
#define IPU_TOP_BUSY_O		(63)

//
// Bitfield Structure
//
typedef struct {
	u32 BSTOP;   // top data of the bit stream (32 bit)
	u32 BUSY;    // busy flag
} tIPU_TOP;

#define IPU_CTRL_IFC_M		(0x0f<< 0)
#define IPU_CTRL_OFC_M		(0x0f<< 4)
#define IPU_CTRL_CBP_M		(0x3f<< 8)
#define IPU_CTRL_ECD_M		(0x01<<14)
#define IPU_CTRL_SCD_M		(0x01<<15)
#define IPU_CTRL_IDP_M		(0x03<<16)
#define IPU_CTRL_AS_M		(0x01<<20)
#define IPU_CTRL_IVF_M		(0x01<<21)
#define IPU_CTRL_QST_M		(0x01<<22)
#define IPU_CTRL_MP1_M		(0x01<<23)
#define IPU_CTRL_PCT_M		(0x07<<24)
#define IPU_CTRL_RST_M		(0x01<<30)
#define IPU_CTRL_BUSY_M		(0x01<<31)

#define IPU_CTRL_IFC_O		( 0)
#define IPU_CTRL_OFC_O		( 4)
#define IPU_CTRL_CBP_O		( 8)
#define IPU_CTRL_ECD_O		(14)
#define IPU_CTRL_SCD_O		(15)
#define IPU_CTRL_IDP_O		(16)
#define IPU_CTRL_AS_O		(20)
#define IPU_CTRL_IVF_O		(21)
#define IPU_CTRL_QST_O		(22)
#define IPU_CTRL_MP1_O		(23)
#define IPU_CTRL_PCT_O		(24)
#define IPU_CTRL_RST_O		(30)
#define IPU_CTRL_BUSY_O		(31)


//
// Bitfield Structure
//
typedef struct {
	u32 IFC;	// Input FIFO counter
	u32 OFC;	// Output FIFO counter
	u32 CBP;	// Coded block pattern
	u32 ECD;	// Error code pattern
	u32 SCD;	// Start code detected
	u32 IDP;	// Intra DC precision
	u32 AS;		// Alternate scan
	u32 IVF;	// Intra VLC format
	u32 QST;	// Q scale step
	u32 MP1;	// MPEG1 bit strea
	u32 PCT;	// Picture Type
	u32 BUSY;	// Busy
} tIPU_CTRL;



//

#define IPU_BP_BP_M		(0x7f<< 0)
#define IPU_BP_IFC_M	(0x0f<< 8)
#define IPU_BP_FP_M		(0x03<<16)

#define IPU_BP_BP_O		( 0)
#define IPU_BP_IFC_O	( 8)
#define IPU_BP_FP_O		(16)


//
// Bitfield Structure
//
typedef struct {
	u32 BP;		// Bit stream point
	u32 IFC;	// Input FIFO counter
	u32 FP;		// FIFO point
} tIPU_BP;
#ifdef __WIN32__
#pragma pack()
#endif

// IPU

#define IPU_CMD_OPTION_M	(0xfffffff<< 0)
#define IPU_CMD_CODE_M		(0x0f     <<28)

#define IPU_CMD_OPTION_O	( 0)
#define IPU_CMD_CODE_O		(28)

//
// Bitfield Structure
//
typedef struct {
	u32 OPTION:28;	// Command optin
	u32 CODE  : 4;	// Command code
} tIPU_CMD_write;

typedef struct {
	u32 FB  : 6;	
	u32 UN2 :10;
	u32 QSC : 5;	
	u32 UN1 : 3;	
	u32 DTD : 1;	
	u32 SGN : 1;	
	u32 DTE : 1;	
	u32 OFM : 1;	
	u32 cmd : 4;	
} tIPU_CMD_IDEC;

typedef struct {
	u32 FB  : 6;	
	u32 UN2 :10;
	u32 QSC : 5;	
	u32 UN1 : 4;	
	u32 DT  : 1;	
	u32 DCR : 1;	
	u32 MBI : 1;	
	u32 cmd : 4;	
} tIPU_CMD_BDEC;

typedef struct {
	u32 MBC :11;
	u32 UN2 :15;
	u32 DTE : 1;	
	u32 OFM : 1;	
	u32 cmd : 4;	
} tIPU_CMD_CSC;

#define SCE_IPU_BCLR	0x0
#define SCE_IPU_IDEC	0x1
#define SCE_IPU_BDEC	0x2
#define SCE_IPU_VDEC	0x3
#define SCE_IPU_FDEC	0x4
#define SCE_IPU_SETIQ	0x5
#define SCE_IPU_SETVQ	0x6
#define SCE_IPU_CSC		0x7
#define SCE_IPU_PACK	0x8
#define SCE_IPU_SETTH	0x9

typedef struct {
  tIPU_CMD  cmd;
  tIPU_TOP  top;
  tIPU_CTRL ctrl;
  tIPU_BP   bp;
} IPUregisters;

IPUregisters ipuRegs;

typedef struct {
	int stopIN;
	int stopOUT;
	int size;
} PauseRegister;

PauseRegister pReg;

void dmaIPU0();
void dmaIPU1();

void IPU_write32(u32 mem,u32 value);
void IPU_write64(u32 mem,u64 value);
u32 IPU_read32(u32 mem);
u64 IPU_read64(u32 mem);

int getBits(u8 *address, u32 size, u32 advance);
int GETMOREFROMFIFO();
int fifo_wread(void *value, int size);

#endif
