/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*

RAM
---
0x00100000-0x01ffffff this is the physical address for the ram.its cached there
0x20100000-0x21ffffff uncached 
0x30100000-0x31ffffff uncached & acceleretade 
0xa0000000-0xa1ffffff MIRROR might...???
0x80000000-0x81ffffff MIRROR might... ????  

scratch pad
----------
0x70000000-0x70003fff scratch pad

BIOS
----
0x1FC00000 - 0x1FFFFFFF un-cached
0x9FC00000 - 0x9FFFFFFF cached
0xBFC00000 - 0xBFFFFFFF un-cached
*/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/stat.h>

#include "Common.h"
#include "PsxMem.h"
#include "R3000A.h"
#include "VU0.h"
#include "VUmicro.h"

//#define FULLTLB

static u16 ba0R16(u32 mem) {
#ifdef MEM_LOG
	MEM_LOG("ba00000 Memory read16 address %x\n", mem);
#endif
	if (mem == 0xba000006) {
		static int ba6;
		ba6++;
		if (ba6 == 3) ba6 = 0;
		return ba6;
	}
	return 0;
}


void memMapMem(u32 base) {
	int i;

	for (i=0; i<0x02000; i++) memLUTRK[i + 0x00000 + base] = (uptr)&psM[i << 12];
	for (i=0; i<0x00400; i++) memLUTRK[i + 0x1fc00 + base] = (uptr)&psR[i << 12];
	for (i=0; i<0x00040; i++) memLUTRK[i + 0x1e000 + base] = (uptr)&psR1[i << 12];
	for (i=0; i<0x00080; i++) memLUTRK[i + 0x1e400 + base] = (uptr)&psR2[i << 12];
	for (i=0; i<0x001C0; i++) memLUTRK[i + 0x1e040 + base] = (uptr)&psER[i << 12];
	for (i=0; i<0x00800; i++) memLUTRK[i + 0x1c000 + base] = (uptr)&psxM[(i & 0x1ff) << 12];
	for (i=0; i<0x00004; i++) memLUTRK[i + 0x11000 + base] = (uptr)VU0.Micro;
	for (i=0; i<0x00004; i++) memLUTRK[i + 0x11004 + base] = (uptr)VU0.Mem;
	for (i=0; i<0x00004; i++) memLUTRK[i + 0x11008 + base] = (uptr)&VU1.Micro[i << 12];
	for (i=0; i<0x00004; i++) memLUTRK[i + 0x1100c + base] = (uptr)&VU1.Mem[i << 12];

	for (i=0; i<0x02000; i++) memLUTWK[i + 0x00000 + base] = (uptr)&psM[i << 12];
	for (i=0; i<0x00400; i++) memLUTWK[i + 0x1fc00 + base] = (uptr)&psR[i << 12];
	for (i=0; i<0x00040; i++) memLUTWK[i + 0x1e000 + base] = (uptr)&psR1[i << 12];
	for (i=0; i<0x00080; i++) memLUTWK[i + 0x1e400 + base] = (uptr)&psR2[i << 12];
	for (i=0; i<0x001C0; i++) memLUTWK[i + 0x1e040 + base] = (uptr)&psER[i << 12];
	for (i=0; i<0x00800; i++) memLUTWK[i + 0x1c000 + base] = (uptr)&psxM[(i & 0x1ff) << 12];
	for (i=0; i<0x00004; i++) memLUTWK[i + 0x11000 + base] = (uptr)VU0.Micro;
	for (i=0; i<0x00004; i++) memLUTWK[i + 0x11004 + base] = (uptr)VU0.Mem;
	for (i=0; i<0x00004; i++) memLUTWK[i + 0x11008 + base] = (uptr)&VU1.Micro[i << 12];
	for (i=0; i<0x00004; i++) memLUTWK[i + 0x1100c + base] = (uptr)&VU1.Mem[i << 12];

	for (i=0; i<0x00010; i++) memLUTRK[i + 0x10000 + base] = 1; // hwm
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x1f800 + base] = 2; // psh
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x1f400 + base] = 3; // psh4
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x18000 + base] = 4; // b80
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x1a000 + base] = 5; // ba0
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x12000 + base] = 6; // gsm
	for (i=0; i<0x00010; i++) memLUTRK[i + 0x14000 + base] = 7; // dev9

	for (i=0; i<0x00010; i++) memLUTWK[i + 0x10000 + base] = 1; // hwm
	for (i=0; i<0x00010; i++) memLUTWK[i + 0x1f800 + base] = 2; // psh
	for (i=0; i<0x00010; i++) memLUTWK[i + 0x1f400 + base] = 3; // psh4
	for (i=0; i<0x00010; i++) memLUTWK[i + 0x1a000 + base] = 5; // ba0
	for (i=0; i<0x00010; i++) memLUTWK[i + 0x12000 + base] = 6; // gsm
	for (i=0; i<0x00010; i++) memLUTWK[i + 0x14000 + base] = 7; // dev9
}

void memMapKernelMem() {
	memMapMem(0xa0000);
	memMapMem(0x80000);
	memMapMem(0x00000);
}

void memMapSupervisorMem() {
}

void memMapUserMem() {
}

void *_align16(void *ptr) {
	if (ptr == NULL) return NULL;
	return (void*)(((uptr)ptr+0x10) & ~0xf);
}

int memInit() {
	int i;

	memLUTRK = (uptr*)malloc(0x100000 * sizeof(uptr));
	memLUTWK = (uptr*)malloc(0x100000 * sizeof(uptr));
	memLUTRU = (uptr*)malloc(0x100000 * sizeof(uptr));
	memLUTWU = (uptr*)malloc(0x100000 * sizeof(uptr));

	_psM  = (char*)malloc(0x02000010); psM  = _align16(_psM);
	_psR  = (char*)malloc(0x00400010); psR  = _align16(_psR);
	_psR1 = (char*)malloc(0x00080010); psR1 = _align16(_psR1);
	_psR2 = (char*)malloc(0x00080010); psR2 = _align16(_psR2);
	_psER = (char*)malloc(0x001C0010); psER = _align16(_psER);
	_psS  = (char*)malloc(0x00004010); psS  = _align16(_psS);
	if (memLUTRK == NULL || memLUTWK == NULL ||
		memLUTRU == NULL || memLUTWU == NULL || 
		psM  == NULL || psR  == NULL || psR1 == NULL || 
		psR2 == NULL || psER == NULL || psS == NULL) {
		SysMessage(_("Error allocating memory")); return -1;
	}

	memset(memLUTRK, 0, 0x100000 * 4);
	memset(memLUTWK, 0, 0x100000 * 4);
	memset(memLUTRU, 0, 0x100000 * 4);
	memset(memLUTWU, 0, 0x100000 * 4);

	if (psxInit() == -1) return -1;

	for (i=0x00000; i<0x00004; i++) memLUTRK[i + 0x70000] = (uptr)&psS[i << 12];
	for (i=0x00000; i<0x00004; i++) memLUTWK[i + 0x70000] = (uptr)&psS[i << 12];

	for (i=0x00000; i<0x00004; i++) memLUTRU[i + 0x70000] = (uptr)&psS[i << 12];
	for (i=0x00000; i<0x00004; i++) memLUTWU[i + 0x70000] = (uptr)&psS[i << 12];

	memMapKernelMem();
	memMapSupervisorMem();
	memMapUserMem();

	memSetKernelMode();

	return 0;
}

void loadBiosRom(char *ext, u8 *dest) {
	struct stat buf;
	char Bios1[256];
	char Bios[256];
	FILE *fp;
	char *ptr;
	int i;

	strcpy(Bios, Config.BiosDir);
	strcat(Bios, Config.Bios);

	sprintf(Bios1, "%s.%s", Bios, ext);
	if (stat(Bios1, &buf) != -1) {	
		fp = fopen(Bios1, "rb");
		fread(dest, 1, buf.st_size, fp);
		fclose(fp);
		return;
	}

	sprintf(Bios1, "%s", Bios);
	ptr = Bios1; i = strlen(Bios1);
	while (i > 0) { if (ptr[i] == '.') break; i--; }
	ptr[i+1] = 0;
	strcat(Bios1, ext);
	if (stat(Bios1, &buf) != -1) {	
		fp = fopen(Bios1, "rb");
		fread(dest, 1, buf.st_size, fp);
		fclose(fp);
		return;
	}

	sprintf(Bios1, "%s%s.bin", Config.BiosDir, ext);
	if (stat(Bios1, &buf) != -1) {	
		fp = fopen(Bios1, "rb");
		fread(dest, 1, buf.st_size, fp);
		fclose(fp);
		return;
	}
	SysPrintf("\n\n\n");
	SysPrintf("**************\n");
	SysPrintf("%s NOT FOUND\n", ext);
	SysPrintf("**************\n\n\n");
}

void memReset() {
	struct stat buf;
	char Bios[256];
	FILE *fp;

	memset(psM, 0, 0x02000000);
	memset(psS, 0, 0x00004000);

	strcpy(Bios, Config.BiosDir);
	strcat(Bios, Config.Bios);

	if (stat(Bios, &buf) == -1) {	
		SysMessage(_("Unable to load bios: '%s', PCSX2 can't run without that"), Bios);
		return;
	}

	fp = fopen(Bios, "rb");
	fread(psR, 1, buf.st_size, fp);
	fclose(fp);

	BiosVersion = GetBiosVersion();
	SysPrintf("Bios Version %d.%d\n", BiosVersion >> 8, BiosVersion & 0xff);

	//injectIRX("host.irx");	//not fully tested; still buggy

	loadBiosRom("rom1", psR1);
	loadBiosRom("rom2", psR2);
	loadBiosRom("erom", psER);
}

void memSetKernelMode() {
	memLUTR = memLUTRK;
	memLUTW = memLUTWK;
}

void memSetSupervisorMode() {
}

void memSetUserMode() {
#ifdef FULLTLB
	memLUTR = memLUTRU;
	memLUTW = memLUTWU;
#endif
}

void memSetPageAddr(u32 vaddr, u32 paddr) {
//	SysPrintf("memSetPageAddr: %8.8x -> %8.8x\n", vaddr, paddr);

	memLUTRU[vaddr >> 12] = memLUTRK[paddr >> 12];
	memLUTWU[vaddr >> 12] = memLUTWK[paddr >> 12];

	memLUTRK[vaddr >> 12] = memLUTRK[paddr >> 12];
	memLUTWK[vaddr >> 12] = memLUTWK[paddr >> 12];
}

void memClearPageAddr(u32 vaddr) {
//	SysPrintf("memClearPageAddr: %8.8x\n", vaddr);

	if ((vaddr & 0xffffc000) == 0x70000000) return;
	memLUTRU[vaddr >> 12] = 0;
	memLUTWU[vaddr >> 12] = 0;

#ifdef FULLTLB
	memLUTRK[vaddr >> 12] = 0;
	memLUTWK[vaddr >> 12] = 0;
#endif
}

void memShutdown() {
	FREE(_psM);
	FREE(_psR);
	FREE(_psR1);
	FREE(_psR2);
	FREE(_psER);
	FREE(_psS);
	FREE(memLUTRK);
	FREE(memLUTWK);
	FREE(memLUTRU);
	FREE(memLUTWU);
}

int  memRead8 (u32 mem, u8  *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
		*out = *(u8 *)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = hwRead8(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = psxHwRead8(mem & ~0xa0000000); return 0;
		case 3: // psh4
			*out = psxHw4Read8(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = gsRead8(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = DEV9read8(mem & ~0xa4000000);
			SysPrintf("DEV9 read8 %8.8lx: %2.2lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead8RS (u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(s8 *)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(s8 *)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (s8)hwRead8(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (s8)psxHwRead8(mem & ~0xa0000000); return 0;
		case 3: // psh4
			*out = (s8)psxHw4Read8(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = (s8)gsRead8(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (s8)DEV9read8(mem & ~0xa4000000);
			SysPrintf("DEV9 read8 %8.8lx: %2.2lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead8RU (u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(u8 *)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(u8 *)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (u8)hwRead8(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (u8)psxHwRead8(mem & ~0xa0000000); return 0;
		case 3: // psh4
			*out = (u8)psxHw4Read8(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = (u8)gsRead8(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (u8)DEV9read8(mem & ~0xa4000000);
			SysPrintf("DEV9 read8 %8.8lx: %2.2lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead16(u32 mem, u16 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
		*out = *(u16*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = hwRead16(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = psxHwRead16(mem & ~0xa0000000); return 0;
		case 4: // b80
#ifdef MEM_LOG
			MEM_LOG("b800000 Memory read16 address %x\n", mem);
#endif
			*out = 0; return 0;
		case 5: // ba0
			*out = ba0R16(mem); return 0;
		case 6: // gsm
			*out = gsRead16(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = DEV9read16(mem & ~0xa4000000);
			SysPrintf("DEV9 read16 %8.8lx: %4.4lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read16  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead16RS(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(s16*)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(s16*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (s16)hwRead16(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (s16)psxHwRead16(mem & ~0xa0000000); return 0;
		case 4: // b80
#ifdef MEM_LOG
			MEM_LOG("b800000 Memory read16 address %x\n", mem);
#endif
			*out = 0; return 0;
		case 5: // ba0
			*out = (s16)ba0R16(mem); return 0;
		case 6: // gsm
			*out = (s16)gsRead16(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (s16)DEV9read16(mem & ~0xa4000000);
			SysPrintf("DEV9 read16 %8.8lx: %4.4lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read16  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead16RU(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(u16*)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(u16*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (u16)hwRead16(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (u16)psxHwRead16(mem & ~0xa0000000); return 0;
		case 4: // b80
#ifdef MEM_LOG
			MEM_LOG("b800000 Memory read16 address %x\n", mem);
#endif
			*out = 0; return 0;
		case 5: // ba0
			*out = (u16)ba0R16(mem); return 0;
		case 6: // gsm
			*out = (u16)gsRead16(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (u16)DEV9read16(mem & ~0xa4000000);
			SysPrintf("DEV9 read16 %8.8lx: %4.4lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read16  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead32(u32 mem, u32 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
		*out = *(u32*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = hwRead32(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = psxHwRead32(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = gsRead32(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = DEV9read32(mem & ~0xa4000000);
			SysPrintf("DEV9 read32 %8.8lx: %8.8lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x (Status=%8.8x)\n", mem, cpuRegs.CP0.n.Status);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead32RS(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(s32*)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(s32*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (s32)hwRead32(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (s32)psxHwRead32(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = (s32)gsRead32(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (s32)DEV9read32(mem & ~0xa4000000);
			SysPrintf("DEV9 read32 %8.8lx: %8.8lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x (Status=%8.8x)\n", mem, cpuRegs.CP0.n.Status);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead32RU(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(u32*)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(u32*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = (u32)hwRead32(mem & ~0xa0000000); return 0;
		case 2: // psh
			*out = (u32)psxHwRead32(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = (u32)gsRead32(mem & ~0xa0000000); return 0;
		case 7: // dev9
			*out = (u32)DEV9read32(mem & ~0xa4000000);
			SysPrintf("DEV9 read32 %8.8lx: %8.8lx\n", mem & ~0xa4000000, *out);
			return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read32  from address %8.8x (Status=%8.8x)\n", mem, cpuRegs.CP0.n.Status);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead64(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u8 *tmp = readCache(mem);
			*out = *(u64*)(tmp+(mem&0xf));
			return 0;
		}
#endif
		*out = *(u64*)(p + (mem & 0xfff));
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			*out = hwRead64(mem & ~0xa0000000); return 0;
		case 6: // gsm
			*out = gsRead64(mem & ~0xa0000000); return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read64  from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

int  memRead128(u32 mem, u64 *out)  {
	char *p;

	p = (char *)(memLUTR[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			u64 *tmp = (u64*)readCache(mem);
			out[0] = tmp[0];
			out[1] = tmp[1];
			return 0;
		}
#endif
		p+= mem & 0xfff;
		out[0] = ((u64*)p)[0];
		out[1] = ((u64*)p)[1];
		return 0;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwRead128(mem & ~0xa0000000, out); return 0;
		case 6: // gsm
			out[0] = gsRead64((mem  ) & ~0xa0000000);
			out[1] = gsRead64((mem+8) & ~0xa0000000); return 0;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory read128 from address %8.8x\n", mem);
#endif
	cpuTlbMissR(mem, cpuRegs.branch);

	return -1;
}

void memWrite8 (u32 mem, u8  value)   {
	char *p;

	p = (char *)(memLUTW[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			writeCache8(mem, value);
			return;
		}
#endif
		*(u8 *)(p + (mem & 0xfff)) = value;
		if (!Config.Cpu) {
			REC_CLEARM(mem&(~3));
//			PSXREC_CLEARM(mem & 0x1ffffc);
		}
		return;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwWrite8(mem & ~0xa0000000, value); return;
		case 2: // psh
			psxHwWrite8(mem & ~0xa0000000, value); return;
		case 3: // psh4
			psxHw4Write8(mem & ~0xa0000000, value); return;
		case 6: // gsm
			gsWrite8(mem & ~0xa0000000, value); return;
		case 7: // dev9
			DEV9write8(mem & ~0xa4000000, value);
			SysPrintf("DEV9 write8 %8.8lx: %2.2lx\n", mem & ~0xa4000000, value);
			return;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory write8   to  address %x with data %2.2x\n", mem, value);
#endif
	cpuTlbMissW(mem, cpuRegs.branch);
}

void memWrite16(u32 mem, u16 value)   {
	char *p;

	p = (char *)(memLUTW[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			writeCache16(mem, value);
			return;
		}
#endif
		*(u16*)(p + (mem & 0xfff)) = value;
		if (!Config.Cpu) {
			REC_CLEARM(mem&(~1));
//			PSXREC_CLEARM(mem & 0x1ffffe);
		}
		return;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwWrite16(mem & ~0xa0000000, value); return;
		case 2: // psh
			psxHwWrite16(mem & ~0xa0000000, value); return;
		case 5: // ba0
#ifdef MEM_LOG
			MEM_LOG("ba00000 Memory write16 to  address %x with data %x\n", mem, value);
#endif
			return;
		case 6: // gsm
			gsWrite16(mem & ~0xa0000000, value); return;
		case 7: // dev9
			DEV9write16(mem & ~0xa4000000, value);
			SysPrintf("DEV9 write16 %8.8lx: %4.4lx\n", mem & ~0xa4000000, value);
			return;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory write16  to  address %x with data %4.4x\n", mem, value);
#endif
	cpuTlbMissW(mem, cpuRegs.branch);
}

void memWrite32(u32 mem, u32 value)   {
	char *p;

//	if (mem >= 0x70000000 && mem < 0x70004000) { Log=1;
//		SysPrintf("psS32 write %x: %x\n", mem, value); }
	p = (char *)(memLUTW[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			writeCache32(mem, value);
			return;
		}
#endif
		*(u32*)(p + (mem & 0xfff)) = value;
		if (!Config.Cpu) {
			REC_CLEARM(mem);
//			PSXREC_CLEARM(mem & 0x1fffff);
		}
		return;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwWrite32(mem & ~0xa0000000, value); return;
		case 2: // psh
			psxHwWrite32(mem & ~0xa0000000, value); return;
		case 6: // gsm
			gsWrite32(mem & ~0xa0000000, value); return;
		case 7: // dev9
			DEV9write32(mem & ~0xa4000000, value);
			SysPrintf("DEV9 write32 %8.8lx: %8.8lx\n", mem & ~0xa4000000, value);
			return;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory write32  to  address %x with data %8.8x\n", mem, value);
#endif
	cpuTlbMissW(mem, cpuRegs.branch);
}

void memWrite64(u32 mem, u64 value)   {
	char *p;

	p = (char *)(memLUTW[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			writeCache64(mem, value);
			return;
		}
#endif
/*		__asm __volatile (
			"movq %1, %%mm0\n"
			"movq %%mm0, %0\n"
			"emms\n"
			: "=m"(*(u64*)(p + (mem & 0xfff))) : "m"(value)
		);*/
		*(u64*)(p + (mem & 0xfff)) = value;
		if (!Config.Cpu) {
			REC_CLEARM(mem);
			REC_CLEARM(mem+4);
/*			PSXREC_CLEARM((mem)   & 0x1fffff);
			PSXREC_CLEARM((mem+4) & 0x1fffff);*/
		}
		return;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwWrite64(mem & ~0xa0000000, value); return;
		case 6: // gsm
			gsWrite64(mem & ~0xa0000000, value); return;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory write64  to  address %x with data %8.8x_%8.8x\n", mem, (u32)(value>>32), (u32)value);
#endif
	cpuTlbMissW(mem, cpuRegs.branch);
}

void memWrite128(u32 mem, u64 *value) {
	char *p;

	p = (char *)(memLUTW[mem >> 12]);
	if ((uptr)p > 0x10) {
#ifdef ENABLECACHE
		if ((mem & 0x20000000) == 0) {
			writeCache128(mem, value);
			return;
		}
#endif
		p+= mem & 0xfff;
		((u64*)p)[0] = value[0];
		((u64*)p)[1] = value[1];
		if (!Config.Cpu) {
			REC_CLEARM(mem);
			REC_CLEARM(mem+4);
			REC_CLEARM(mem+8);
			REC_CLEARM(mem+12);
/*			PSXREC_CLEARM((mem)    & 0x1fffff);
			PSXREC_CLEARM((mem+4)  & 0x1fffff);
			PSXREC_CLEARM((mem+8)  & 0x1fffff);
			PSXREC_CLEARM((mem+12) & 0x1fffff);*/
		}
		return;
	}

	switch ((uptr)p) {
		case 1: // hwm
			hwWrite128(mem & ~0xa0000000, value); return;
		case 6: // gsm
			mem &= ~0xa0000000;
			gsWrite64(mem,   value[0]);
			gsWrite64(mem+8, value[1]); return;
	}

#ifdef MEM_LOG
	MEM_LOG("Unknown Memory write128 to  address %x with data %8.8x_%8.8x_%8.8x_%8.8x\n", mem, ((u32*)value)[3], ((u32*)value)[2], ((u32*)value)[1], ((u32*)value)[0]);
#endif
	cpuTlbMissW(mem, cpuRegs.branch);
}

