/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __MEMORY_H__
#define __MEMORY_H__

#include "Common.h"


//#define ENABLECACHE

//new memory model

u8  *psM; //32mb Main Ram
u8  *_psM;

#define psMs8(mem)	(*(s8 *)&psM[(mem) & 0x1ffffff])
#define psMs16(mem)	(*(s16*)&psM[(mem) & 0x1ffffff])
#define psMs32(mem)	(*(s32*)&psM[(mem) & 0x1ffffff])
#define psMs64(mem)	(*(s64*)&psM[(mem) & 0x1ffffff])
#define psMu8(mem)	(*(u8 *)&psM[(mem) & 0x1ffffff])
#define psMu16(mem)	(*(u16*)&psM[(mem) & 0x1ffffff])
#define psMu32(mem)	(*(u32*)&psM[(mem) & 0x1ffffff])
#define psMu64(mem)	(*(u64*)&psM[(mem) & 0x1ffffff])

u8  *psR; //4mb rom area
u8  *_psR;

#define psRs8(mem)	(*(s8 *)&psR[(mem) & 0x3fffff])
#define psRs16(mem)	(*(s16*)&psR[(mem) & 0x3fffff])
#define psRs32(mem)	(*(s32*)&psR[(mem) & 0x3fffff])
#define psRs64(mem)	(*(s64*)&psR[(mem) & 0x3fffff])
#define psRu8(mem)	(*(u8 *)&psR[(mem) & 0x3fffff])
#define psRu16(mem)	(*(u16*)&psR[(mem) & 0x3fffff])
#define psRu32(mem)	(*(u32*)&psR[(mem) & 0x3fffff])
#define psRu64(mem)	(*(u64*)&psR[(mem) & 0x3fffff])

u8  *psR1; //256kb rom1 area (actually 196kb, but can't mask this)
u8  *_psR1;

#define psR1s8(mem)		(*(s8 *)&psR1[(mem) & 0x3ffff])
#define psR1s16(mem)	(*(s16*)&psR1[(mem) & 0x3ffff])
#define psR1s32(mem)	(*(s32*)&psR1[(mem) & 0x3ffff])
#define psR1s64(mem)	(*(s64*)&psR1[(mem) & 0x3ffff])
#define psR1u8(mem)		(*(u8 *)&psR1[(mem) & 0x3ffff])
#define psR1u16(mem)	(*(u16*)&psR1[(mem) & 0x3ffff])
#define psR1u32(mem)	(*(u32*)&psR1[(mem) & 0x3ffff])
#define psR1u64(mem)	(*(u64*)&psR1[(mem) & 0x3ffff])

u8  *psR2; // 0x00080000
u8  *_psR2;

#define psR2s8(mem)		(*(s8 *)&psR2[(mem) & 0x3ffff])
#define psR2s16(mem)	(*(s16*)&psR2[(mem) & 0x3ffff])
#define psR2s32(mem)	(*(s32*)&psR2[(mem) & 0x3ffff])
#define psR2s64(mem)	(*(s64*)&psR2[(mem) & 0x3ffff])
#define psR2u8(mem)		(*(u8 *)&psR2[(mem) & 0x3ffff])
#define psR2u16(mem)	(*(u16*)&psR2[(mem) & 0x3ffff])
#define psR2u32(mem)	(*(u32*)&psR2[(mem) & 0x3ffff])
#define psR2u64(mem)	(*(u64*)&psR2[(mem) & 0x3ffff])

u8  *psER; // 0x001C0000
u8  *_psER;

#define psERs8(mem)		(*(s8 *)&psER[(mem) & 0x3ffff])
#define psERs16(mem)	(*(s16*)&psER[(mem) & 0x3ffff])
#define psERs32(mem)	(*(s32*)&psER[(mem) & 0x3ffff])
#define psERs64(mem)	(*(s64*)&psER[(mem) & 0x3ffff])
#define psERu8(mem)		(*(u8 *)&psER[(mem) & 0x3ffff])
#define psERu16(mem)	(*(u16*)&psER[(mem) & 0x3ffff])
#define psERu32(mem)	(*(u32*)&psER[(mem) & 0x3ffff])
#define psERu64(mem)	(*(u64*)&psER[(mem) & 0x3ffff])

u8  *psS; //0.015 mb, scratch pad
u8  *_psS;

#define psSs8(mem)	(*(s8 *)&psS[(mem) & 0x3fff])
#define psSs16(mem)	(*(s16*)&psS[(mem) & 0x3fff])
#define psSs32(mem)	(*(s32*)&psS[(mem) & 0x3fff])
#define psSs64(mem)	(*(s64*)&psS[(mem) & 0x3fff])
#define psSu8(mem)	(*(u8 *)&psS[(mem) & 0x3fff])
#define psSu16(mem)	(*(u16*)&psS[(mem) & 0x3fff])
#define psSu32(mem)	(*(u32*)&psS[(mem) & 0x3fff])
#define psSu64(mem)	(*(u64*)&psS[(mem) & 0x3fff])

uptr *memLUTR;
uptr *memLUTW;
uptr *memLUTRK;
uptr *memLUTWK;
uptr *memLUTRU;
uptr *memLUTWU;

#define PSM(mem)	(memLUTR[(mem) >> 12] == 0 ? NULL : (void*)(memLUTR[(mem) >> 12] + ((mem) & 0xfff)))
#define _PSM(mem)	((void*)(memLUTR[(mem) >> 12] + ((mem) & 0xfff)))
#define PSMs8(mem)	(*(s8 *)_PSM(mem))
#define PSMs16(mem)	(*(s16*)_PSM(mem))
#define PSMs32(mem)	(*(s32*)_PSM(mem))
#define PSMs64(mem)	(*(s64*)_PSM(mem))
#define PSMu8(mem)	(*(u8 *)_PSM(mem))
#define PSMu16(mem)	(*(u16*)_PSM(mem))
#define PSMu32(mem)	(*(u32*)_PSM(mem))
#define PSMu64(mem)	(*(u64*)_PSM(mem))

#ifndef DISABLE_REC
extern uptr *recLUT;

#define PC_REC(x)	(recLUT[(x) >> 16] + ((x) & 0xffff))
#define PC_REC32(x) (*(u32*)PC_REC(x))

#define REC_CLEARM(mem) if (recLUT[(mem) >> 16]) PC_REC32(mem) = 0;
#else

#define REC_CLEARM(mem)

#endif

#define FREE(ptr) if (ptr != NULL) { free(ptr); ptr = NULL; }


int  memInit();
void memReset();
void memSetKernelMode();
void memSetSupervisorMode();
void memSetUserMode();
void memSetPageAddr(u32 vaddr, u32 paddr);
void memClearPageAddr(u32 vaddr);
void memShutdown();

int  memRead8(u32 mem, u8  *out);
int  memRead8RS(u32 mem, u64 *out);
int  memRead8RU(u32 mem, u64 *out);
int  memRead16(u32 mem, u16 *out);
int  memRead16RS(u32 mem, u64 *out);
int  memRead16RU(u32 mem, u64 *out);
int  memRead32(u32 mem, u32 *out);
int  memRead32RS(u32 mem, u64 *out);
int  memRead32RU(u32 mem, u64 *out);
int  memRead64(u32 mem, u64 *out);
int  memRead128(u32 mem, u64 *out);
void memWrite8 (u32 mem, u8  value);
void memWrite16(u32 mem, u16 value);
void memWrite32(u32 mem, u32 value);
void memWrite64(u32 mem, u64 value);
void memWrite128(u32 mem, u64 *value);

void *_align16(void *ptr);

#endif
