/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "PsxCommon.h"

static int cnts = 6;

static void psxRcntUpd16(u32 index) {
	if (index == 2 && psxCounters[index].mode & 1) {
		psxCounters[index].Cycle = 0xffffffff;
		return;
	}

	psxCounters[index].sCycle  = psxRegs.cycle;
	psxCounters[index].sCycleT = psxRegs.cycle;

	psxCounters[index].Cycle  = ((0xffff - (psxCounters[index].count&0xffff)) * psxCounters[index].rate) / BIAS;
	psxCounters[index].CycleT = ((psxCounters[index].target - psxRcntRcount16(index)) * psxCounters[index].rate) / BIAS;
}

static void psxRcntUpd32(u32 index) {
	psxCounters[index].sCycle  = psxRegs.cycle;
	psxCounters[index].sCycleT = psxRegs.cycle;

	psxCounters[index].Cycle  = ((0xffff - (psxCounters[index].count&0xffff)) * psxCounters[index].rate) / BIAS;
	psxCounters[index].CycleT = ((psxCounters[index].target - psxRcntRcount32(index)) * psxCounters[index].rate) / BIAS;
/*	if (index == 5) {
		SysPrintf("psxRcntUpd32[5]: sCycle=%x, Cycle=%x, CycleT=%x\n", 
			psxCounters[index].sCycle, psxCounters[index].Cycle, psxCounters[index].CycleT);
	}*/
}

static void psxRcntReset16(u32 index) {
	psxCounters[index].count = 0;

	psxCounters[index].mode&= ~0x08000000;
	psxRcntUpd16(index);
}

static void psxRcntReset32(u32 index) {
	psxCounters[index].count = 0;

	psxCounters[index].mode&= ~0x08000000;
	psxRcntUpd32(index);
}

static void psxRcntSet() {
	u32 c;
	int i;

	psxNextCounter = 0xffffffff;
	psxNextsCounter = psxRegs.cycle;

	for (i=0; i<cnts; i++) {
		c = psxCounters[i].Cycle - (psxRegs.cycle - psxCounters[i].sCycle);
		if (c < psxNextCounter) {
			psxNextCounter = c;
		}
		c = psxCounters[i].CycleT - (psxRegs.cycle - psxCounters[i].sCycleT);
		if (c < psxNextCounter) {
			psxNextCounter = c;
		}
	}
//	SysPrintf("psxRcntSet: %x (cycle)\n", psxNextCounter, psxRegs.cycle);
}

void psxRcntInit() {
	int i;

	memset(psxCounters, 0, sizeof(psxCounters));

	for (i=0; i<6; i++) {
		psxCounters[i].rate = 1;
		psxCounters[i].target = 0xffff;
	}

	psxCounters[0].interrupt = 0x10;
	psxCounters[1].interrupt = 0x20;
	psxCounters[2].interrupt = 0x40;

	psxCounters[3].interrupt = 0x04000;
	psxCounters[4].interrupt = 0x08000;
	psxCounters[5].interrupt = 0x10000;

	if (SPU2async != NULL) {
		cnts = 7;

		psxCounters[6].rate = 768 * 64;
		psxCounters[6].target = 1;
		psxCounters[6].mode = 0x8;
	} else cnts = 6;

	for (i=0; i<3; i++)
		psxRcntUpd16(i);
	for (i=3; i<6; i++)
		psxRcntUpd32(i);
	for (i=6; i<cnts; i++)
		psxRcntUpd16(i);
	psxRcntSet();
}

void psxVSyncStart() {
	psxHu32(0x1070)|= 1;
}

void psxVSyncEnd() {
	cdvdVsync();
	psxHu32(0x1070)|= 0x800;
}


void psxRcntUpdate() {
	int i;

	for (i=0; i<3; i++) {
		if ((psxRegs.cycle - psxCounters[i].sCycleT) >= psxCounters[i].CycleT &&
			(psxCounters[i].mode & 0x08000000) == 0) {
//			if (i == 5 && Log) SysPrintf("cycle %x >= %x (CycleT)\n", (psxRegs.cycle - psxCounters[i].sCycleT), psxCounters[i].CycleT);
			if (!(psxCounters[i].mode & 0x08)) {
				psxCounters[i].mode|= 0x08000000;
			} else { // Reset on target
				psxRcntReset16(i);
			}
			psxCounters[i].mode|= 0x0800; // Target flag
			if (psxCounters[i].mode & 0x10) { // Target interrupt
				if (psxCounters[i].mode & 0x40) { // Repeat interrupt
					psxHu32(0x1070)|= psxCounters[i].interrupt;
					psxCounters[i].mode|= 0x0400; // Interrupt flag
				} else {
					if ((psxCounters[i].mode & 0x00100000) == 0) {
						psxCounters[i].mode|= 0x00100000; // Internal target interrupt flag
						psxHu32(0x1070)|= psxCounters[i].interrupt;
						psxCounters[i].mode|= 0x0400; // Interrupt flag
					}
				}
			}
		}
		if ((psxRegs.cycle - psxCounters[i].sCycle) >= psxCounters[i].Cycle) {
//			if (i == 5 && Log) SysPrintf("cycle %x >= %x (Cycle)\n", (psxRegs.cycle - psxCounters[i].sCycle), psxCounters[i].Cycle);
			psxCounters[i].mode|= 0x1000; // Overflow flag
			if (psxCounters[i].mode & 0x0020) { // Overflow interrupt
				if (psxCounters[i].mode & 0x0040) { // Repeat interrupt
					psxHu32(0x1070)|= psxCounters[i].interrupt;
					psxCounters[i].mode|= 0x0400; // Interrupt flag
				} else {
					if ((psxCounters[i].mode & 0x200000) == 0) {
						psxCounters[i].mode|= 0x200000; // Internal Overflow interrupt flag
						psxHu32(0x1070)|= psxCounters[i].interrupt;
						psxCounters[i].mode|= 0x0400; // Interrupt flag
					}
				}
			}

			psxRcntReset16(i);
/*			if (i == 5 && Log) {
				SysPrintf("psxRcntCycle[5]: sCycle=%x, Cycle=%x (count=%x)\n", 
					psxCounters[i].sCycle, psxCounters[i].Cycle, psxCounters[i].count);
			}*/
		}
	}
	for (i=3; i<6; i++) {
		if ((psxRegs.cycle - psxCounters[i].sCycleT) >= psxCounters[i].CycleT &&
			(psxCounters[i].mode & 0x08000000) == 0) {
//			if (i == 5) SysPrintf("cycle %x >= %x (CycleT); count=%x\n", (psxRegs.cycle - psxCounters[i].sCycleT), psxCounters[i].CycleT, psxRcntRcount32(i));
			if (!(psxCounters[i].mode & 0x08)) {
				psxCounters[i].mode|= 0x08000000;
			} else { // Reset on target
				psxRcntReset32(i);
			}
			psxCounters[i].mode|= 0x0800; // Target flag
			if (psxCounters[i].mode & 0x10) { // Target interrupt
//				if (i == 5) SysPrintf("target interrupt %x (count=%x)\n", psxRegs.cycle, psxRcntRcount32(i));
				if (psxCounters[i].mode & 0x40) { // Repeat interrupt
					psxHu32(0x1070)|= psxCounters[i].interrupt;
					psxCounters[i].mode|= 0x0400; // Interrupt flag
				} else {
					if ((psxCounters[i].mode & 0x00100000) == 0) {
						psxCounters[i].mode|= 0x00100000; // Internal target interrupt flag
						psxHu32(0x1070)|= psxCounters[i].interrupt;
						psxCounters[i].mode|= 0x0400; // Interrupt flag
					}
				}
			}
		}
		if ((psxRegs.cycle - psxCounters[i].sCycle) >= psxCounters[i].Cycle) {
//			if (i == 5) SysPrintf("cycle %x >= %x (Cycle); count=%x\n", (psxRegs.cycle - psxCounters[i].sCycle), psxCounters[i].Cycle, psxRcntRcount32(i));
			if ((psxCounters[i].count >> 16) == 0xffff) {
//				if (i == 5) SysPrintf("overflow interrupt %x\n", psxRcntRcount32(i));
				psxCounters[i].mode|= 0x1000; // Overflow flag
				if (psxCounters[i].mode & 0x0020) { // Overflow interrupt
					if (psxCounters[i].mode & 0x0040) { // Repeat interrupt
						psxHu32(0x1070)|= psxCounters[i].interrupt;
						psxCounters[i].mode|= 0x0400; // Interrupt flag
					} else {
						if ((psxCounters[i].mode & 0x200000) == 0) {
							psxCounters[i].mode|= 0x200000; // Internal Overflow interrupt flag
							psxHu32(0x1070)|= psxCounters[i].interrupt;
							psxCounters[i].mode|= 0x0400; // Interrupt flag
						}
					}
				}
				psxRcntReset32(i);
			} else {
				psxCounters[i].sCycle = psxRegs.cycle;
				psxCounters[i].Cycle  = ((0xffff - (psxCounters[i].count&0xffff)) * psxCounters[i].rate) / BIAS;
				psxCounters[i].count+= 0x10000;
/*				if (i == 4 && Log) {
					SysPrintf("psxRcntCycle[%d]: sCycle=%x, Cycle=%x (count=%x)\n", i,
						psxCounters[i].sCycle, psxCounters[i].Cycle, psxCounters[i].count);
				}*/
			}
		}
	}

	if (cnts >= 7) {
		if ((psxRegs.cycle - psxCounters[6].sCycleT) >= psxCounters[6].CycleT) {
			SPU2async((psxRegs.cycle - psxCounters[6].sCycle) * BIAS);
			psxRcntReset16(6);
		}
	}

	psxRcntSet();
}

void psxRcntWcount16(u32 index, u32 value) {
//	SysPrintf("writeCcount[%d] = %x\n", index, value);
//	PSXCPU_LOG("writeCcount[%d] = %x\n", index, value);
	psxCounters[index].count = value;
	psxRcntUpd16(index);
	psxRcntSet();
}

void psxRcntWcount32(u32 index, u32 value) {
//	SysPrintf("writeCcount32[%d] = %x\n", index, value);
//	PSXCPU_LOG("writeCcount[%d] = %x\n", index, value);
	psxCounters[index].count = value;
	psxRcntUpd32(index);
	psxRcntSet();
}

void psxRcnt0Wmode(u32 value)  {
//	SysPrintf("writeCmode[0] = %lx\n", value);
	psxCounters[0].mode = value;
	if (value & 0x100) { // Pixel Clock (13.5Mhz)
		psxCounters[0].rate = 3;
	} else {
		psxCounters[0].rate = 1;
	}
	// Need to set a rate and target
	psxRcntUpd16(0);
	psxRcntSet();
}

void psxRcnt1Wmode(u32 value)  {
//	SysPrintf("writeCmode[1] = %lx\n", value);
	psxCounters[1].mode = value;
	if (value & 0x100) {
		psxCounters[1].rate = PSXHBLANK;
	} else {
		psxCounters[1].rate = 1;
	}
	// Need to set a rate and target
	psxRcntUpd16(1);
	psxRcntSet();
}

void psxRcnt2Wmode(u32 value)  {
//	SysPrintf("writeCmode[2] = %lx\n", value);
	psxCounters[2].mode = value;
	if (value & 0x200) {
		psxCounters[2].rate = 8; // 1/8 speed
	} else {
		psxCounters[2].rate = 1; // normal speed
	}
	// Need to set a rate and target
	psxRcntUpd16(2);
	psxRcntSet();
}

void psxRcnt3Wmode(u32 value)  {
//	SysPrintf("writeCmode[3] = %lx\n", value);
	psxCounters[3].mode = value;
	psxCounters[3].rate = 1; // normal speed
	// Need to set a rate and target
	psxRcntReset32(3);
	psxRcntSet();
}

void psxRcnt4Wmode(u32 value)  {
//	SysPrintf("writeCmode[4] = %lx\n", value);
	psxCounters[4].mode = value;
	switch (value & 0x6000) {
		case 0x0000: psxCounters[4].rate = 1;   break; // normal speed
		case 0x2000: psxCounters[4].rate = 8;   break; // 1/8    speed
		case 0x4000: psxCounters[4].rate = 16;  break; // 1/16   speed
		case 0x6000: psxCounters[4].rate = 256; break; // 1/256  speed
	}
	// Need to set a rate and target
	psxRcntReset32(4);
	psxRcntSet();
}

void psxRcnt5Wmode(u32 value)  {
//	SysPrintf("writeCmode[5] = %lx\n", value);
	psxCounters[5].mode = value;
	switch (value & 0x6000) {
		case 0x0000: psxCounters[5].rate = 1;   break; // normal speed
		case 0x2000: psxCounters[5].rate = 8;   break; // 1/8    speed
		case 0x4000: psxCounters[5].rate = 16;  break; // 1/16   speed
		case 0x6000: psxCounters[5].rate = 256; break; // 1/256  speed
	}
	// Need to set a rate and target
	psxRcntReset32(5);
	psxRcntSet();
}

void psxRcntWtarget16(u32 index, u32 value) {
//	SysPrintf("writeCtarget[%ld] = %lx\n", index, value);
	psxCounters[index].target = value;
	psxRcntUpd16(index);
	psxRcntSet();
}

void psxRcntWtarget32(u32 index, u32 value) {
//	SysPrintf("writeCtarget32[%ld] = %lx (count=%lx) ; ", index, value, psxRcntRcount32(index));
	psxCounters[index].target = value;
	psxCounters[index].mode&= ~0x08000000;
	psxCounters[index].sCycleT = psxRegs.cycle;
	psxCounters[index].CycleT = ((psxCounters[index].target - psxRcntRcount32(index)) * psxCounters[index].rate) / BIAS;
//	SysPrintf("sCycleT: %x CycleT: %x\n", psxCounters[index].sCycleT, psxCounters[index].CycleT);
//	psxRcntUpd32(index);
//	psxRcntReset32(index);
	psxRcntSet();
}

u16 psxRcntRcount16(u32 index) {
	return (u16)(psxCounters[index].count + (BIAS * (psxRegs.cycle - psxCounters[index].sCycle)) / psxCounters[index].rate);
}

u32 psxRcntRcount32(u32 index) {
	u32 ret = psxCounters[index].count + (BIAS * (psxRegs.cycle - psxCounters[index].sCycle)) / psxCounters[index].rate;
//	if (index == 5) SysPrintf("psxRcntRcount32: %x (cycle=%x)\n", ret, psxRegs.cycle);
	return ret;
}

int psxRcntFreeze(gzFile f, int Mode) {
	gzfreezel(psxCounters);

	return 0;
}
