/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2004  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "PsxCommon.h"

// Dma0/1   in Mdec.c
// Dma3     in CdRom.c
// Dma8     in PsxSpd.c
// Dma11/12 in PsxSio2.c

void psxDma4(u32 madr, u32 bcr, u32 chcr) { // SPU
	int size;

	switch (chcr) {
		case 0x01000201: //cpu to spu transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 4 - SPU mem2spu *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
			SPU2writeDMA4Mem((u16 *)PSXM(madr), size);
			PSX_INT(4, (size * 80) / BIAS);
			break;
		case 0x01000200: //spu to cpu transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 4 - SPU spu2mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
    		SPU2readDMA4Mem((u16 *)PSXM(madr), size);
			PSX_INT(4, (size * 80) / BIAS);
			break;
#ifdef PSXDMA_LOG
		default:
			PSXDMA_LOG("*** DMA 4 - SPU unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
}

int  psxDma4Interrupt() {
	HW_DMA4_CHCR &= ~0x01000000;
	psxDmaInterrupt(4);
	SPU2interruptDMA4();
	return 1;
}

void psxDma2(u32 madr, u32 bcr, u32 chcr) { // GPU
	switch(chcr) {
		case 0x01000200: // vram2mem
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 2 - GPU vram2mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_readDataMem((u32 *)PSXM(madr), (bcr >> 16) * (bcr & 0xffff));
			psxCpu->Clear(madr, ((bcr >> 16) * (bcr & 0xffff)) / 4);
			break;

		case 0x01000201: // mem2vram
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 2 - GPU mem2vram *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_writeDataMem((u32 *)PSXM(madr), (bcr >> 16) * (bcr & 0xffff));
			break;

		case 0x01000401: // dma chain
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 2 - GPU dma chain *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			GPU_dmaChain((u32 *)psxM, madr & 0x1fffff);
			break;
#ifdef PSXDMA_LOG
		default:
			PSXDMA_LOG("*** DMA 2 - GPU unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
	HW_DMA2_CHCR &= ~0x01000000;
	psxDmaInterrupt(2);
}

void psxDma6(u32 madr, u32 bcr, u32 chcr) {
	u32 *mem = (u32 *)PSXM(madr);

#ifdef PSXDMA_LOG
	PSXDMA_LOG("*** DMA 6 - OT *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif

	if (chcr == 0x11000002) {
		while (bcr--) {
			*mem-- = (madr - 4) & 0xffffff;
			madr -= 4;
		}
		mem++; *mem = 0xffffff;
	} else {
		// Unknown option
#ifdef PSXDMA_LOG
		PSXDMA_LOG("*** DMA 6 - OT unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
	}
	HW_DMA6_CHCR &= ~0x01000000;
	psxDmaInterrupt(6);
}

void psxDma7(u32 madr, u32 bcr, u32 chcr) {
	int size;

	switch (chcr) {
		case 0x01000201: //cpu to spu2 transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 7 - SPU2 mem2spu *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
			SPU2writeDMA7Mem((u16 *)PSXM(madr), size);
			PSX_INT(7, (size * 80) / BIAS);
			break;
		case 0x01000200: //spu2 to cpu transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 7 - SPU2 spu2mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
    		SPU2readDMA7Mem((u16 *)PSXM(madr), size);
			PSX_INT(7, (size * 80) / BIAS);
			break;
#ifdef PSXDMA_LOG
		default:
			PSXDMA_LOG("*** DMA 7 - SPU2 unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
}

int  psxDma7Interrupt() {
	HW_DMA7_CHCR &= ~0x01000000;
	psxDmaInterrupt2(0);
	SPU2interruptDMA7();
	return 1;
}

void psxDma9(u32 madr, u32 bcr, u32 chcr) {

	DMACh *dma = (DMACh*)&psH[0xc000];

#ifdef SIF_LOG
	SIF_LOG("IOP: dmaSIF0 chcr = %lx, madr = %lx, bcr = %lx, tadr = %lx\n",	chcr, madr, bcr, HW_DMA9_TADR);
#endif

	if (dma->chcr & 0x100 && HW_DMA9_CHCR & 0x01000000) {
		SIF0Dma();
	}
}

void psxDma10(u32 madr, u32 bcr, u32 chcr) {
	DMACh *dma = (DMACh*)&psH[0xc400];

#ifdef SIF_LOG
	SIF_LOG("IOP: dmaSIF1 chcr = %lx, madr = %lx, bcr = %lx\n",	chcr, madr, bcr);
#endif

	if (dma->chcr & 0x100 && HW_DMA10_CHCR & 0x01000000) {
		SIF1Dma();
	}
}

void psxDma8(u32 madr, u32 bcr, u32 chcr) {
	int size;

	switch (chcr & 0x01000201) {
		case 0x01000201: //cpu to dev9 transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 8 - DEV9 mem2dev9 *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
			DEV9writeDMA8Mem(PSXM(madr), size);
			break;
		case 0x01000200: //dev9 to cpu transfer
#ifdef PSXDMA_LOG
			PSXDMA_LOG("*** DMA 8 - DEV9 dev9mem *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
#endif
			size = (bcr >> 16) * (bcr & 0xffff) * 4;
			DEV9readDMA8Mem(PSXM(madr), size);
			break;
#ifdef PSXDMA_LOG
		default:
			PSXDMA_LOG("*** DMA 8 - DEV9 unknown *** %lx addr = %lx size = %lx\n", chcr, madr, bcr);
			break;
#endif
	}
	HW_DMA8_CHCR &= ~0x01000000;
	psxDmaInterrupt2(1);
}

int  dev9Interrupt() {
	if (dev9Handler == NULL) goto irq;
	if (dev9Handler() != 1) return 1;
irq:
	psxHu32(0x1070)|= 1<<13;
	return 1;
}

void dev9Irq(int cycles) {
	PSX_INT(20, cycles);
}

int  usbInterrupt() {
	if (usbHandler == NULL) goto irq;
	if (usbHandler() != 1) return 1;
irq:
	psxHu32(0x1070)|= 1<<22;
	return 1;
}

void usbIrq(int cycles) {
	PSX_INT(21, cycles);
}

void fwIrq() {
	psxHu32(0x1070)|= 1<<24;
}

void spu2Irq() {
	psxHu32(0x1070)|= 1<<9;
}

