/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "Memory.h"
#include "Hw.h"
#include "Debug.h"
#include "VU0.h"
#include "R3000A.h"
#include "VUmicro.h"


int cpuInit() {
	int ret;

	cpudetectInit();
#ifndef DISABLE_REC
	if (Config.Cpu) Cpu = &intCpu;
	else Cpu = &recCpu;
#else
	Cpu = &intCpu;
#endif
	ret = Cpu->Init();
	if (ret == -1 && Config.Cpu == 0) {
		SysMessage(_("Error initializing Recompiler, switching to Interpreter"));
		Config.Cpu = 1;
		Cpu = &intCpu;
		ret = Cpu->Init();
	}

	if (hwInit() == -1) return -1;
	if (vu0Init() == -1) return -1;
	if (vu1Init() == -1) return -1;
	if (memInit() == -1) return -1;
	Log = 0;

	return ret;
}

void cpuReset() {
	Cpu->Reset();

	memReset();

	memset(&cpuRegs, 0, sizeof(cpuRegs));
	memset(&fpuRegs, 0, sizeof(fpuRegs));
	memset(&tlb, 0, sizeof(tlb));

	cpuRegs.pc = 0xbfc00000; ///set pc reg to stack

	cpuRegs.CP0.n.Status.val = 0x10900000; // COP0 enabled | BEV = 1 | TS = 1
	cpuRegs.CP0.n.PRid   = 0x00002e20; // PRevID = Revision ID, same as R5900
	fpuRegs.fprc[0]   = 0x00002e00; // fpu Revision..
	fpuRegs.fprc[31]  = 0x01000001; // fpu Status/Control

    vu0Reset();
    vu1Reset();
    vif0Reset();
    vif1Reset();
	hwReset();
	rcntInit();
	psxReset();
}

void cpuShutdown() {
	memShutdown();
	hwShutdown();
//	biosShutdown();
	psxShutdown();
	vu0Shutdown();
	vu1Shutdown();
	disR5900FreeSyms();

	Cpu->Shutdown();
}

void cpuException(u32 code, u32 bd) {
	u32 offset;
//	BIOS_LOG("cpuException %x\n", cpuRegs.pc);
//	if (cpuRegs.CP0.n.Status.b.EXL == 1) {
/*	if (cpuRegs.cycle > 0x0b2f4008) {*/
/*	if ((code & 0x800400) == 0x800400) {
		Log=1;
		SysPrintf("cpuException %x, %x, status=%x, code=%x\n", cpuRegs.pc, cpuRegs.cycle, cpuRegs.CP0.n.Status.val, code);
	}*/
	cpuRegs.CP0.n.Cause = code;
	if (bd) {
		SysPrintf("branch delay!!\n");
	}

	if (cpuRegs.CP0.n.Status.b.EXL == 0) {
		cpuRegs.CP0.n.Status.b.EXL = 1;
		if (code & 0x800000) {
			cpuRegs.CP0.n.EPC = cpuRegs.pc;
			offset = 0x200;
		} else {
			cpuRegs.CP0.n.EPC = cpuRegs.pc - 4;
			offset = 0x180;
		}
	} else {
		SysPrintf("cpuException: Status.EXL = 1\n");
		return;
		offset = 0x180;
		SysPrintf("cpuException: Status.EXL = 1\n");
	}

	if (cpuRegs.CP0.n.Status.b.BEV == 0) {
		cpuRegs.pc = 0x80000000 + offset;
	} else {
		cpuRegs.pc = 0xBFC00200 + offset;
	}

	UpdateCP0Status();
}

void cpuTlbMiss(u32 addr, u32 bd, u32 excode) {
	SysPrintf("cpuTlbMiss %x, %x, status=%x, code=%x\n", cpuRegs.pc, cpuRegs.cycle, cpuRegs.CP0.n.Status.val, excode);
	if (bd) {
		SysPrintf("branch delay!!\n");
	}
	cpuRegs.CP0.n.BadVAddr = addr;
	cpuRegs.CP0.n.Context &= 0xFF80000F;
	cpuRegs.CP0.n.Context |= (addr >> 9) & 0x007FFFF0;
	cpuRegs.CP0.n.EntryHi = (addr & 0xFFFFE000) | (cpuRegs.CP0.n.EntryHi & 0x1FFF);

	cpuRegs.CP0.n.Cause = excode;
	if (!(cpuRegs.CP0.n.Status.val & 0x2)) { // EXL bit
		cpuRegs.CP0.n.EPC = cpuRegs.pc - 4;
	}

	if ((cpuRegs.CP0.n.Status.val & 0x1) == 0) {
		cpuRegs.pc = 0x80000000;
	} else {
		cpuRegs.pc = 0x80000180;
	}

	cpuRegs.CP0.n.Status.b.EXL = 1;
	UpdateCP0Status();
//	Log=1; varLog|= 0x40000000;
}

void cpuTlbMissR(u32 addr, u32 bd) {
	cpuTlbMiss(addr, bd, EXC_CODE_TLBL);
}

void cpuTlbMissW(u32 addr, u32 bd) {
	cpuTlbMiss(addr, bd, EXC_CODE_TLBS);
}



void JumpCheckSym(u32 addr, u32 pc) {
#if 0
//	if (addr == 0x80051770) { SysPrintf("Log!: %s\n", PSM(cpuRegs.GPR.n.a0.UL[0])); Log=1; varLog|= 0x40000000; }
	if (addr == 0x8002f150) { SysPrintf("printk: %s\n", PSM(cpuRegs.GPR.n.a0.UL[0])); }
	if (addr == 0x8002aba0) return;
	if (addr == 0x8002f450) return;
	if (addr == 0x800dd520) return;
//	if (addr == 0x80049300) SysPrintf("register_blkdev: %x\n", cpuRegs.GPR.n.a0.UL[0]);
	if (addr == 0x8013cb70) { SysPrintf("change_root: %x\n", cpuRegs.GPR.n.a0.UL[0]); }
//	if (addr == 0x8013d1e8) { SysPrintf("Log!\n"); Log++; if (Log==2) exit(0); varLog|= 0x40000000; }
//	if (addr == 0x00234e88) { SysPrintf("StoreImage\n"); Log=1; /*psMu32(0x234e88) = 0x03e00008; psMu32(0x234e8c) = 0;*/ }
#endif
/*	if ((pc >= 0x00131D50 &&
		 pc <  0x00132454) ||
		(pc >= 0x00786a90 &&
		 pc <  0x00786ac8))*/
	if (varLog & 0x40000000) {
		char *str;
		char *strf;

		str = disR5900GetSym(addr);
		if (str != NULL) {
			strf = disR5900GetUpperSym(pc);
			if (strf) {
				SysPrintf("Func %8.8x: %s (called by %8.8x: %s)\n", addr, str, pc, strf);
			} else {
				SysPrintf("Func %8.8x: %s (called by %x)\n", addr, str, pc);
			}
			if (!strcmp(str, "printf")) { SysPrintf("%s\n", (char*)PSM(cpuRegs.GPR.n.a0.UL[0])); }
			if (!strcmp(str, "printk")) { SysPrintf("%s\n", (char*)PSM(cpuRegs.GPR.n.a0.UL[0])); }
		}
	}
}

void JumpCheckSymRet(u32 addr) {
	if (varLog & 0x40000000) {
		char *str;
		str = disR5900GetUpperSym(addr);
		if (str != NULL) {
			SysPrintf("Return       : %s, v0=%8.8x\n", str, cpuRegs.GPR.n.v0.UL[0]);
//			if (!strcmp(str, "printf")) { SysPrintf("%s\n", (char*)PSM(cpuRegs.GPR.n.a0.UL[0])); }
		}
	}
}

__inline void _cpuTestMissingINTC() {
	if (cpuRegs.CP0.n.Status.val & 0x400 &&
		psHu32(INTC_STAT) & psHu32(INTC_MASK)) {
		if ((cpuRegs.interrupt & (1 << 30)) == 0) {
			SysPrintf("*PCSX2*: Error, missing INTC Interrupt\n");
		}
	}
}

__inline void _cpuTestMissingDMAC() {
	if (cpuRegs.CP0.n.Status.val & 0x800 &&
		(psHu16(0xe012) & psHu16(0xe010) || 
		 psHu16(0xe010) & 0x8000)) {
		if ((cpuRegs.interrupt & (1 << 31)) == 0) {
			SysPrintf("*PCSX2*: Error, missing DMAC Interrupt\n");
		}
	}
}

void cpuTestMissingHwInts() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) == 0x10001) {
		_cpuTestMissingINTC();
		_cpuTestMissingDMAC();
//		_cpuTestTIMR();
	}
}

#define TESTINT(n, callback) \
	if (cpuRegs.interrupt & (1 << n)) { \
		if ((cpuRegs.cycle - cpuRegs.sCycle[n]) >= cpuRegs.eCycle[n]) { \
			if (callback() == 1) { \
				cpuRegs.interrupt&= ~(1 << n); \
			} \
		} \
	}

void _cpuTestInterrupts() {
	TESTINT(0, vif0Interrupt);
	TESTINT(1, vif1Interrupt);
	TESTINT(2, gsInterrupt);
//	TESTINT(3, ipu0Interrupt);
//	TESTINT(4, ipu1Interrupt);
/*	TESTINT(5, sif0Interrupt);
	TESTINT(6, sif1Interrupt);
	TESTINT(7, EEsif0Interrupt);
	TESTINT(8, EEsif1Interrupt);*/
	TESTINT(30, intcInterrupt);
	TESTINT(31, dmacInterrupt);
}

void cpuBranchTest()  {
	if ((cpuRegs.cycle - nextsCounter) >= nextCounter)
		rcntUpdate();

	if (cpuRegs.interrupt) {
		_cpuTestInterrupts();
	}

//	cpuTestHwInts();
#ifdef CPU_LOG
	cpuTestMissingHwInts();
#endif

	cpuRegs.EEsCycle+= cpuRegs.cycle - cpuRegs.EEoCycle;
	cpuRegs.EEoCycle = cpuRegs.cycle;
/*	if (cpuRegs.EEsCycle < -100) {
		cpuRegs.IOPoCycle = psxRegs.cycle;
		cpuRegs.EEsCycle = -100;
	}*/

	while (cpuRegs.EEsCycle > 0) {
#ifdef PSXCPU_LOG
		if (Log) { PSXCPU_LOG("*iopExecute* %8.8x (eeCycle=%8.8x : iopCycle=%8.8x)\n", psxRegs.pc, cpuRegs.cycle, psxRegs.cycle); } 
#endif
		psxCpu->ExecuteBlock();

		cpuRegs.EEsCycle-= (psxRegs.cycle - cpuRegs.IOPoCycle) * (PS2CLK / PSXCLK);
		cpuRegs.IOPoCycle = psxRegs.cycle;
	}

	if (VU0.VI[REG_VPU_STAT].UL & 0x1) {
		Cpu->ExecuteVU0Block();
	}
	if (VU0.VI[REG_VPU_STAT].UL & 0x100) {
		Cpu->ExecuteVU1Block();
	}

/*	if (cpuRegs.cycle > 0x052d347e &&
		cpuRegs.cycle < 0x052f46fa) Log=1; else Log=0;*/
//	if (cpuRegs.cycle > 0x010c2d00) if (!Log) { SysPrintf("Log!\n"); Log=1; varLog|=0x40000000;}
//	if (cpuRegs.cycle > 0x0f201504) if (!Log) { SysPrintf("Log!\n"); Log=1; varLog|=0x40000000; }
//	if (cpuRegs.cycle > 0x02b38700) if (!Log) { SysPrintf("Log!\n"); Log=1; varLog|=0x40008404; }
//	if (cpuRegs.cycle > 0x1cd53700) { exit(1); }
//	if (cpuRegs.cycle > 0x261a537c) { SaveState("DOA2.auto"); exit(1); }
//	if (Log) if (Log++ == 10) Log=0;
//	JumpCheckSym(cpuRegs.pc);
/*	{
		static u32 mem;
		if (mem != psMu32(0x1fffb60)) {
			mem = psMu32(0x1fffb60);
			SysPrintf("0x1fffb60 = %x\n", mem);
		}
	}*/
}

__inline void _cpuTestINTC() {
	if (cpuRegs.CP0.n.Status.val & 0x400 &&
		psHu32(INTC_STAT) & psHu32(INTC_MASK)) {
		if ((cpuRegs.interrupt & (1 << 30)) == 0) {
//			SysPrintf("_cpuTestINTC %x\n", psHu32(INTC_STAT) & psHu32(INTC_MASK));
			INT(30, 4);
		}
	}
}

__inline void _cpuTestDMAC() {
	if (cpuRegs.CP0.n.Status.val & 0x800 &&
		(psHu16(0xe012) & psHu16(0xe010) || 
		 psHu16(0xe010) & 0x8000)) {
		if ((cpuRegs.interrupt & (1 << 31)) == 0) {
			INT(31, 4);
		}
	}
}

__inline void _cpuTestTIMR() {
	if (cpuRegs.CP0.n.Status.val & 0x8000 &&
		cpuRegs.CP0.n.Count >= cpuRegs.CP0.n.Compare) {
		SysPrintf("timr intr: %x, %x\n", cpuRegs.CP0.n.Count, cpuRegs.CP0.n.Compare);
		//Log=1;
		cpuException(0x808000, 0);
	}
}

void cpuTestHwInts() {
	cpuTestINTCInts();
	cpuTestDMACInts();
	cpuTestTIMRInts();
}

void cpuTestINTCInts() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) == 0x10001) {
		_cpuTestINTC();
	}
}

void cpuTestDMACInts() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) == 0x10001) {
		_cpuTestDMAC();
	}
}

void cpuTestTIMRInts() {
	if ((cpuRegs.CP0.n.Status.val & 0x10007) == 0x10001) {
		_cpuTestTIMR();
	}
}

void cpuExecuteBios() {
	SysPrintf("* PCSX2 *: ExecuteBios\n");

	while (cpuRegs.pc != 0x00200008 &&
		   cpuRegs.pc != 0x00100008) {
		Cpu->ExecuteBlock();
	}
	SysPrintf("* PCSX2 *: ExecuteBios Complete\n");
	GSprintf(5, "PCSX v" PCSX2_VERSION "\nExecuteBios Complete\n");
//	Log=1;
}

void cpuRestartCPU() {
	Cpu->Shutdown();
#ifndef DISABLE_REC
	if (Config.Cpu) Cpu = &intCpu;
	else Cpu = &recCpu;
#else
	Cpu = &intCpu;
#endif
	if (Cpu->Init() == -1) {
		SysClose();
		exit(1);
	}
	Cpu->Reset();
	psxRestartCPU();
}

