/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "Common.h"
#include "SPR.h"

DMACh *spr0;
DMACh *spr1;

void sprInit() {
	spr0 = (DMACh*)&psH[0xD000];
	spr1 = (DMACh*)&psH[0xD400];
}

__inline static void SPR0transfer(u32 *data, int size) {
/*	while (size > 0) {
#ifdef SPR_LOG
		SPR_LOG("SPR1transfer: %x\n", *data);
#endif
		data++; size--;
	}*/
	memcpy(data, &psS[spr0->sadr & 0x3fff], size*4);
	spr0->sadr+= size*4;
}

int  _SPR0chain() {
	u32 qwc = spr0->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;

	pMem = dmaGetAddr(spr0->madr);
	if (pMem == NULL) return -1;

	SPR0transfer(pMem, qwc*4);
	spr0->madr+= spr0->qwc*16;
	spr0->qwc = 0;
	return 0;
}

#define SPR0chain() \
	if (spr0->qwc) { \
		if (_SPR0chain() == -1) break; \
/*		cycles+= dma->qwc / BIAS;*/ /* guessing */ \
	}

void _SPR0interleave() {
	int qwc = spr0->qwc;
	int sqwc = psHu32(DMAC_SQWC) & 0xff;
	int tqwc = (psHu32(DMAC_SQWC) >> 16) & 0xff;

	while (qwc > 0) {
		spr0->qwc = tqwc; qwc-= tqwc;
		if (_SPR0chain() != 0) break;

		spr0->madr+= sqwc*16; qwc-= sqwc;
	}

	spr0->chcr &=~ 0x100;
	hwDmacIrq(8);
}

void _dmaSPR0() {
	u32 *ptag;
	int id;

	if ((spr0->chcr & 0xc) == 0x8) { // Interleave Mode
		_SPR0interleave();
		return;
	}

	// Transfer Dn_QWC from SPR to Dn_MADR
	if (_SPR0chain() != 0) {
		spr0->chcr&= ~0x100;
		hwDmacIrq(8);
		return;
	}

	if ((spr0->chcr & 0xc) == 0) { // Normal Mode
		spr0->chcr&= ~0x100;
		hwDmacIrq(8);
		return;
	}

	// Destination Chain Mode

	while (spr0->chcr & 0x100) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)&psS[spr0->sadr & 0x3fff];
		spr0->sadr+= 16;

		// Transfer dma tag if tte is set
		if (spr0->chcr & 0x40) SPR0transfer(ptag+2, 2);

		spr0->chcr = ( spr0->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

		id        = (ptag[0] >> 28) & 0x7;
		spr0->qwc  = (u16)ptag[0];
		spr0->madr = ptag[1];

#ifdef SPR_LOG
		SPR_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], spr0->qwc, id, spr0->madr);
#endif
		
		switch (id) {
			case 0: // cnts
				psHu32(DMAC_STADR) = spr0->madr;
				break;

			case 1: // cnt
				break;

			case 7: // end
				spr0->chcr &=~ (1<<8);
				break;
		}
		SPR0chain();

/*		if (spr0->chcr & 0x80 && ptag[0] >> 31) {
#ifdef SPR_LOG
			SPR_LOG("dmaIrq Set\n");
#endif
			spr0->chcr&= ~0x100;
			hwDmacIrq(8);
			return;
		}*/
	}

	spr0->chcr &=~ 0x100;
	hwDmacIrq(8);
}

void dmaSPR0() { // fromSPR
#ifdef SPR_LOG
	SPR_LOG("dmaSPR0 chcr = %lx, madr = %lx, qwc  = %lx, sadr = %lx\n",
			spr0->chcr, spr0->madr, spr0->qwc, spr0->sadr);
#endif

	if ((psHu32(DMAC_CTRL) & 0xC) == 0xC) { // GIF MFIFO
		MFIFO_GIFtransfer((DMACh*)&psH[0xA000], (spr0->sadr & 0x3fff) | 0x80000000,spr0->qwc);
		spr0->madr += (spr0->qwc<<4);
		spr0->chcr&= ~0x100;
		hwDmacIrq(8);
	} else
	if ((psHu32(DMAC_CTRL) & 0xC) == 0x8) { // VIF1 MFIFO
		MFIFO_VIF1transfer((DMACh*)&psH[0x9000], (spr0->sadr & 0x3fff) | 0x80000000, spr0->qwc);
		spr0->madr += (spr0->qwc<<4);
		spr0->chcr&= ~0x100;
		hwDmacIrq(8);
	} else {
		_dmaSPR0();
	}

}

__inline static void SPR1transfer(u32 *data, int size) {
/*	{
		int i;
		for (i=0; i<size; i++) {
#ifdef SPR_LOG 
		   SPR_LOG( "SPR1transfer[0x%x]: 0x%x\n", (spr1->sadr+i*4) & 0x3fff, data[i] );
#endif
		}
	}*/
	memcpy(&psS[spr1->sadr & 0x3fff], data, size*4);
	spr1->sadr+= size*4;
}

int  _SPR1chain() {
	u32 qwc = spr1->qwc;
	u32 *pMem;

	if (qwc == 0) return 0;

	pMem = dmaGetAddr(spr1->madr);
	if (pMem == NULL) return -1;

	SPR1transfer(pMem, qwc*4);
	spr1->madr+= spr1->qwc*16;
	spr1->qwc = 0;
	return qwc;
}

#define SPR1chain() \
	if (spr1->qwc) { \
		if (_SPR1chain() == -1) break; \
/*		cycles+= dma->qwc / BIAS;*/ /* guessing */ \
	}

void _SPR1interleave() {
	int qwc = spr1->qwc;
	int sqwc = psHu32(DMAC_SQWC) & 0xff;
	int tqwc = (psHu32(DMAC_SQWC) >> 16) & 0xff;

	while (qwc > 0) {
		spr1->qwc = tqwc; qwc-= tqwc;
		while (1) { SPR1chain(); break; }

		spr1->madr+= sqwc*16; qwc-= sqwc;
	}

	spr1->chcr &=~ 0x100;
	hwDmacIrq(9);
}

void dmaSPR1() { // toSPR
	u32 *ptag;
	u32 temp,finalAddress;
	int id, done=0;

#ifdef SPR_LOG
	SPR_LOG("dmaSPR1 chcr = 0x%x, madr = 0x%x, qwc  = 0x%x\n"
			"        tadr = 0x%x, sadr = 0x%x\n",
			spr1->chcr, spr1->madr, spr1->qwc,
			spr1->tadr, spr1->sadr);
#endif

	if ((spr1->chcr & 0xc) == 0x8) { // Interleave Mode
		_SPR1interleave();
		return;
	}

	// Transfer Dn_QWC from Dn_MADR to SPR1
	while (1) { SPR1chain(); break; }

	if ((spr1->chcr & 0xc) == 0) { // Normal Mode
		spr1->chcr &=~ 0x100;
		hwDmacIrq(9);
		return;
	}

	// Chain Mode

	while (done == 0) {  // Loop while Dn_CHCR.STR is 1
		ptag = (u32*)dmaGetAddr(spr1->tadr);
		if (ptag == NULL) {
			psHu32(DMAC_STAT)|= 1<<15;
			break;
		}

		// Transfer dma tag if tte is set
/*		if (spr1->chcr & 0x40) {
#ifdef SPR_LOG
			SPR_LOG("SPR TTE: %x_%x\n", ptag[3], ptag[2]);
#endif
			spr1->sadr-= 8;
			SPR1transfer(ptag+2, 2);
			spr1->sadr+= 8;
		}*/

		spr1->chcr = ( spr1->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );

		id        = (ptag[0] >> 28) & 0x7;
		spr1->qwc  = (u16)ptag[0];
		spr1->madr = ptag[1];

#ifdef SPR_LOG
		SPR_LOG("dmaChain %8.8x_%8.8x size=%d, id=%d, addr=%lx\n",
				ptag[1], ptag[0], spr1->qwc, id, spr1->madr);
#endif
		
		switch (id) {
			case 0: // refe
				spr1->tadr += 16;
				done = 1;
				break;

			case 1: // cnt
				spr1->madr = spr1->tadr + 16;
				// Set the taddr to the next tag
				spr1->tadr += 16 + (spr1->qwc * 16);
				break;

			case 2: // next
				temp = spr1->madr;
				spr1->madr = spr1->tadr + 16;
				spr1->tadr = temp;
				break;

			case 3: // ref
			case 4: // refs
				spr1->tadr += 16;
				break;

			case 5: // call
				temp = spr1->madr;
				finalAddress = spr1->tadr + 16 + (spr1->qwc * 16);

				if (spr1->asr0 == 0) 
					spr1->asr0 = finalAddress;
				else
					spr1->asr1 = finalAddress;

				spr1->madr = spr1->tadr + 16;
				spr1->tadr = temp;
				break;

			case 6: // ret
				spr1->madr = spr1->tadr + 16;
				if (spr1->asr1 != 0) {
					spr1->tadr = spr1->asr1;
					spr1->asr1 = 0;
				} else {
					if(spr1->asr0 != 0) {
						spr1->tadr = spr1->asr0;
						spr1->asr0 = 0;
					} else {
						spr1->tadr = 0;
						done = 1;
					}
				}
				break;

			case 7: // end
				spr1->madr = spr1->tadr + 16;
				done = 1;
				break;
		}
		SPR1chain();

/*		if (spr1->chcr & 0x80 && ptag[0] >> 31) {
			_dmaGetAddr(ptag, spr1->tadr, 9);
			spr1->chcr = ( spr1->chcr & 0xFFFF ) | ( (*ptag) & 0xFFFF0000 );
#ifdef SPR_LOG
			SPR_LOG("dmaIrq Set\n");
#endif
			spr1->chcr&= ~0x100;
			hwDmacIrq(9);
			return;
		}*/
	}

	spr1->chcr &=~ 0x100;
	hwDmacIrq(9);
}

