/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2005  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "VU0.h"
#include "VUmicro.h"
/*****************************************/
/*          NEW FLAGS                    */ //By asadr. Thnkx F|RES :p
/*****************************************/


__inline void vuUpdateDI(VURegs * VU) {
	u32 Flag_S = 0;
	u32 Flag_I = 0;
	u32 Flag_D = 0;

	/*
	FLAG D - I
	*/
	if ( (VU->macflag & 0x00F0 ) != 0 ) Flag_S = 1;
	Flag_I = (VU->statusflag >> 4) & 0x1;
	Flag_D = (VU->statusflag >> 5) & 0x1;

	VU->statusflag|= (Flag_I | Flag_S) << 10;
	VU->statusflag|= (Flag_D | Flag_S) << 11;
}

void VU_UPDATE_STAT_REG_DI0(VURegs * VU) {
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI1(VURegs * VU) {
	VU->statusflag|= 0x20; 
	VU->statusflag&=~0x10; 
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI2(VURegs * VU) {
	VU->statusflag|= 0x10; 
	VU->statusflag&=~0x20; 
	vuUpdateDI(VU);
}

void VU_UPDATE_STAT_REG_DI3(VURegs * VU) {
	VU->statusflag&=~0x20; 
	VU->statusflag&=~0x10; 
	vuUpdateDI(VU);
}

u32 VU_MACx_UPDATE(VURegs * VU, double x) {
	u64 v = *(u64*)&x;
	int exp = (v >> 52) & 0x7ff;
	int s = v >> 63;
	float f;

	if (s) VU->macflag|= 0x0080; 
	else   VU->macflag&=~0x0080;
	if (exp == 0) {
		VU->macflag|= 0x0008; 
		VU->macflag&=~0x8800;
		if (s) {
			return 0x80000000;
		} else {
			return 0x00000000;
		}
	} else {
		VU->macflag&=~0x0008;
		if (exp > 0x47e) {
			if (s) {
				VU->macflag|= 0x0800;
				VU->macflag&=~0x8000;
				return 0x7f800000 | ((v >> 29) & 0x7fffff);
			} else {
				VU->macflag|= 0x8000;
				VU->macflag&=~0x0800;
				return 0xff800000 | ((v >> 29) & 0x7fffff);
			}
		} else {
			VU->macflag&=~0x8800;
		}
	}

	if (exp == 0x47e) {
		if (s) {
			return 0x7f800000 | ((v >> 29) & 0x7fffff);
		} else {
			return 0xff800000 | ((v >> 29) & 0x7fffff);
		}
	}

	f = (float)x;
	return *(u32*)&f;
}

void VU_MACx_CLEAR(VURegs * VU) {
	VU->macflag&= ~0x8888;
}

u32 VU_MACy_UPDATE(VURegs * VU, double y) {
	u64 v = *(u64*)&y;
	int exp = (v >> 52) & 0x7ff;
	int s = v >> 63;
	float f;

	if (s) VU->macflag|= 0x0040; 
	else   VU->macflag&=~0x0040;
	if (exp == 0) {
		VU->macflag|= 0x0004; 
		VU->macflag&=~0x4400;
		if (s) {
			return 0x80000000;
		} else {
			return 0x00000000;
		}
	} else {
		VU->macflag&=~0x0004;
		if (exp > 0x47e) {
			if (s) {
				VU->macflag|= 0x0400;
				VU->macflag&=~0x4000;
				return 0x7f800000 | ((v >> 29) & 0x7fffff);
			} else {
				VU->macflag|= 0x4000;
				VU->macflag&=~0x0400;
				return 0xff800000 | ((v >> 29) & 0x7fffff);
			}
		} else {
			VU->macflag&=~0x4400;
		}
	}

	if (exp == 0x47e) {
		if (s) {
			return 0x7f800000 | ((v >> 29) & 0x7fffff);
		} else {
			return 0xff800000 | ((v >> 29) & 0x7fffff);
		}
	}

	f = (float)y;
	return *(u32*)&f;
}

void VU_MACy_CLEAR(VURegs * VU) {
	VU->macflag&= ~0x4444;
}

u32 VU_MACz_UPDATE(VURegs * VU, double z) {
	u64 v = *(u64*)&z;
	int exp = (v >> 52) & 0x7ff;
	int s = v >> 63;
	float f;

	if (s) VU->macflag|= 0x0020; 
	else   VU->macflag&=~0x0020;
	if (exp == 0) {
		VU->macflag|= 0x0002; 
		VU->macflag&=~0x2200;
		if (s) {
			return 0x80000000;
		} else {
			return 0x00000000;
		}
	} else {
		VU->macflag&=~0x0002;
		if (exp > 0x47e) {
			if (s) {
				VU->macflag|= 0x0200;
				VU->macflag&=~0x2000;
				return 0x7f800000 | ((v >> 29) & 0x7fffff);
			} else {
				VU->macflag|= 0x2000;
				VU->macflag&=~0x0200;
				return 0xff800000 | ((v >> 29) & 0x7fffff);
			}
		} else {
			VU->macflag&=~0x2200;
		}
	}

	if (exp == 0x47e) {
		if (s) {
			return 0x7f800000 | ((v >> 29) & 0x7fffff);
		} else {
			return 0xff800000 | ((v >> 29) & 0x7fffff);
		}
	}

	f = (float)z;
	return *(u32*)&f;
}

void VU_MACz_CLEAR(VURegs * VU) {
	VU->macflag&= ~0x2222;
}

u32 VU_MACw_UPDATE(VURegs * VU, double w) {
	u64 v = *(u64*)&w;
	int exp = (v >> 52) & 0x7ff;
	int s = v >> 63;
	float f;

	if (s) VU->macflag|= 0x0010; 
	else   VU->macflag&=~0x0010;
	if (exp == 0) {
		VU->macflag|= 0x0001; 
		VU->macflag&=~0x1100;
		if (s) {
			return 0x80000000;
		} else {
			return 0x00000000;
		}
	} else {
		VU->macflag&=~0x0001;
		if (exp > 0x47e) {
			if (s) {
				VU->macflag|= 0x0100;
				VU->macflag&=~0x1000;
				return 0x7f800000 | ((v >> 29) & 0x7fffff);
			} else {
				VU->macflag|= 0x1000;
				VU->macflag&=~0x0100;
				return 0xff800000 | ((v >> 29) & 0x7fffff);
			}
		} else {
			VU->macflag&=~0x1100;
		}
	}

	if (exp == 0x47e) {
		if (s) {
			return 0x7f800000 | ((v >> 29) & 0x7fffff);
		} else {
			return 0xff800000 | ((v >> 29) & 0x7fffff);
		}
	}

	f = (float)w;
	return *(u32*)&f;
}

void VU_MACw_CLEAR(VURegs * VU) {
	VU->macflag&= ~0x1111;
}

void VU_STAT_UPDATE(VURegs * VU) {
	int newZ=0, newS=0, newU=0, newO=0;
	u32 statusflag = VU->statusflag;

	if (VU->macflag & 0x000F) newZ = 1;
	if (VU->macflag & 0x00F0) newS = 1;
	if (VU->macflag & 0x0F00) newU = 1;
	if (VU->macflag & 0xF000) newO = 1;

	VU->statusflag = 0;
	VU->statusflag|= ((statusflag & 0x1) | newZ) << 6;
	VU->statusflag|= ((statusflag & 0x2) | newS) << 7;
	VU->statusflag|= ((statusflag & 0x4) | newU) << 8;
	VU->statusflag|= ((statusflag & 0x8) | newO) << 9;

	VU->statusflag|= newZ << 0;
	VU->statusflag|= newS << 1;
	VU->statusflag|= newU << 2;
	VU->statusflag|= newO << 3;
}

