/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"


#ifdef __WIN32__
#pragma warning(disable:4244)
#pragma warning(disable:4761)
#endif

/*********************************************************
* Register arithmetic                                    *
* Format:  OP rd, rs, rt                                 *
*********************************************************/

#ifndef ARITHMETIC_RECOMPILE

REC_FUNC(ADD);
REC_FUNC(ADDU);
REC_FUNC(DADD);
REC_FUNC(DADDU);
REC_FUNC(SUB);
REC_FUNC(SUBU);
REC_FUNC(DSUB);
REC_FUNC(DSUBU);
REC_FUNC(AND);
REC_FUNC(OR);
REC_FUNC(XOR);
REC_FUNC(NOR);
REC_FUNC(SLT);
REC_FUNC(SLTU);

#else

////////////////////////////////////////////////////
void recADD( void ) 
{
   if ( ! _Rd_ )
   {
      return;
   }

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Rt_ != 0 )
	   {
		   ADD32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   }
	   CDQ( );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );

}

////////////////////////////////////////////////////
void recADDU( void ) 
{
	recADD( );
}

////////////////////////////////////////////////////
void recDADD( void ) 
{
   if ( ! _Rd_ )
   {
      return;
   }
	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 1 ] );
	   if ( _Rt_ != 0 )
	   {
		   ADD32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
		   ADC32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
	   }
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );  
}

////////////////////////////////////////////////////
void recDADDU( void )
{
	recDADD( );
}

////////////////////////////////////////////////////
void recSUB( void ) 
{
       if ( ! _Rd_ ) return;
	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Rt_ != 0 )
	   {
		   SUB32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   }
	   CDQ( );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );
   
}

////////////////////////////////////////////////////
void recSUBU( void ) 
{
	recSUB( );
}

////////////////////////////////////////////////////
void recDSUB( void ) 
{
      if ( ! _Rd_ ) return;

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 1 ] );
	   if ( _Rt_ != 0 )
	   {
		   SUB32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
		   SBB32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
	   }
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
	   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );

}

////////////////////////////////////////////////////
void recDSUBU( void ) 
{
	recDSUB( );
}

////////////////////////////////////////////////////
void recAND( void )
{
   if ( ! _Rd_ )
   {
      return;
   }
	   if ( ( _Rt_ == 0 ) || ( _Rs_ == 0 ) )
	   {
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], 0 );	
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], 0 );	
	   }
      else
      {
	      MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
	      MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
	      PANDRtoR( MM0, MM1);
	      MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
	      SetMMXstate();
      }

}

////////////////////////////////////////////////////
void recOR( void )
{
   if ( ! _Rd_ )
   {
      return;
   }

	   if ( ( _Rs_ == 0 ) && ( _Rt_ == 0  ) )
	   {
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], 0x0 );
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], 0x0 );
	   }
      else if ( _Rs_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
		   SetMMXstate();
	   } 
      else if ( _Rt_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
		   SetMMXstate();
	   }
      else
      {
	      MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
	      MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
	      PORRtoR( MM0, MM1 );
	      MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
	      SetMMXstate();
      }
}

////////////////////////////////////////////////////
void recXOR( void ) 
{
   if ( ! _Rd_ )
   {
      return;
   }

	   if ( ( _Rs_ == 0 ) && ( _Rt_ == 0 ) )
	   {
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ],0x0);
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ],0x0);
		   return;
	   }

	   if ( _Rs_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
		   SetMMXstate();
		   return;
	   }

	   if ( _Rt_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
		   SetMMXstate();
		   return;
	   }

	   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
	   MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
	   PXORRtoR( MM0, MM1);
	   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ], MM0 );
	   SetMMXstate();
}

////////////////////////////////////////////////////
void recNOR( void ) 
{
   if ( ! _Rd_ )
   {
      return;
   }

	   if ( ( _Rs_ == 0 ) && ( _Rt_ == 0 ) )
	   {
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ],0xffffffff);
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ],0xffffffff);
		   return;
	   }

	   if ( _Rs_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ] );
		   PCMPEQDRtoR( MM1,MM1);
		   PXORRtoR( MM0,MM1);
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ],MM0);
		   SetMMXstate();
		   return;
	   }

	   if ( _Rt_ == 0 )
	   {
		   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
		   PCMPEQDRtoR( MM1,MM1);
		   PXORRtoR( MM0,MM1);
		   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ],MM0);
		   SetMMXstate();
		   return;
	   }

	   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ] );
	   PCMPEQDRtoR( MM1,MM1);
	   PORMtoR( MM0,(u32)&cpuRegs.GPR.r[ _Rt_ ] );
	   PXORRtoR( MM0,MM1);
	   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ],MM0);
	   SetMMXstate();
}

////////////////////////////////////////////////////
void recSLT( void )
{
   if ( ! _Rd_ )
   {
      return;
   }


      if ( _Rs_ == 0 )
      {
         XOR32RtoR( ECX, ECX );
         MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].SL[ 0 ] );
         XOR32RtoR( EDX, EDX );
         SUB32RtoR( ECX, EAX );
         MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].SL[ 1 ] );
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].SL[ 1 ] );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 1 ], EDX );
         SBB32RtoR( ECX, EAX );
         SETS8R( EDX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 0 ], EDX );
      }
      else if ( _Rt_ == 0 )
      {
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].SL[ 0 ] );
         XOR32RtoR( EAX, EAX );
         XOR32RtoR( EDX, EDX );
         SUB32RtoR( ECX, EAX );
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].SL[ 1 ] );
         SBB32RtoR( ECX, EAX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 1 ], EDX );
         SETS8R( EDX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 0 ], EDX );
      }
      else
      {
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].SL[ 0 ] );
         MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].SL[ 0 ] );
         XOR32RtoR( EDX, EDX );
         SUB32RtoR( ECX, EAX );
         MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].SL[ 1 ] );
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].SL[ 1 ] );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 1 ], EDX );
         SBB32RtoR( ECX, EAX );
         SETS8R( EDX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 0 ], EDX );
      }
}

////////////////////////////////////////////////////
void recSLTU( void )
{
   if ( ! _Rd_ )
   {
      return;
   }
      if ( _Rs_ == 0 )
      {
         XOR32RtoR( ECX, ECX );
         MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
         XOR32RtoR( EAX, EAX );
         CMP32RtoR( ECX, EDX );
         SETB8R( EAX );
         MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
         JNE8( 5 );
         CMP32RtoR( ECX, EDX );
         SETB8R( EAX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 1 ], ECX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 0 ], EAX );
      }
      else if ( _Rt_ == 0 )
      {
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 1 ] );
         XOR32RtoR( EDX, EDX );
         XOR32RtoR( EAX, EAX );
         CMP32RtoR( ECX, EDX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 1 ], EDX );
         SETB8R( EAX );
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
         JNE8( 5 );
         CMP32RtoR( ECX, EDX );
         SETB8R( EAX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].SL[ 0 ], EAX );
      }
      else
      {
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 1 ] );
         MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
         XOR32RtoR( EAX, EAX );
         CMP32RtoR( ECX, EDX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EAX );
         SETB8R( EAX );
         MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
         MOV32MtoR( EDX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
         JNE8( 5 );
         CMP32RtoR( ECX, EDX );
         SETB8R( EAX );
         MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
      }

}

#endif
