/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"
#include "VU0.h"

#ifdef __WIN32__
#pragma warning(disable:4244)
#pragma warning(disable:4761)
#endif

/*********************************************************
* Load and store for GPR                                 *
* Format:  OP rt, offset(base)                           *
*********************************************************/
#ifndef LOADSTORE_RECOMPILE

REC_FUNC(LB);
REC_FUNC(LBU);
REC_FUNC(LH);
REC_FUNC(LHU);
REC_FUNC(LW);
REC_FUNC(LWU);
REC_FUNC(LWL);
REC_FUNC(LWR);
REC_FUNC(LD);
REC_FUNC(LDR);
REC_FUNC(LDL);
REC_FUNC(LQ);
REC_FUNC(SB);
REC_FUNC(SH);
REC_FUNC(SW);
REC_FUNC(SWL);
REC_FUNC(SWR);
REC_FUNC(SD);
REC_FUNC(SDL);
REC_FUNC(SDR);
REC_FUNC(SQ);
REC_FUNC(LWC1);
REC_FUNC(SWC1);
REC_FUNC(LQC2);
REC_FUNC(SQC2);

#else

u64 retValue;
u64 dummyValue[ 4 ];

////////////////////////////////////////////////////
void recLB( void ) 
{


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead8 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ ) 
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   MOVSX32R8toR( EAX, EAX );
		   CDQ( );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], EDX );
         SETLINK8( linkEnd );
	   }
   
}

////////////////////////////////////////////////////
void recLBU( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead8 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ ) 
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   MOVZX32R8toR( EAX, EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], 0 );
         SETLINK8( linkEnd );
	   }
   
}

////////////////////////////////////////////////////
void recLH( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead16 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ )
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   MOVSX32R16toR( EAX, EAX );
		   CDQ( );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], EDX );
         SETLINK8( linkEnd );
	   }
  
}

////////////////////////////////////////////////////
void recLHU( void )
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead16 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ ) 
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   MOVZX32R16toR( EAX, EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], 0 );
         SETLINK8( linkEnd );
	   }
   
}

////////////////////////////////////////////////////
void recLW( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead32 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ ) 
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   CDQ( );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], EDX );
         SETLINK8( linkEnd );
	   }
   
}

////////////////////////////////////////////////////
void recLWU( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&retValue );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead32 );
      ADD32ItoR( ESP, 8 );
	   if ( _Rt_ )
      {
         u8* linkEnd;
         TEST32RtoR( EAX, EAX );
         linkEnd = JNZ8( 0 );
         MOV32MtoR( EAX, (u32)&retValue );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ], EAX );
		   MOV32ItoM( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ], 0 );
         SETLINK8( linkEnd );
	   }
   
}

////////////////////////////////////////////////////
void recLWL( void ) 
{

	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)LWL );
   
}

////////////////////////////////////////////////////
void recLWR( void ) 
{
 
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)LWR );
   
}

////////////////////////////////////////////////////
void recLD( void )
{


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      if ( _Rt_ )
      {
         PUSH32I( (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
      }
      else
      {
         PUSH32I( (u32)&retValue );
      }
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead64 );
      ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recLDL( void ) 
{

	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)LDL );
   
}

////////////////////////////////////////////////////
void recLDR( void ) 
{

	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)LDR );
   
}

////////////////////////////////////////////////////
void recLQ( void ) 
{


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_);
	   }
	   AND32ItoR( EAX, ~0xf );

      if ( _Rt_ )
      {
         PUSH32I( (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
      }
      else
      {
         PUSH32I( (u32)&dummyValue );
      }
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead128 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSB( void )
{
	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_);
	   }
	   PUSH32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite8 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSH( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
	   PUSH32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite16 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSW( void ) 
{
 

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
	   PUSH32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite32 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSWL( void ) 
{

	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)SWL );
   
}

////////////////////////////////////////////////////
void recSWR( void ) 
{
 
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
   	   CALLFunc( (u32)SWR );
   
}

////////////////////////////////////////////////////
void recSD( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
	   PUSH32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
	   PUSH32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite64 );
	   ADD32ItoR( ESP, 12 );
   
}

////////////////////////////////////////////////////
void recSDL( void ) 
{
 
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)SDL );
   
}

////////////////////////////////////////////////////
void recSDR( void ) 
{

	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	   MOV32ItoM( (u32)&cpuRegs.pc, pc );
	   iFlushCall();
	   CALLFunc( (u32)SDR );
   
}

////////////////////////////////////////////////////
void recSQ( void ) 
{
  


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
	   AND32ItoR( EAX, ~0xf );

	   PUSH32I( (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite128 );
	   ADD32ItoR( ESP, 8 );
   
}

/*********************************************************
* Load and store for COP1                                *
* Format:  OP rt, offset(base)                           *
*********************************************************/

////////////////////////////////////////////////////
void recLWC1( void )
{
	_freeXMMregs();

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )	
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }
      PUSH32I( (u32)&fpuRegs.fpr[ _Rt_ ].UL );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead32 );
      ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSWC1( void )
{
	_freeXMMregs();



	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )	
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }

	   PUSH32M( (u32)&fpuRegs.fpr[ _Rt_ ].UL );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite32 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////

#define _Ft_ _Rt_
#define _Fs_ _Rd_
#define _Fd_ _Sa_

void recLQC2( void ) 
{

	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_);
	   }

      if ( _Rt_ )
      {
         PUSH32I( (u32)&VU0.VF[_Ft_].UD[0] );
      }
      else
      {
         PUSH32I( (u32)&dummyValue );
      }
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memRead128 );
	   ADD32ItoR( ESP, 8 );
   
}

////////////////////////////////////////////////////
void recSQC2( void ) 
{


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	   if ( _Imm_ != 0 )
	   {
		   ADD32ItoR( EAX, _Imm_ );
	   }

	   PUSH32I( (u32)&VU0.VF[_Ft_].UD[0] );
	   PUSH32R( EAX );
	   iFlushCall();
	   CALLFunc( (u32)memWrite128 );
	   ADD32ItoR( ESP, 8 );
   
}

#endif
