/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2005  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************************************************
*   MMI opcodes                                          *
*                                                        *
*********************************************************/
#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"
#include "iMMI.h"

#ifndef MMI_RECOMPILE

REC_FUNC( MADD );
REC_FUNC( MADDU );
REC_FUNC( MADD1 );
REC_FUNC( MADDU1 );
REC_FUNC( PLZCW );

#ifndef MMI0_RECOMPILE

REC_FUNC( MMI0 );

#endif

#ifndef MMI1_RECOMPILE

REC_FUNC( MMI1 );

#endif

#ifndef MMI2_RECOMPILE

REC_FUNC( MMI2 );

#endif

#ifndef MMI3_RECOMPILE

REC_FUNC( MMI3 );

#endif

REC_FUNC( MFHI1 );
REC_FUNC( MFLO1 );
REC_FUNC( MTHI1 );
REC_FUNC( MTLO1 );
REC_FUNC( MULT1 );
REC_FUNC( MULTU1 );
REC_FUNC( DIV1 );
REC_FUNC( DIVU1 );
REC_FUNC( PMFHL );
REC_FUNC( PMTHL );

REC_FUNC( PSRLW );
REC_FUNC( PSRLH );

REC_FUNC( PSRAH );
REC_FUNC( PSRAW );

REC_FUNC( PSLLH );
REC_FUNC( PSLLW );

#else

REC_FUNC( MADD );
REC_FUNC( MADDU );
REC_FUNC( MADD1 );
REC_FUNC( MADDU1 );
REC_FUNC( PLZCW );
REC_FUNC( PMFHL );
REC_FUNC( PMTHL );

#ifndef MMI0_RECOMPILE

REC_FUNC( MMI0 );

#endif

#ifndef MMI1_RECOMPILE

REC_FUNC( MMI1 );

#endif

#ifndef MMI2_RECOMPILE

REC_FUNC( MMI2 );

#endif

#ifndef MMI3_RECOMPILE

REC_FUNC( MMI3 );

#endif

////////////////////////////////////////////////////
void recPSRLH( void )
{


	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSRLWItoR( MM0, _Sa_ );
	PSRLWItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSRLW( void )
{


	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSRLDItoR( MM0, _Sa_ );
	PSRLDItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSRAH( void )
{

	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSRAWItoR( MM0, _Sa_ );
	PSRAWItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSRAW( void )
{

	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSRADItoR( MM0, _Sa_ );
	PSRADItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSLLH( void )
{

	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSLLWItoR( MM0, _Sa_ );
	PSLLWItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSLLW( void )
{


	if ( ( _Rd_ == 0 ) || ( _Sa_ == 0 ) )
   {
      return;
   }
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PSLLDItoR( MM0, _Sa_ );
	PSLLDItoR( MM1, _Sa_ );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

/*
void recMADD( void ) 
{
}

void recMADDU( void ) 
{
}

void recPLZCW( void ) 
{
}
*/  

#ifdef MMI0_RECOMPILE

void recMMI0( void )
{
	recMMI0t[ _Sa_ ]( );
}

#endif

#ifdef MMI1_RECOMPILE

void recMMI1( void )
{
	recMMI1t[ _Sa_ ]( );
}

#endif

#ifdef MMI2_RECOMPILE

void recMMI2( void )
{
	recMMI2t[ _Sa_ ]( );
}

#endif

#ifdef MMI3_RECOMPILE

////////////////////////////////////////////////////
void recMMI3( void )
{
	recMMI3t[ _Sa_ ]( );
}

#endif

////////////////////////////////////////////////////
void recMTHI1( void )
{


	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQRtoM( (u32)&cpuRegs.HI.UD[ 1 ], MM0 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recMTLO1( void )
{


	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQRtoM( (u32)&cpuRegs.LO.UD[ 1 ], MM0 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recMFHI1( void )
{


	if ( ! _Rd_ ) return;

	MOVQMtoR( MM0, (u32)&cpuRegs.HI.UD[ 1 ] );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recMFLO1( void )
{


	if ( ! _Rd_ ) return;

	MOVQMtoR( MM0, (u32)&cpuRegs.LO.UD[ 1 ] );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recMULT1( void ) 
{


	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
   //	XOR32RtoR( EDX, EDX );
   //	CDQ();
	   IMUL32M( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );

	   MOV32RtoR( ECX, EDX );
	   MOV32RtoM( (u32)&cpuRegs.LO.UL[ 2 ], EAX );
	   CDQ( );
	   MOV32RtoM( (u32)&cpuRegs.LO.UL[ 3 ], EDX );
	   if ( _Rd_ != 0 )
	   {
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
		   MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );
	   }

	   MOV32RtoR( EAX, ECX );
	   CDQ( );
	   MOV32RtoM( (u32)&cpuRegs.HI.UL[ 2 ], EAX );
	   MOV32RtoM( (u32)&cpuRegs.HI.UL[ 3 ], EDX );
   
}

////////////////////////////////////////////////////
void recMULTU1( void ) 
{


	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
//	XOR32RtoR( EDX, EDX );
	MUL32M  ( (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );

	MOV32RtoR( ECX, EDX );
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 2 ], EAX );
	CDQ();
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 3 ], EDX );
	if ( _Rd_ != 0 )
	{
		MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
		MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EDX );
	}

	MOV32RtoR( EAX, ECX );
	CDQ();
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 2 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 3 ], EDX );

}

////////////////////////////////////////////////////
void recDIV1( void )
{


	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	CMP32ItoR( ECX, 0 );
	j8Ptr[ 0 ] = JE8(0 );

	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
//	XOR32RtoR( EDX, EDX );
	CDQ( );
	IDIV32R  ( ECX );

	MOV32RtoR( ECX, EDX );
	CDQ( );
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 2 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 3 ], EDX );

	MOV32RtoR( EAX, ECX );
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 2 ], ECX );
	CDQ( );
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 3 ], EDX );
	x86SetJ8( j8Ptr[ 0 ] );
}

////////////////////////////////////////////////////
void recDIVU1( void ) 
{


	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	CMP32ItoR( ECX, 0 );
	j8Ptr[ 0 ] = JE8(0 );

	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );

	XOR32RtoR( EDX, EDX );
	DIV32R  ( ECX );

	MOV32RtoR( ECX, EDX );
	CDQ();
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 2 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.LO.UL[ 3 ], EDX );

	MOV32RtoR( EAX, ECX );
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 2 ], ECX );
	CDQ();
	MOV32RtoM( (u32)&cpuRegs.HI.UL[ 3 ], EDX );
	x86SetJ8(j8Ptr[ 0 ] );
}

#endif

/*

void recMADD1( void ) 
{
}

void recMADDU1( void ) 
{
}

void recPMFHL( void ) 
{
}

void recPMTHL( void ) 
{
}

void recPSLLH( void ) 
{
}

void recPSRLH( void ) 
{
}

void recPSRAH( void ) 
{
}

void recPSLLW( void ) 
{
}

void recPSRLW( void ) 
{
}

void recPSRAW( void ) 
{
}*/

/*********************************************************
*   MMI0 opcodes                                         *
*                                                        *
*********************************************************/
#ifndef MMI0_RECOMPILE

REC_FUNC( PADDB );
REC_FUNC( PADDH );
REC_FUNC( PADDW );
REC_FUNC( PADDSB );
REC_FUNC( PADDSH );
REC_FUNC( PADDSW );
REC_FUNC( PSUBB );
REC_FUNC( PSUBH );
REC_FUNC( PSUBW );
REC_FUNC( PSUBSB );
REC_FUNC( PSUBSH );
REC_FUNC( PSUBSW );

REC_FUNC( PMAXW ); //x86 dont have PMAXSD :(  we can use PCMPGTD
REC_FUNC( PMAXH );        

REC_FUNC( PCGTW );
REC_FUNC( PCGTH );
REC_FUNC( PCGTB );

REC_FUNC( PEXTLW );

REC_FUNC( PPACW );        
REC_FUNC( PEXTLH );
REC_FUNC( PPACH );        
REC_FUNC( PEXTLB );
REC_FUNC( PPACB );
REC_FUNC( PEXT5 );
REC_FUNC( PPAC5 );

#else


////////////////////////////////////////////////////
REC_FUNC( PMAXW ); //x86 dont have PMAXSD :(  we can use PCMPGTD
////////////////////////////////////////////////////
REC_FUNC( PPACW );        
////////////////////////////////////////////////////
REC_FUNC( PEXTLH );
////////////////////////////////////////////////////
REC_FUNC( PPACH );        
////////////////////////////////////////////////////
REC_FUNC( PEXTLB );
////////////////////////////////////////////////////
REC_FUNC( PPACB );
////////////////////////////////////////////////////
REC_FUNC( PEXT5 );
////////////////////////////////////////////////////
REC_FUNC( PPAC5 );

////////////////////////////////////////////////////
void recPMAXH( void ) 
{
	//TODO:optimize RS | RT== 0 PXOR MMx, MMx,no second load.
	if ( ! _Rd_ ) return;


   if(cpucaps.hasStreamingSIMDExtensions) //if supports SSE or MMX2 for AMDS
   {
	 MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	 MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	 PMAXSWRtoR( MM0, MM1 );
	 MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	 MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	 PMAXSWRtoR( MM2, MM3);
	 MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	 MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM2);
	 SetMMXstate();
   }
   else
   {
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code ); 
	   MOV32ItoM( (u32)&cpuRegs.pc, pc ); 
	   CALLFunc( (u32)PMAXH ); 
   }

}

////////////////////////////////////////////////////
void recPCGTB( void )
{
	//TODO:optimize RS | RT== 0
	if ( ! _Rd_ ) return;



	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	PCMPGTBRtoR( MM0, MM1 );
	
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PCMPGTBRtoR( MM2, MM3);
	
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM2);
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPCGTH( void )
{
	//TODO:optimize RS | RT== 0
	if ( ! _Rd_ ) return;



   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );

	PCMPGTWRtoR( MM0, MM1 );
	PCMPGTWRtoR( MM2, MM3);
	
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM2);
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPCGTW( void )
{
	//TODO:optimize RS | RT== 0
	if ( ! _Rd_ ) return;


   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );

	PCMPGTDRtoR( MM0, MM1 );
	PCMPGTDRtoR( MM2, MM3);
	
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM2);
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDSB( void ) 
{
	if ( ! _Rd_ ) return;



   if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ], MM0 );
		MOVQMtoR( (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ], MM2);
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDSBRtoR( MM0, MM2);
	PADDSBRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDSH( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDSWRtoR( MM0, MM2);
	PADDSWRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDSW( void ) 
{


	if ( ! _Rd_ ) return;

   if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDSDRtoR( MM0, MM2);
	PADDSDRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBSB( void ) 
{
	if ( ! _Rd_ ) return;



   if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBSBRtoR( MM0, MM2);
	PSUBSBRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBSH( void ) 
{
	if ( ! _Rd_ ) return;



   if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBSWRtoR( MM0, MM2);
	PSUBSWRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBSW( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBSDRtoR( MM0, MM2 );
	PSUBSDRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDB( void )
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDBRtoR( MM0, MM2 );
	PADDBRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDH( void )
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDWRtoR( MM0, MM2 );
	PADDWRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDW( void ) 
{
	if ( ! _Rd_ ) return;


	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDDRtoR( MM0, MM2 );
	PADDDRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBB( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBBRtoR( MM0, MM2 );
	PSUBBRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBH( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBWRtoR( MM0, MM2 );
	PSUBWRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPSUBW( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}

	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PSUBDRtoR( MM0, MM2);
	PSUBDRtoR( MM1, MM3);
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPEXTLW( void ) 
{
	if ( ! _Rd_ ) return;


	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 0 ] );
	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 0 ] );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], ECX );

	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 1 ] );
	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 1 ] );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 3 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 2 ], ECX );
}

/*
void recPMAXW( void ) 
{
}
     
void recPPACW( void ) 
{
}        

void recPEXTLH( void ) 
{
}

void recPPACH( void ) 
{
}        

void recPEXTLB( void ) 
{
}

void recPPACB( void ) 
{
}        

void recPEXT5( void ) 
{
}

void recPPAC5( void ) 
{
}
*/

#endif

/*********************************************************
*   MMI1 opcodes                                         *
*                                                        *
*********************************************************/
#ifndef MMI1_RECOMPILE

REC_FUNC( PABSW );
REC_FUNC( PABSH );

REC_FUNC( PMINW ); 
REC_FUNC( PADSBH );
REC_FUNC( PMINH );
REC_FUNC( PCEQB );   
REC_FUNC( PCEQH );
REC_FUNC( PCEQW );

REC_FUNC( PADDUB );
REC_FUNC( PADDUH );
REC_FUNC( PADDUW );

REC_FUNC( PSUBUB );
REC_FUNC( PSUBUH );
REC_FUNC( PSUBUW );

REC_FUNC( PEXTUW );   
REC_FUNC( PEXTUH );
REC_FUNC( PEXTUB );
REC_FUNC( QFSRV ); 

#else

////////////////////////////////////////////////////
REC_FUNC( PABSW );
////////////////////////////////////////////////////
REC_FUNC( PABSH );

////////////////////////////////////////////////////
REC_FUNC( PMINW ); 
////////////////////////////////////////////////////
REC_FUNC( PADSBH );

////////////////////////////////////////////////////
REC_FUNC( PADDUW );

////////////////////////////////////////////////////
REC_FUNC( PSUBUB );
////////////////////////////////////////////////////
REC_FUNC( PSUBUH );
////////////////////////////////////////////////////
REC_FUNC( PSUBUW );

////////////////////////////////////////////////////
REC_FUNC( PEXTUH );
////////////////////////////////////////////////////
REC_FUNC( PEXTUB );
////////////////////////////////////////////////////
REC_FUNC( QFSRV ); 

/*
void recPABSW( void ) 
{
}

void recPABSH( void ) 
{
}

void recPMINH( void ) 
{
}

void recPMINW( void ) 
{
}

void recPADSBH( void ) 
{
}

void recPADDUH( void ) 
{
}

void recPADDUW( void ) 
{
}

void recPSUBUW( void ) 
{
}

void recPSUBUH( void ) 
{
}

void recPEXTUH( void ) 
{
}
*/

////////////////////////////////////////////////////
void recPEXTUW( void ) 
{
	if ( ! _Rd_ ) return;



	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 2 ] );
	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 2 ] );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], ECX );
	
	MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ _Rs_ ].UL[ 3 ] );
	MOV32MtoR( ECX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UL[ 3 ] );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 3 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 2 ], ECX );
}

////////////////////////////////////////////////////
void recPMINH( void ) 
{
	if ( ! _Rd_ ) return;


   if(cpucaps.hasStreamingSIMDExtensions) //if supports SSE or MMX2 for AMDS
   {
	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PMINSWRtoR( MM0, MM2 );
	PMINSWRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
   }
   else
   {
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code ); 
	   MOV32ItoM( (u32)&cpuRegs.pc, pc ); 
	   CALLFunc( (u32)PMINH ); 
   }

}

////////////////////////////////////////////////////
void recPCEQB( void )
{
	if ( ! _Rd_ ) return;


	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PCMPEQBRtoR( MM0, MM2 );
	PCMPEQBRtoR( MM1, MM3 );

	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPCEQH( void )
{
	if ( ! _Rd_ ) return;


	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PCMPEQWRtoR( MM0, MM2 );
	PCMPEQWRtoR( MM1, MM3 );
	
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPCEQW( void )
{
	if ( ! _Rd_ ) return;



	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PCMPEQDRtoR( MM0, MM2 );
	PCMPEQDRtoR( MM1, MM3 );
	
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDUB( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}
	
	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDUSBRtoR( MM0, MM2 );
	PADDUSBRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPADDUH( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 )
	{
		if ( _Rt_ != 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}
		else
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}
	}
	else
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	}
	
	if ( _Rt_ == 0 )
	{
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	else
	{
		MOVQMtoR( MM2, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM3, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	}
	PADDUSWRtoR( MM0, MM2 );
	PADDUSWRtoR( MM1, MM3 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}


/*
void recPSUBUB( void ) 
{
}

void recPEXTUB( void ) 
{
}

void recQFSRV( void ) 
{
}
*/

#endif
/*********************************************************
*   MMI2 opcodes                                         *
*                                                        *
*********************************************************/
#ifndef MMI2_RECOMPILE

REC_FUNC( PMFHI );
REC_FUNC( PMFLO );
REC_FUNC( PCPYLD );
REC_FUNC( PAND );
REC_FUNC( PXOR ); 

REC_FUNC( PMADDW );
REC_FUNC( PSLLVW );
REC_FUNC( PSRLVW ); 
REC_FUNC( PMSUBW );
REC_FUNC( PINTH );
REC_FUNC( PMULTW );
REC_FUNC( PDIVW );
REC_FUNC( PMADDH );
REC_FUNC( PHMADH );
REC_FUNC( PMSUBH );
REC_FUNC( PHMSBH );
REC_FUNC( PEXEH );
REC_FUNC( PREVH ); 
REC_FUNC( PMULTH );
REC_FUNC( PDIVBW );
REC_FUNC( PEXEW );
REC_FUNC( PROT3W ); 

#else

////////////////////////////////////////////////////
REC_FUNC( PMADDW );
////////////////////////////////////////////////////
REC_FUNC( PSLLVW );
////////////////////////////////////////////////////
REC_FUNC( PSRLVW ); 
////////////////////////////////////////////////////
REC_FUNC( PMSUBW );
////////////////////////////////////////////////////
REC_FUNC( PINTH );
////////////////////////////////////////////////////
REC_FUNC( PMULTW );
////////////////////////////////////////////////////
REC_FUNC( PDIVW );
////////////////////////////////////////////////////
REC_FUNC( PMADDH );
////////////////////////////////////////////////////
REC_FUNC( PHMADH );
////////////////////////////////////////////////////
REC_FUNC( PMSUBH );
////////////////////////////////////////////////////
REC_FUNC( PHMSBH );
////////////////////////////////////////////////////
REC_FUNC( PEXEH );
////////////////////////////////////////////////////
REC_FUNC( PREVH ); 
////////////////////////////////////////////////////
REC_FUNC( PMULTH );
////////////////////////////////////////////////////
REC_FUNC( PDIVBW );
////////////////////////////////////////////////////
REC_FUNC( PEXEW );
////////////////////////////////////////////////////
REC_FUNC( PROT3W ); 

////////////////////////////////////////////////////
void recPMFHI( void )
{
	if ( ! _Rd_ ) return;



	if (cpucaps.hasStreamingSIMDExtensions)
   {
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.HI.UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
	} 
   else 
   {
		MOVQMtoR( MM0, (u32)&cpuRegs.HI.UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.HI.UD[ 1 ] );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

////////////////////////////////////////////////////
void recPMFLO( void )
{
	if ( ! _Rd_ ) return;



	if (cpucaps.hasStreamingSIMDExtensions) 
   {
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.LO.UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
	} 
   else 
   {
		MOVQMtoR( MM0, (u32)&cpuRegs.LO.UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.LO.UD[ 1 ] );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

////////////////////////////////////////////////////
void recPAND( void ) 
{
	if ( ! _Rd_ ) return;


	if (cpucaps.hasStreamingSIMDExtensions) 
   {
		if ( _Rt_ == 0 || _Rs_ == 0 )
		{
			XORPSRtoR( XMM0,XMM0 );
			MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
			return;
		}

		MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		ANDPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
	} 
   else 
   {
		if ( _Rt_ == 0 || _Rs_ == 0 ) 
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}

		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		PANDMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		PANDMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

////////////////////////////////////////////////////
void recPXOR( void ) 
{
	if ( ! _Rd_ ) return;



	if (cpucaps.hasStreamingSIMDExtensions)
   {
		if ( _Rt_ == 0 && _Rs_ == 0 )
		{
			XORPSRtoR( XMM0,XMM0 );
			MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
			return;
		}

		if ( _Rt_ == 0 )
		{
			MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
			MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
			return;
		}
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		XORPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
	} 
   else 
   {
		if ( _Rt_ == 0 && _Rs_ == 0 )
		{
			PXORRtoR( MM0, MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
			SetMMXstate();
			return;
		}

		if ( _Rs_ == 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );

			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}

		if ( _Rt_ == 0 )
		{
			MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
			MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );

			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
			MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
			SetMMXstate();
			return;
		}

		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		PXORMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		PXORMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );

		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

/*
void recPMADDW( void ) 
{
}

void recPSLLVW( void ) 
{
}

void recPSRLVW( void ) 
{
} 

void recPMSUBW( void ) 
{
}

void recPMFLO( void ) 
{
}

void recPINTH( void ) 
{
}

void recPMULTW( void ) 
{
}

void recPDIVW( void ) 
{
}
*/

////////////////////////////////////////////////////
void recPCPYLD( void ) 
{
	if ( ! _Rd_ ) return;

 

	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM1 );
	SetMMXstate();
}

/*
void recPMADDH( void ) 
{
}

void recPHMADH( void ) 
{
}

void recPAND( void ) 
{
}

void recPXOR( void ) 
{
} 

void recPMSUBH( void ) 
{
}

void recPHMSBH( void ) 
{
}

void recPEXEH( void ) 
{
}

void recPREVH( void ) 
{
} 

void recPMULTH( void ) 
{
}

void recPDIVBW( void ) 
{
}

void recPEXEW( void ) 
{
}

void recPROT3W( void ) 
{
}
*/
#endif
/*********************************************************
*   MMI3 opcodes                                         *
*                                                        *
*********************************************************/
#ifndef MMI3_RECOMPILE

REC_FUNC( PMADDUW );
REC_FUNC( PSRAVW ); 
REC_FUNC( PMTHI );
REC_FUNC( PMTLO );
REC_FUNC( PINTEH );
REC_FUNC( PMULTUW );
REC_FUNC( PDIVUW );
REC_FUNC( PCPYUD );
REC_FUNC( POR );
REC_FUNC( PNOR );  
REC_FUNC( PCPYH ); 
REC_FUNC( PEXCW );
REC_FUNC( PEXCH );

#else

////////////////////////////////////////////////////
REC_FUNC( PMADDUW );
////////////////////////////////////////////////////
REC_FUNC( PSRAVW ); 

////////////////////////////////////////////////////
REC_FUNC( PINTEH );
////////////////////////////////////////////////////
REC_FUNC( PMULTUW );
////////////////////////////////////////////////////
REC_FUNC( PDIVUW );

////////////////////////////////////////////////////
REC_FUNC( PEXCW );
////////////////////////////////////////////////////
REC_FUNC( PEXCH );

////////////////////////////////////////////////////
void recPNOR( void ) 
{
	if ( ! _Rd_ ) return;



	if ( _Rs_ == 0 && _Rt_ == 0 )
	{
		PCMPEQDRtoR( MM0, MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM0 );
		SetMMXstate();
		return;
	}

	if ( _Rs_ == 0 )
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
		PCMPEQDRtoR( MM2, MM2);
		PXORRtoR( MM0, MM2);
		PXORRtoR( MM1, MM2);
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}
	
	if ( _Rt_ == 0 )
	{
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		PCMPEQDRtoR( MM2, MM2);
		PXORRtoR( MM0, MM2);
		PXORRtoR( MM1, MM2);
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
		return;
	}

	MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
	MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
	PORMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	PORMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
	PCMPEQDRtoR( MM2, MM2 );
	PXORRtoR( MM0, MM2 );
	PXORRtoR( MM1, MM2 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
	MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
	SetMMXstate();
}

////////////////////////////////////////////////////
void recPMTHI( void ) 
{


	if (cpucaps.hasStreamingSIMDExtensions)
   {
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.HI.UD[ 0 ],XMM0 );
	} 
   else 
   {
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		MOVQRtoM( (u32)&cpuRegs.HI.UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.HI.UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

////////////////////////////////////////////////////
void recPMTLO( void ) 
{


	if (cpucaps.hasStreamingSIMDExtensions)
   {
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVAPSRtoM( (u32)&cpuRegs.LO.UD[ 0 ],XMM0 );
	}
   else 
   {
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		MOVQRtoM( (u32)&cpuRegs.LO.UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.LO.UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

/*
void recPMADDUW( void ) 
{
}

void recPSRAVW( void ) 
{
} 

void recPINTEH( void ) 
{
}

void recPMULTUW( void ) 
{
}

void recPDIVUW( void ) 
{
}
*/

////////////////////////////////////////////////////
void recPCPYUD( void ) 
{
   if ( ! _Rd_ ) return;



   MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
   MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
   MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
   SetMMXstate();
}

////////////////////////////////////////////////////
void recPOR( void ) 
{
	if ( ! _Rd_ ) return;



	if (cpucaps.hasStreamingSIMDExtensions)
   {
		MOVAPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		if ( _Rt_ != 0 )
		{
			ORPSMtoR( XMM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
		}
		MOVAPSRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ],XMM0 );
	}
   else 
   {
		MOVQMtoR( MM0, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 0 ] );
		MOVQMtoR( MM1, (u32)&cpuRegs.GPR.r[ _Rs_ ].UD[ 1 ] );
		if ( _Rt_ != 0 )
		{
			PORMtoR ( MM0, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
			PORMtoR ( MM1, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 1 ] );
		}
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 0 ], MM0 );
		MOVQRtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UD[ 1 ], MM1 );
		SetMMXstate();
	}
}

////////////////////////////////////////////////////
void recPCPYH( void ) 
{
	if ( ! _Rd_ ) return;



	PUSH32R( EBX );
	MOVZX32M16toR( EAX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 0 ] );
	MOVZX32M16toR( EBX, (u32)&cpuRegs.GPR.r[ _Rt_ ].UD[ 2 ] );
	MOV32RtoR( ECX, EAX );
	MOV32RtoR( EDX, EBX );
	SHL32ItoR( EAX, 16 );
	SHL32ItoR( EBX, 16 );
	OR32RtoR( EAX, ECX );
	OR32RtoR( EBX, EDX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 0 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 1 ], EAX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 2 ], EBX );
	MOV32RtoM( (u32)&cpuRegs.GPR.r[ _Rd_ ].UL[ 3 ], EBX );
	POP32R( EBX );
} 

/*
void recPEXCH( void ) 
{
}

void recPEXCW( void ) 
{
}
*/

#endif


