/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2005  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"
#include "iR5900AritImm.h"
#include "iR5900Arit.h"
#include "iR5900MultDiv.h"
#include "iR5900Shift.h"
#include "iR5900Branch.h"
#include "iR5900Jump.h"
#include "iR5900LoadStore.h"
#include "iR5900Move.h"
#include "iMMI.h"
#include "iFPU.h"
#include "iCP0.h"
#include "iVU0micro.h"
#include "iVU1micro.h"
#include "VU.h"
#include "VU0.h"


#ifdef __MSCW32__
#pragma warning(disable:4244)
#pragma warning(disable:4761)
#endif

uptr *recLUT;

#undef PC_REC
#undef PC_REC8
#undef PC_REC16
#undef PC_REC32
#define PC_REC( X )    (recLUT[ X >> 16 ] + ( X & 0xffff ) )
#define PC_REC8( X )   (*(u8 *)PC_REC( X ))
#define PC_REC16( X )  (*(u16 *)PC_REC( X ))
#define PC_REC32( X )  (*(u32 *)PC_REC( X ))

#define X86

char *recMem;	   /* the recompiled blocks will be here */
char *recRAM;	   /* and the ptr to the blocks here */
char *recROM;	   /* and here */
char *recROM1;	   /* also here */
char *recPtr;
char *recMemVU0;	/* VU0 blocks */
char *recVU0;	   /* VU0 mem */
char *recMemVU1;	/* VU1 blocks */
char *recVU1;	   /* VU1 mem */
char *recPtrVU0;
char *recPtrVU1;

u32 hasSSE;
u32 has3DNOW;
u32 has3DNOW2;
u32 hasMMX2;
u32 hasSSE2;

// Temp to test
BOOL ConfigNewRec = TRUE;

//
u32 pc;			         /* recompiler pc */
int count;		         /* recompiler intruction count */
int branch;		         /* set for branch */
u32 target;		         /* branch target */
u64 _imm;		         /* temp immediate */
u64 _sa;			         /* temp sa */
u32 pcold;
int RunCount = 0;


#ifdef __x86_64__
char *txt0 = "RAX = %x : RDX = %x : RCX = %x\n";
char *txt0RC = "EAX = %x : EBX = %x : ECX = %x : EDX = %x : ESI = %x : EDI = %x\n";
char *txt1 = "REG[%d] = %x_%x\n";
char *txt2 = "M32 = %x\n";
#else
char *txt0 = "EAX = %x : ECX = %x : EDX = %x\n";
char *txt0RC = "EAX = %x : EBX = %x : ECX = %x : EDX = %x : ESI = %x : EDI = %x\n";
char *txt1 = "REG[%d] = %x_%x\n";
char *txt2 = "M32 = %x\n";
#endif

void recRecompile( void );
void recCOP22( void );


////////////////////////////////////////////////////
void Run( void )
{
	RunCount++;
	SysPrintf( "%x\n", RunCount );
}

////////////////////////////////////////////////////
/* uses ECX, result in ECX */
void ExtSign32to64( int reg ) 
{
	XOR32RtoR( ECX, ECX );
	TEST32ItoR( reg, 0x80000000 );
	j8Ptr[ 7 ] = JZ8( 0 );
	MOV32ItoR( ECX, 0xffffffff );
	x86SetJ8( j8Ptr[ 7 ] );
}

////////////////////////////////////////////////////
void iRet( BOOL freeRegs ) {
	/* store cycle */
	count = ( pc - pcold ) / 4;

	branch = 0;

	iFlushCall();
	ADD32ItoM( (uptr)&cpuRegs.cycle, count );

	/* store Count */
	ADD32ItoM( (uptr)&cpuRegs.CP0.n.Count, count * 2 );

#ifdef __x86_64__
	ADD64ItoR(ESP, 8);
#endif
	RET( );
}

////////////////////////////////////////////////////
void LogX86( void ) {
#ifdef __x86_64__
	SUB64ItoR(RSP, 15*8);
	MOV32RtoRm(RSP, RAX);
	MOV32RtoRm(RSP, RBX);
	MOV32RtoRm(RSP, RCX);
	MOV32RtoRm(RSP, RDX);
	MOV32RtoRm(RSP, RSI);
	MOV32RtoRm(RSP, RDI);
	MOV32RtoRm(RSP, RBP);
	MOV32RtoRm(RSP, R8 );
	MOV32RtoRm(RSP, R9 );
	MOV32RtoRm(RSP, R10);
	MOV32RtoRm(RSP, R11);
	MOV32RtoRm(RSP, R12);
	MOV32RtoRm(RSP, R13);
	MOV32RtoRm(RSP, R14);
	MOV32RtoRm(RSP, R15);

	MOV64ItoR(RSI, RDX);
	MOV64MtoR(RDI, (u64)&txt2);
	CALLFunc( (u32)SysPrintf );

	MOV32RmtoR(RAX, RSP);
	MOV32RmtoR(RBX, RSP);
	MOV32RmtoR(RCX, RSP);
	MOV32RmtoR(RDX, RSP);
	MOV32RmtoR(RSI, RSP);
	MOV32RmtoR(RDI, RSP);
	MOV32RmtoR(RBP, RSP);
	MOV32RmtoR(R8 , RSP);
	MOV32RmtoR(R9 , RSP);
	MOV32RmtoR(R10, RSP);
	MOV32RmtoR(R11, RSP);
	MOV32RmtoR(R12, RSP);
	MOV32RmtoR(R13, RSP);
	MOV32RmtoR(R14, RSP);
	MOV32RmtoR(R15, RSP);
	ADD64ItoR(RSP, 15*8);
#else
	PUSHA32( );

	PUSH32R( EDX );
	PUSH32R( ECX );
	PUSH32R( EAX );
	PUSH32M( (u32)&txt0 );
	CALLFunc( (u32)SysPrintf );
	ADD32ItoR( ESP, 4 * 4 );

	POPA32( );
#endif
}

////////////////////////////////////////////////////
void LogM32( u32 mem ) {
	PUSH32M( mem );
	PUSH32M( (u32)&txt2 );
	CALLFunc( (u32)SysPrintf );
	ADD32ItoR( ESP, 4 * 2 );
}

////////////////////////////////////////////////////
void LogR64( int reg ) {
	PUSH32M( (u32)&cpuRegs.GPR.r[ reg ].UL[ 0 ] );
	PUSH32M( (u32)&cpuRegs.GPR.r[ reg ].UL[ 1 ] );
	PUSH32I( reg );
	PUSH32M( (u32)&txt1 );
	CALLFunc( (u32)SysPrintf );
	ADD32ItoR(ESP, 4 * 4);
}

////////////////////////////////////////////////////
void iDumpBlock( char * ptr ) {
	FILE *f;
	char command[ 256 ];
	char filename[ 256 ];
	u32 i;

	SysPrintf( "dump1 %x:%x, %x\n", cpuRegs.pc, pc, cpuRegs.cycle );
#ifdef __WIN32__
	CreateDirectory("dumps", NULL);
	sprintf( filename, "dumps\\dump%.8X_%s.txt", cpuRegs.pc, Config.Regcaching ? "yes": "no" );
#else
	mkdir("dumps", 0755);
	sprintf( filename, "dumps/dump%.8X_%s.txt", cpuRegs.pc, Config.Regcaching ? "yes": "no" );
#endif

	for ( i = cpuRegs.pc; i < pc; i += 4 ) {
		SysPrintf( "%s\n", disR5900Fasm( PSMu32( i ), i ) );
	}

	fflush( stdout );
	f = fopen( "dump1", "wb" );
	fwrite( ptr, 1, (u32)x86Ptr - (u32)ptr, f );
	fclose( f );

#ifdef __x86_64__
	sprintf( command, "objdump -D --target=binary --architecture=i386:x86-64 dump1 > %s", filename );
#else
	sprintf( command, "objdump -D --target=binary --architecture=i386 dump1 > %s", filename );
#endif
	system( command );

	f = fopen( filename, "at" );
	fprintf( f, "\n\n" );
	for ( i = cpuRegs.pc; i < pc; i += 4 ) {
		fprintf( f, "%s\n", disR5900Fasm( PSMu32( i ), i ) );
	}
	fclose( f );
}


void iDumpRegisters( u32 _addr )
{
	FILE *f;
   char filename[ 256 ];

   sprintf( filename, "regdump\\r%.8X.%s", _addr, Config.Regcaching ? "yes": "no" );
   f = fopen( filename, "wb" );
   fwrite( &cpuRegs, 1, sizeof( cpuRegs ), f );
   fclose( f );
}

void iLogCallOrder( u32 _addr )
{
	FILE *f;

   f = fopen( "regdump\\callorder.txt", "at" );
   fprintf( f, "%.8X\n", _addr );
   fflush( f );
   fclose( f );
}

#define TYPE_UNUSED 0
#define TYPE_TEMP	1
#define TYPE_GPRREG	2

typedef struct {
	int type;
	int gprreg;
	int mode;
	int needed;
} _x86regs;

#ifdef __x86_64__
#define X86REGS 16
#else
#define X86REGS 8
#endif

_x86regs x86regs[X86REGS];

void _initX86regs() {
	memset(x86regs, 0, sizeof(x86regs));
	x86regs[EBX].needed = 2;
	x86regs[EBP].needed = 2;
	x86regs[ESP].needed = 2;
}

int  _getFreeX86reg() {
	int i;

	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].type == TYPE_UNUSED) return i;
	}
	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].needed) continue;

		_freeX86reg(i);
		return i;
	}
	SysPrintf("*PCSX2*: X86 ERROR\n");

	return -1;
}

int _allocTempX86reg(int x86reg) {
	if (x86reg == -1) {
		x86reg = _getFreeX86reg();
	}

	x86regs[x86reg].type = TYPE_TEMP;
	x86regs[x86reg].needed = 1;

//SysPrintf("_allocTempX86reg => %d\n", x86reg);
	return x86reg;
}

int _allocGPRtoX86reg(int x86reg, int gprreg, int mode) {
	int i;

	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].type != TYPE_GPRREG) continue;
		if (x86regs[i].gprreg != gprreg) continue;

		x86regs[i].mode|= mode;
//SysPrintf("_allocVFtoX86reg => %d (inuse)\n", i);
		return i;
	}

	if (x86reg == -1) {
		x86reg = _getFreeX86reg();
	}

	x86regs[x86reg].type = TYPE_GPRREG;
	x86regs[x86reg].gprreg = gprreg;
	x86regs[x86reg].mode = mode;
	x86regs[x86reg].needed = 1;
	if (mode & MODE_READ) {
		MOV64MtoR(x86reg, (uptr)&cpuRegs.GPR.r[x86regs[x86reg].gprreg].UD[0]);
	}

//SysPrintf("_allocGPRtoX86reg => %d\n", x86reg);
	return x86reg;
}

void _addNeededGPRtoX86reg(int gprreg) {
	int i;

	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].type != TYPE_GPRREG) continue;
		if (x86regs[i].gprreg != gprreg) continue;

		x86regs[i].needed = 1;
	}
}

void _clearNeededX86regs() {
	int i;

	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].needed == 2) continue;
		x86regs[i].needed = 0;
	}
}

void _freeX86reg(int x86reg) {
	if (x86regs[x86reg].type == TYPE_GPRREG &&
		x86regs[x86reg].mode & MODE_WRITE) {
		MOV64RtoM((uptr)&cpuRegs.GPR.r[x86regs[x86reg].gprreg].UD[0], x86reg);

	}
	x86regs[x86reg].type = TYPE_UNUSED;
}

void _freeX86regs() {
	int i;

//SysPrintf("_freeX86regs\n");
	for (i=0; i<X86REGS; i++) {
		if (x86regs[i].type == TYPE_UNUSED) continue;
		if (x86regs[i].needed == 2) continue;

		_freeX86reg(i);
	}
}


#define VU_VFx_ADDR(x)  (u32)&VU->VF[x].UL[0]
#define VU_ACCx_ADDR    (u32)&VU->ACC.UL[0]


#define XMMTYPE_TEMP	0
#define XMMTYPE_VFREG	1
#define XMMTYPE_ACC		2
#define XMMTYPE_FPREG	3

typedef struct {
	int inuse;
	int type;
	VURegs *VU;
	int reg;
	int mode;
	int needed;
} _xmmregs;

#ifdef __x86_64__
#define XMMREGS 16
#else
#define XMMREGS 8
#endif

_xmmregs xmmregs[XMMREGS];

void _initXMMregs() {
	memset(xmmregs, 0, sizeof(xmmregs));
}

int  _getFreeXMMreg() {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) return i;
	}
	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].needed == 1) continue;

		_freeXMMreg(i);
		return i;
	}
	SysPrintf("*PCSX2*: VUrec ERROR\n");

	return -1;
}

int _allocTempXMMreg(int xmmreg) {
	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg();
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = XMMTYPE_TEMP;
	xmmregs[xmmreg].needed = 1;

//SysPrintf("_allocTempXMMreg => %d\n", xmmreg);
	return xmmreg;
}

int _allocVFtoXMMreg(VURegs *VU, int xmmreg, int vfreg, int mode) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_VFREG) continue;
		if (xmmregs[i].reg != vfreg) continue;

		xmmregs[i].mode|= mode;
//SysPrintf("_allocVFtoXMMreg => %d (inuse)\n", i);
		return i;
	}

	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg();
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = XMMTYPE_VFREG;
	xmmregs[xmmreg].reg = vfreg;
	xmmregs[xmmreg].mode = mode;
	xmmregs[xmmreg].needed = 1;
	xmmregs[xmmreg].VU = VU;
	if (mode & MODE_READ) {
		MOVAPSMtoR(xmmreg, VU_VFx_ADDR(xmmregs[xmmreg].reg));
	}

//SysPrintf("_allocVFtoXMMreg => %d\n", xmmreg);
	return xmmreg;
}

int _allocACCtoXMMreg(VURegs *VU, int xmmreg, int mode) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_ACC) continue;

		xmmregs[i].mode|= mode;
		return i;
	}

	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg();
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = XMMTYPE_ACC;
	xmmregs[xmmreg].mode = mode;
	xmmregs[xmmreg].needed = 1;
	xmmregs[xmmreg].VU = VU;
	if (mode & MODE_READ) {
		MOVAPSMtoR(xmmreg, VU_ACCx_ADDR);
	}

	return xmmreg;
}

int _allocFPtoXMMreg(int xmmreg, int fpreg, int mode) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_FPREG) continue;
		if (xmmregs[i].reg != fpreg) continue;

		xmmregs[i].mode|= mode;
//SysPrintf("_allocVFtoXMMreg => %d (inuse)\n", i);
		return i;
	}

	if (xmmreg == -1) {
		xmmreg = _getFreeXMMreg();
	}

	xmmregs[xmmreg].inuse = 1;
	xmmregs[xmmreg].type = XMMTYPE_FPREG;
	xmmregs[xmmreg].reg = fpreg;
	xmmregs[xmmreg].mode = mode;
	xmmregs[xmmreg].needed = 1;
	if (mode & MODE_READ) {
		MOVSSMtoR(xmmreg, (u32)&fpuRegs.fpr[xmmregs[xmmreg].reg].f);
	}

//SysPrintf("_allocVFtoXMMreg => %d\n", xmmreg);
	return xmmreg;
}

void _addNeededVFtoXMMreg(int vfreg) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_VFREG) continue;
		if (xmmregs[i].reg != vfreg) continue;

		xmmregs[i].needed = 1;
	}
}

void _addNeededACCtoXMMreg() {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_ACC) continue;

		xmmregs[i].needed = 1;
	}
}

void _addNeededFPtoXMMreg(int fpreg) {
	int i;

	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;
		if (xmmregs[i].type != XMMTYPE_FPREG) continue;
		if (xmmregs[i].reg != fpreg) continue;

		xmmregs[i].needed = 1;
	}
}

void _clearNeededXMMregs() {
	int i;

	for (i=0; i<XMMREGS; i++) {
		xmmregs[i].needed = 0;
	}
}

void _freeXMMreg(int xmmreg) {
	if (xmmregs[xmmreg].type == XMMTYPE_VFREG &&
		xmmregs[xmmreg].mode & MODE_WRITE) {
		VURegs *VU = xmmregs[xmmreg].VU;
		MOVAPSRtoM(VU_VFx_ADDR(xmmregs[xmmreg].reg), xmmreg);
	}
	if (xmmregs[xmmreg].type == XMMTYPE_ACC &&
		xmmregs[xmmreg].mode & MODE_WRITE) {
		VURegs *VU = xmmregs[xmmreg].VU;
		MOVAPSRtoM(VU_ACCx_ADDR, xmmreg);
	}
	if (xmmregs[xmmreg].type == XMMTYPE_FPREG &&
		xmmregs[xmmreg].mode & MODE_WRITE) {
		MOVSSRtoM((u32)&fpuRegs.fpr[xmmregs[xmmreg].reg], xmmreg);
	}
	xmmregs[xmmreg].inuse = 0;
}

void _freeXMMregs() {
	int i;

//SysPrintf("_freeXMMregs\n");
	for (i=0; i<XMMREGS; i++) {
		if (xmmregs[i].inuse == 0) continue;

		_freeXMMreg(i);
	}
}


////////////////////////////////////////////////////
int recInit( void ) 
{
	int i;

   remove( "callorder.txt" );


	recLUT = (uptr*) malloc( 0x010000 * sizeof(uptr) );
	memset( recLUT, 0, 0x010000 * sizeof(uptr) );
	recMem = SysMmap(0x10000000, 0x00800000);
	recRAM = (char*) malloc( 0x02000000 );
	recROM = (char*) malloc( 0x00400000 );
	recROM1= (char*) malloc( 0x00040000 );
	recMemVU0 = SysMmap(0x10800000, 0x00100000);
	recMemVU1 = SysMmap(0x10900000, 0x00100000);

	recVU0 = (char*) malloc( 0x00001000 );
	recVU1 = (char*) malloc( 0x00004000 );
	if ( ( recRAM == NULL ) || ( recROM    == NULL ) || ( recROM1   == NULL ) || 
		  ( recMem == NULL ) || ( recLUT    == NULL ) || ( recMemVU0 == NULL ) ||
		  ( recVU0 == NULL ) || ( recMemVU1 == NULL ) || ( recVU1    == NULL ) ) 
   {
		SysMessage( _( "Error allocating memory" ) ); 
      return -1;
	}

	for ( i = 0x0000; i < 0x0200; i++ )
   {
      recLUT[ i + 0x0000 ] = (uptr)&recRAM[ i << 16 ];
	   recLUT[ i + 0x2000 ] = (uptr)&recRAM[ i << 16 ];
	   recLUT[ i + 0x3000 ] = (uptr)&recRAM[ i << 16 ];
   }

	for ( i = 0x0000; i < 0x0040; i++ )
   {
      recLUT[ i + 0x1fc0 ] = (uptr)&recROM[ i << 16 ];
	   recLUT[ i + 0x9fc0 ] = (uptr)&recROM[ i << 16 ];
	   recLUT[ i + 0xbfc0 ] = (uptr)&recROM[ i << 16 ];
   }

	for ( i = 0x0000; i < 0x0004; i++ )
   {
      recLUT[ i + 0x1e00 ] = (uptr)&recROM1[ i << 16 ];
	   recLUT[ i + 0x9e00 ] = (uptr)&recROM1[ i << 16 ];
	   recLUT[ i + 0xbe00 ] = (uptr)&recROM1[ i << 16 ];
   }

	memcpy( recLUT + 0x8000, recLUT, 0x2000 * sizeof(uptr) );
	memcpy( recLUT + 0xa000, recLUT, 0x2000 * sizeof(uptr) );

	
	SysPrintf( "x86Init: \n" );
	SysPrintf( "\tCPU vender name =  %s\n", cpuinfo.x86ID );
	SysPrintf( "\tFamilyID	=  %x\n", cpuinfo.x86StepID );
	SysPrintf( "\tx86Family =  %s\n", cpuinfo.x86Fam );
	SysPrintf( "\tCPU speed =  %d MHZ\n", cpuinfo.cpuspeed);
	SysPrintf( "\tx86PType  =  %s\n", cpuinfo.x86Type );
	SysPrintf( "\tx86Flags  =  %8.8x\n", cpuinfo.x86Flags );
	SysPrintf( "\tx86EFlags =  %8.8x\n", cpuinfo.x86EFlags );
   SysPrintf( "Features: \n" );
   SysPrintf( "\t%sDetected MMX\n",    cpucaps.hasMultimediaExtensions     ? "" : "Not " );
   SysPrintf( "\t%sDetected SSE\n",    cpucaps.hasStreamingSIMDExtensions  ? "" : "Not " );
   SysPrintf( "\t%sDetected SSE2\n",   cpucaps.hasStreamingSIMD2Extensions ? "" : "Not " );
	if ( cpuinfo.x86ID[0] == 'A' ) //AMD cpu
	{
      SysPrintf( " Extented AMD Features: \n" );
      SysPrintf( "\t%sDetected MMX2\n",     cpucaps.hasMultimediaExtensionsExt       ? "" : "Not " );
      SysPrintf( "\t%sDetected 3DNOW\n",    cpucaps.has3DNOWInstructionExtensions    ? "" : "Not " );
      SysPrintf( "\t%sDetected 3DNOW2\n",   cpucaps.has3DNOWInstructionExtensionsExt ? "" : "Not " );
	}
	if ( !( cpucaps.hasMultimediaExtensions  ) )
   {
		SysMessage( _( "Processor doesn't supports MMX, can't run recompiler without that" ) );
		return -1;
	}
		
	x86FpuState = FPU_STATE;

	return 0;
}

////////////////////////////////////////////////////
void recReset( void ) {
	memset( recRAM,  0, 0x02000000 );
	memset( recROM,  0, 0x00400000 );
	memset( recROM1, 0, 0x00040000 );
	memset( recVU0,  0, 0x00001000 );
	memset( recVU1,  0, 0x00004000 );

	recPtr = recMem;
	recPtrVU0 = recMemVU0;
	recPtrVU1 = recMemVU1;
	x86FpuState = FPU_STATE;
	iCWstate = 0;

	branch = 0;
	_imm = 0;
	_sa  = 0;
}

void recResetVU1( void ) {
	memset( recVU1,  0, 0x00004000 );

	recPtrVU1 = recMemVU1;
	x86FpuState = FPU_STATE;
	iCWstate = 0;

	branch = 0;
	_imm = 0;
	_sa  = 0;
}

////////////////////////////////////////////////////
void recError( void ) {
	SysReset( );
	ClosePlugins( );
	SysMessage( _( "Unrecoverable error while running recompiler" ) );
	SysRunGui( );
}

void recGetNextCode() {
	char * p = (char *)PSM( pc );
	if ( p == NULL ) {
		recError( );
	}

	cpuRegs.code = *(u32 *)p;
}

void _test() {
	Int_OpcodePrintTable[cpuRegs.code >> 26]();
}

////////////////////////////////////////////////////
void recompileNextInstruction( void ) {
	recGetNextCode();
//	SysPrintf("%s\n", disR5900Fasm(cpuRegs.code, pc));
	pc += 4;
	recBSC[ cpuRegs.code >> 26 ]( );
}

////////////////////////////////////////////////////
/* set a pending branch */
void SetBranch( ) {
   branch = 1;

   recompileNextInstruction( );

   MOV32MtoR( EAX, (u32)&target );
   MOV32RtoM( (u32)&cpuRegs.pc, EAX );
   iFlushCall();
   CALLFunc( (u32)cpuBranchTest );
}

void SetBranchReg( u32 reg ) {
   branch = 1;

   recGetNextCode();
   if (cpuRegs.code) {
	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ reg ].UL[ 0 ] );
	   MOV32RtoM( (u32)&target, EAX );

	   recompileNextInstruction( );

	   MOV32MtoR( EAX, (u32)&target );
	   MOV32RtoM( (u32)&cpuRegs.pc, EAX );
   } else {
	   MOV32MtoR( EAX, (u32)&cpuRegs.GPR.r[ reg ].UL[ 0 ] );
	   MOV32RtoM( (u32)&cpuRegs.pc, EAX );
	   pc += 4;
   }

   iFlushCall();
   CALLFunc( (u32)cpuBranchTest );
}

void SetBranchImm( u32 imm/*, int branchtest*/ ) {
	branch = 1;

	recompileNextInstruction( );
	MOV32ItoM( (u32)&cpuRegs.pc, imm );

//   if (branchtest) {
	iFlushCall();
      CALLFunc( (u32)cpuBranchTest );
/*   } else {
	   	branch = 0;
	   count+= ( pc - pcold ) / 4;
		pc = imm;
		pcold = pc;
   }*/
}

////////////////////////////////////////////////////
void execute( void ) {
	RECFUNC recFunc;
	char * p  = (char*)PC_REC( cpuRegs.pc );

	if ( p != NULL ) {
		recFunc = (RECFUNC)*(u32*)p;
	} else { 
		recError( ); 
		return; 
	}

	if ( recFunc == 0) {
//      u32 addr;
//      GRecLog( "recRecompile: recompiling PC: 0x%.8x\n", cpuRegs.pc );
		recRecompile( );
		recFunc = (RECFUNC)*(u32*)p;

//      addr = cpuRegs.pc;
//      iDumpBlock( (char*)*recFunc );
//      iLogCallOrder( addr );
//      GRecLog( "recRecompile: executing x86 at addr: 0x%.8x\n", *recFunc );
//      iDumpRegisters( addr );
//      GRecLog( "recRecompile: final PC: 0x%.8x\n", cpuRegs.pc );
	}
//	printf("execute %x\n", cpuRegs.pc);
	recFunc( );
//	printf("execute ok\n");
}

////////////////////////////////////////////////////
void recStep( void ) {
}

////////////////////////////////////////////////////
void recExecute( void ) {
	for (;;) execute();
}

////////////////////////////////////////////////////
void recExecuteBlock( void ) {
	execute();
}

////////////////////////////////////////////////////
void recClear( u32 Addr, u32 Size ) {
//   GRecReset( );
	memset( (void*)PC_REC(Addr), 0, Size * 4 );
}

////////////////////////////////////////////////////
void recShutdown( void ) {
	if ( recMem == NULL ) {
      return;
   }
	free( recLUT );
	SysMunmap(0x10000000, 0x00800000);
	free( recRAM );
	free( recROM );
	free( recROM1 );
	free( recVU0 );
	SysMunmap(0x10800000, 0x00100000);
	free( recVU1 );
	SysMunmap(0x10900000, 0x00100000);

	x86Shutdown( );

   remove( "dump1.txt" );
   remove( "dump1" );
}

void recEnableVU0micro(int enable) {
}

void recEnableVU1micro(int enable) {
}


#ifndef EMMS_TRACE

void SetMMXstate() {
	EMMS();
}

void SetFPUstate() {
}

#else

void SetMMXstate() {
	x86FpuState = MMX_STATE;
}

void SetFPUstate() {
	if (x86FpuState==MMX_STATE) {
		if (has3DNOWInstructionExtensions) { FEMMS(); x86FpuState=FPU_STATE; }
		else          {  EMMS(); x86FpuState=FPU_STATE; }
	}
}

#endif

void iFlushCall() {
	if (Config.Regcaching) _freeX86regs();
	_freeXMMregs();
	LoadCW();
	SetFPUstate();
}

////////////////////////////////////////////////////
void recNULL( void ) 
{
}

////////////////////////////////////////////////////
void recREGIMM( void ) 
{
	recREG[ _Rt_ ]( );
}

////////////////////////////////////////////////////
void recSPECIAL( void )
{
	recSPC[ _Funct_ ]( );
}

////////////////////////////////////////////////////
void recCOP0( void )
{
	recCP0[ _Rs_ ]( );
}

////////////////////////////////////////////////////
void recCOP0BC0( void )
{
	recCP0BC0[ ( cpuRegs.code >> 16 ) & 0x03 ]( );
}

////////////////////////////////////////////////////
void recCOP0C0( void ) 
{
	recCP0C0[ _Funct_ ]( );
}

////////////////////////////////////////////////////
void recCOP1( void ) {
	recCP1[ _Rs_ ]( );
}



////////////////////////////////////////////////////
#ifndef CP2_RECOMPILE

REC_SYS(COP2);

#else

void recCOP2( void )
{ 
#ifdef CPU_LOG
	CPU_LOG( "Recompiling COP2:%s\n", disR5900Fasm( cpuRegs.code, cpuRegs.pc ) );
#endif

   if ( Config.VUrec ) //disable the use of vus better this way :P
   {
	   MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code ); 
	   MOV32ItoM( (u32)&cpuRegs.pc, pc ); 
	   iFlushCall();
	   CALLFunc( (u32)COP2 ); 
	   branch = 2; 
   }
   else
   {
      recCOP22( );
   }
}

#endif

////////////////////////////////////////////////////
void recMMI( void ) 
{
	recMMIt[ _Funct_ ]( );
}


/*********************************************************
* Special purpose instructions                           *
* Format:  OP                                            *
*********************************************************/

////////////////////////////////////////////////////
void recSYSCALL( void ) {
	MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	MOV32ItoM( (u32)&cpuRegs.pc, pc );
	iFlushCall();
	CALLFunc( (u32)SYSCALL );
	branch = 2;
}

////////////////////////////////////////////////////
void recBREAK( void ) {
}


/**********************************************************
*    UNHANDLED YET OPCODES
*
**********************************************************/

////////////////////////////////////////////////////
REC_SYS(PREF);
////////////////////////////////////////////////////
REC_SYS(MFSA);
////////////////////////////////////////////////////
REC_SYS(MTSA);
////////////////////////////////////////////////////
REC_SYS(TGE);
////////////////////////////////////////////////////
REC_SYS(TGEU);
////////////////////////////////////////////////////
REC_SYS(TLT);
////////////////////////////////////////////////////
REC_SYS(TLTU);
////////////////////////////////////////////////////
REC_SYS(TEQ);
////////////////////////////////////////////////////
REC_SYS(TNE);
////////////////////////////////////////////////////
REC_SYS(TGEI);
////////////////////////////////////////////////////
REC_SYS(TGEIU);
////////////////////////////////////////////////////
REC_SYS(TLTI);
////////////////////////////////////////////////////
REC_SYS(TLTIU);
////////////////////////////////////////////////////
REC_SYS(TEQI);
////////////////////////////////////////////////////
REC_SYS(TNEI);
////////////////////////////////////////////////////
REC_SYS(MTSAB);
////////////////////////////////////////////////////
REC_SYS(MTSAH);


////////////////////////////////////////////////////
void recCACHE( void ) {
	MOV32ItoM( (u32)&cpuRegs.code, cpuRegs.code );
	MOV32ItoM( (u32)&cpuRegs.pc, pc );
	iFlushCall();
	CALLFunc( (u32)CACHE );
	branch = 2;
}

/*
void recPREF( void ) 
{
}

*/

////////////////////////////////////////////////////
void recSYNC( void )
{
}

/*
void recMFSA( void ) 
{
}

void recMTSA( void ) 
{
}

void recTGE( void ) 
{
}

void recTGEU( void ) 
{
}

void recTLT( void ) 
{
}

void recTLTU( void ) 
{
}

void recTEQ( void ) 
{
}

void recTNE( void ) 
{
}

void recTGEI( void )
{
}

void recTGEIU( void ) 
{
}

void recTLTI( void ) 
{
}

void recTLTIU( void ) 
{
}

void recTEQI( void ) 
{
}

void recTNEI( void ) 
{
}

void recMTSAB( void ) 
{
}

void recMTSAH( void ) 
{
}
*/



void (*recBSC[64] )() = {
    recSPECIAL, recREGIMM, recJ,    recJAL,   recBEQ,  recBNE,  recBLEZ,  recBGTZ,
    recADDI,    recADDIU,  recSLTI, recSLTIU, recANDI, recORI,  recXORI,  recLUI,
    recCOP0,    recCOP1,   recCOP2, recNULL,  recBEQL, recBNEL, recBLEZL, recBGTZL,
    recDADDI,   recDADDIU, recLDL,  recLDR,   recMMI,  recNULL, recLQ,    recSQ,
    recLB,      recLH,     recLWL,  recLW,    recLBU,  recLHU,  recLWR,   recLWU,
    recSB,      recSH,     recSWL,  recSW,    recSDL,  recSDR,  recSWR,   recCACHE,
    recNULL,    recLWC1,   recNULL, recPREF,  recNULL, recNULL, recLQC2,  recLD,
    recNULL,    recSWC1,   recNULL, recNULL,  recNULL, recNULL, recSQC2,  recSD
};

void (*recSPC[64] )() = {
    recSLL,  recNULL,  recSRL,  recSRA,  recSLLV,    recNULL,  recSRLV,   recSRAV,
    recJR,   recJALR,  recMOVZ, recMOVN, recSYSCALL, recBREAK, recNULL,   recSYNC,
    recMFHI, recMTHI,  recMFLO, recMTLO, recDSLLV,   recNULL,  recDSRLV,  recDSRAV,
    recMULT, recMULTU, recDIV,  recDIVU, recNULL,    recNULL,  recNULL,   recNULL,
    recADD,  recADDU,  recSUB,  recSUBU, recAND,     recOR,    recXOR,    recNOR,
    recMFSA, recMTSA,  recSLT,  recSLTU, recDADD,    recDADDU, recDSUB,   recDSUBU,
    recTGE,  recTGEU,  recTLT,  recTLTU, recTEQ,     recNULL,  recTNE,    recNULL,
    recDSLL, recNULL,  recDSRL, recDSRA, recDSLL32,  recNULL,  recDSRL32, recDSRA32
};

void (*recREG[32] )() = {
    recBLTZ,   recBGEZ,   recBLTZL,   recBGEZL,   recNULL, recNULL, recNULL, recNULL,
    recTGEI,   recTGEIU,  recTLTI,    recTLTIU,   recTEQI, recNULL, recTNEI, recNULL,
    recBLTZAL, recBGEZAL, recBLTZALL, recBGEZALL, recNULL, recNULL, recNULL, recNULL,
    recMTSAB,  recMTSAH,  recNULL,    recNULL,    recNULL, recNULL, recNULL, recNULL,
};

void (*recCP0[32] )() = {
    recMFC0,    recNULL, recNULL, recNULL, recMTC0, recNULL, recNULL, recNULL,
    recCOP0BC0, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
    recCOP0C0,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
    recNULL,    recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
};

void (*recCP0BC0[32] )() = {
    recBC0F, recBC0T, recBC0FL, recBC0TL, recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
};

void (*recCP0C0[64] )() = {
    recNULL, recTLBR, recTLBWI, recNULL, recNULL, recNULL, recTLBWR, recNULL,
    recTLBP, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recNULL, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recERET, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recNULL, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recNULL, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recNULL, recNULL, recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
    recEI,   recDI,   recNULL,  recNULL, recNULL, recNULL, recNULL,  recNULL,
};

void (*recCP1[32] )() = {
    recMFC1,     recNULL, recCFC1, recNULL, recMTC1,   recNULL, recCTC1, recNULL,
    recCOP1_BC1, recNULL, recNULL, recNULL, recNULL,   recNULL, recNULL, recNULL,
    recCOP1_S,   recNULL, recNULL, recNULL, recCOP1_W, recNULL, recNULL, recNULL,
    recNULL,     recNULL, recNULL, recNULL, recNULL,   recNULL, recNULL, recNULL,
};

void (*recCP1BC1[32] )() = {
    recBC1F, recBC1T, recBC1FL, recBC1TL, recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
    recNULL, recNULL, recNULL,  recNULL,  recNULL, recNULL, recNULL, recNULL,
};

void (*recCP1S[64] )() = {
	recADD_S,  recSUB_S,  recMUL_S,  recDIV_S, recSQRT_S, recABS_S,  recMOV_S,   recNEG_S, 
	recNULL,   recNULL,   recNULL,   recNULL,  recNULL,   recNULL,   recNULL,    recNULL,   
	recNULL,   recNULL,   recNULL,   recNULL,  recNULL,   recNULL,   recRSQRT_S, recNULL,  
	recADDA_S, recSUBA_S, recMULA_S, recNULL,  recMADD_S, recMSUB_S, recMADDA_S, recMSUBA_S,
	recNULL,   recNULL,   recNULL,   recNULL,  recCVT_W,  recNULL,   recNULL,    recNULL, 
	recMAX_S,  recMIN_S,  recNULL,   recNULL,  recNULL,   recNULL,   recNULL,    recNULL, 
	recC_F,    recNULL,   recC_EQ,   recNULL,  recC_LT,   recNULL,   recC_LE,    recNULL, 
	recNULL,   recNULL,   recNULL,   recNULL,  recNULL,   recNULL,   recNULL,    recNULL, 
};
 
void (*recCP1W[64] )() = { 
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   	
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recCVT_S, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,   
	recNULL,  recNULL, recNULL, recNULL, recNULL, recNULL, recNULL, recNULL,
};

void (*recMMIt[64] )() = {
	 recMADD,  recMADDU,  recNULL,  recNULL,  recPLZCW, recNULL, recNULL,  recNULL,
    recMMI0,  recMMI2,   recNULL,  recNULL,  recNULL,  recNULL, recNULL,  recNULL,
    recMFHI1, recMTHI1,  recMFLO1, recMTLO1, recNULL,  recNULL, recNULL,  recNULL,
    recMULT1, recMULTU1, recDIV1,  recDIVU1, recNULL,  recNULL, recNULL,  recNULL,
    recMADD1, recMADDU1, recNULL,  recNULL,  recNULL,  recNULL, recNULL,  recNULL,
    recMMI1 , recMMI3,   recNULL,  recNULL,  recNULL,  recNULL, recNULL,  recNULL,
    recPMFHL, recPMTHL,  recNULL,  recNULL,  recPSLLH, recNULL, recPSRLH, recPSRAH,
    recNULL,  recNULL,   recNULL,  recNULL,  recPSLLW, recNULL, recPSRLW, recPSRAW,
};

void (*recMMI0t[32] )() = { 
	recPADDW,  recPSUBW,  recPCGTW,  recPMAXW,       
	recPADDH,  recPSUBH,  recPCGTH,  recPMAXH,        
	recPADDB,  recPSUBB,  recPCGTB,  recNULL,
	recNULL,   recNULL,   recNULL,   recNULL,
	recPADDSW, recPSUBSW, recPEXTLW, recPPACW,        
	recPADDSH, recPSUBSH, recPEXTLH, recPPACH,        
	recPADDSB, recPSUBSB, recPEXTLB, recPPACB,        
	recNULL,   recNULL,   recPEXT5,  recPPAC5,
};

void (*recMMI1t[32] )() = { 
	recNULL,   recPABSW,  recPCEQW,  recPMINW, 
	recPADSBH, recPABSH,  recPCEQH,  recPMINH, 
	recNULL,   recNULL,   recPCEQB,  recNULL, 
	recNULL,   recNULL,   recNULL,   recNULL, 
	recPADDUW, recPSUBUW, recPEXTUW, recNULL,  
	recPADDUH, recPSUBUH, recPEXTUH, recNULL, 
	recPADDUB, recPSUBUB, recPEXTUB, recQFSRV, 
	recNULL,   recNULL,   recNULL,   recNULL, 
};

void (*recMMI2t[32] )() = { 
	recPMADDW, recNULL,   recPSLLVW, recPSRLVW, 
	recPMSUBW, recNULL,   recNULL,   recNULL,
	recPMFHI,  recPMFLO,  recPINTH,  recNULL,
	recPMULTW, recPDIVW,  recPCPYLD, recNULL,
	recPMADDH, recPHMADH, recPAND,   recPXOR, 
	recPMSUBH, recPHMSBH, recNULL,   recNULL, 
	recNULL,   recNULL,   recPEXEH,  recPREVH, 
	recPMULTH, recPDIVBW, recPEXEW,  recPROT3W, 
};

void (*recMMI3t[32] )() = { 
	recPMADDUW, recNULL,   recNULL,   recPSRAVW, 
	recNULL,    recNULL,   recNULL,   recNULL,
	recPMTHI,   recPMTLO,  recPINTEH, recNULL,
	recPMULTUW, recPDIVUW, recPCPYUD, recNULL,
	recNULL,    recNULL,   recPOR,    recPNOR,  
	recNULL,    recNULL,   recNULL,   recNULL,
	recNULL,    recNULL,   recPEXCH,  recPCPYH, 
	recNULL,    recNULL,   recPEXCW,  recNULL,
};

//#define ENABLE_BLOCKDUMP

////////////////////////////////////////////////////
void recRecompile( void ) {
#ifdef ENABLE_BLOCKDUMP
	char *ptr;
#endif

	/* if recPtr reached the mem limit reset whole mem */
	if ( ( (uptr)recPtr - (uptr)recMem ) >= ( 0x800000 - 0x8000 ) ) {
		recReset( );
	}
	x86SetPtr( recPtr );

	branch = 0;
	count = 0;

	x86Align(32);
	PC_REC32( cpuRegs.pc ) = (u32)x86Ptr;
#ifdef ENABLE_BLOCKDUMP
	ptr = x86Ptr;
#endif
	pc = cpuRegs.pc;
	pcold = pc;
	x86FpuState = FPU_STATE;
	iCWstate = 0;
	if (Config.Regcaching) _initX86regs();
	_initXMMregs();

#ifdef __x86_64__
	SUB64ItoR(ESP, 8);
#endif

	for ( count = 0; count < 1500; ) {
		recompileNextInstruction( );

		if (branch) break;
	}

	if (!branch) {
		MOV32ItoM( (u32)&cpuRegs.pc, pc );
	}

	iRet( FALSE );

#ifdef ENABLE_BLOCKDUMP
/*	if (Log)*/ iDumpBlock(ptr);
#endif
	recPtr = x86Ptr;
}

R5900cpu recCpu = {
	recInit,
	recReset,
	recStep,
	recExecute,
	recExecuteBlock,
	recExecuteVU0Block,
	recExecuteVU1Block,
	recEnableVU0micro,
	recEnableVU1micro,
	recClear,
	recClearVU0,
	recClearVU1,
	recShutdown
};
