/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Common.h"
#include "InterTables.h"
#include "ix86/ix86.h"
#include "iR5900.h"
#include "iMMI.h"
#include "iFPU.h"
#include "iCP0.h"
#include "VU0.h"
#include "VUmicro.h"
#include "VUflags.h"
#include "iVUmicro.h"
#include "iVUops.h"

#ifdef __MSCW32__
#pragma warning(disable:4244)
#pragma warning(disable:4761)
#endif

//Lower/Upper instructions can use that..
#define _Ft_ (( VU->code >> 16) & 0x1F)  // The rt part of the instruction register 
#define _Fs_ (( VU->code >> 11) & 0x1F)  // The rd part of the instruction register 
#define _Fd_ (( VU->code >>  6) & 0x1F)  // The sa part of the instruction register

#define _X (( VU->code>>24) & 0x1)
#define _Y (( VU->code>>23) & 0x1)
#define _Z (( VU->code>>22) & 0x1)
#define _W (( VU->code>>21) & 0x1)

#define _Fsf_ (( VU->code >> 21) & 0x03)
#define _Ftf_ (( VU->code >> 23) & 0x03)

#define _Imm11_ 	(s32)(VU->code & 0x400 ? 0xfffffc00 | (VU->code & 0x3ff) : VU->code & 0x3ff)
#define _UImm11_	(s32)(VU->code & 0x7ff)

#define VU_VFx_ADDR(x)  (u32)&VU->VF[x].UL[0]
#define VU_VFy_ADDR(x)  (u32)&VU->VF[x].UL[1]
#define VU_VFz_ADDR(x)  (u32)&VU->VF[x].UL[2]
#define VU_VFw_ADDR(x)  (u32)&VU->VF[x].UL[3]

#define VU_REGR_ADDR    (u32)&VU->VI[REG_R]
#define VU_REGI_ADDR    (u32)&VU->VI[REG_I]
#define VU_REGQ_ADDR    (u32)&VU->VI[REG_Q]
#define VU_REGMAC_ADDR  (u32)&VU->VI[REG_MAC_FLAG]

#define VU_VI_ADDR(x)   (u32)&VU->VI[x].UL

#define VU_ACCx_ADDR    (u32)&VU->ACC.UL[0]
#define VU_ACCy_ADDR    (u32)&VU->ACC.UL[1]
#define VU_ACCz_ADDR    (u32)&VU->ACC.UL[2]
#define VU_ACCw_ADDR    (u32)&VU->ACC.UL[3]


/********************************/
/*     HELP FUNCTIONS           */
/********************************/

#define _X_OR_Y   ((( VU->code >> 24 ) & 0x1 ) || (( VU->code >> 23 ) & 0x1 ) )
#define _Z_OR_W   ((( VU->code >> 22 ) & 0x1 ) || (( VU->code >> 21 ) & 0x1 ) )
#define _X_AND_Y  ((( VU->code >> 24 ) & 0x1 ) && (( VU->code >> 23 ) & 0x1 ) )
#define _Z_AND_W  ((( VU->code >> 22 ) & 0x1 ) && (( VU->code >> 21 ) & 0x1 ) )
#define _X_AND_Z  ((( VU->code >> 24 ) & 0x1 ) && (( VU->code >> 22 ) & 0x1 ) )
#define _Y_AND_Z  ((( VU->code >> 23 ) & 0x1 ) && (( VU->code >> 22 ) & 0x1 ) )
#define _Y_AND_W  ((( VU->code >> 23 ) & 0x1 ) && (( VU->code >> 21 ) & 0x1 ) )
#define _X_Y_Z_W  ((( VU->code >> 21 ) & 0xF ) )


float recMult_float_to_int4[4] = { 16.0, 16.0, 16.0, 16.0 };
float recMult_float_to_int12[4] = { 4096.0, 4096.0, 4096.0, 4096.0 };
float recMult_float_to_int15[4] = { 32768.0, 32768.0, 32768.0, 32768.0 };

float recMult_int_to_float4[4] = { 0.0625, 0.0625, 0.0625, 0.0625 };
float recMult_int_to_float12[4] = { 0.000244140625, 0.000244140625, 0.000244140625, 0.000244140625 };
float recMult_int_to_float15[4] = { 0.000030517578125, 0.000030517578125, 0.000030517578125, 0.000030517578125 };


#ifdef __MSCW32__
__declspec(align(16)) 
#endif
u32 SSEmovMask[ 16 ][ 4 ] =
{
{ 0x00000000, 0x00000000, 0x00000000, 0x00000000 },
{ 0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF },
{ 0x00000000, 0x00000000, 0xFFFFFFFF, 0x00000000 },
{ 0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0x00000000, 0xFFFFFFFF, 0x00000000, 0x00000000 },
{ 0x00000000, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF },
{ 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000 },
{ 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0x00000000, 0x00000000, 0x00000000 },
{ 0xFFFFFFFF, 0x00000000, 0x00000000, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0x00000000 },
{ 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000 },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000 },
{ 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF }
};

void _unpackVF_xyzw(int dstreg, int srcreg, int xyzw) {
	switch (xyzw) {
		case 0: PSHUFDRtoR(dstreg, srcreg, 0x00); break;
		case 1: PSHUFDRtoR(dstreg, srcreg, 0x55); break;
		case 2: PSHUFDRtoR(dstreg, srcreg, 0xaa); break;
		case 3: PSHUFDRtoR(dstreg, srcreg, 0xff); break;
	}
}


void _recvuFMACflush(VURegs * VU) {
	int i;

	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 0) continue;

		if ((vucycle - VU->fmac[i].sCycle) >= VU->fmac[i].Cycle) {
#ifdef VUM_LOG
//			if (Log) { VUM_LOG("flushing FMAC pipe[%d]\n", i); }
#endif
			VU->fmac[i].enable = 0;
		}
	}
}

void _recvuFDIVflush(VURegs * VU) {
	if (VU->fdiv.enable == 0) return;

	if ((vucycle - VU->fdiv.sCycle) >= VU->fdiv.Cycle) {
//		SysPrintf("flushing FDIV pipe\n");
		VU->fdiv.enable = 0;
	}
}

void _recvuEFUflush(VURegs * VU) {
	if (VU->efu.enable == 0) return;

	if ((vucycle - VU->efu.sCycle) >= VU->efu.Cycle) {
//		SysPrintf("flushing FDIV pipe\n");
		VU->efu.enable = 0;
	}
}

void _recvuTestPipes(VURegs * VU) {
	_recvuFMACflush(VU);
	_recvuFDIVflush(VU);
	_recvuEFUflush(VU);
}

void _recvuFMACTestStall(VURegs * VU, int reg, int xyzw) {
	int cycle;
	int i;

	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 0) continue;
		if (VU->fmac[i].reg == reg &&
			VU->fmac[i].xyzw & xyzw) break;
	}

	if (i == 8) return;

	cycle = VU->fmac[i].Cycle - (vucycle - VU->fmac[i].sCycle);
	VU->fmac[i].enable = 0;
//	SysPrintf("FMAC stall %d\n", cycle);
	vucycle+= cycle;
	_recvuTestPipes(VU);
}

void _recvuFMACAdd(VURegs * VU, int reg, int xyzw) {
	int i;

	/* find a free fmac pipe */
	for (i=0; i<8; i++) {
		if (VU->fmac[i].enable == 1) continue;
		break;
	}
	if (i==8) {
		SysPrintf("*PCSX2*: error , out of fmacs\n");
	}

#ifdef VUM_LOG
//	if (Log) { VUM_LOG("adding FMAC pipe[%d]; reg %d\n", i, reg); }
#endif
	VU->fmac[i].enable = 1;
	VU->fmac[i].sCycle = vucycle;
	VU->fmac[i].Cycle = 3;
	VU->fmac[i].xyzw = xyzw; 
	VU->fmac[i].reg = reg; 
}

void _recvuFDIVAdd(VURegs * VU, int cycles) {
//	SysPrintf("adding FDIV pipe\n");
	VU->fdiv.enable = 1;
	VU->fdiv.sCycle = vucycle;
	VU->fdiv.Cycle  = cycles;
}

void _recvuEFUAdd(VURegs * VU, int cycles) {
//	SysPrintf("adding EFU pipe\n");
	VU->efu.enable = 1;
	VU->efu.sCycle = vucycle;
	VU->efu.Cycle  = cycles;
}

void _recvuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn) {
	if (VUregsn->VFread0) {
		_recvuFMACTestStall(VU, VUregsn->VFread0, VUregsn->VFr0xyzw);
	}
	if (VUregsn->VFread1) {
		_recvuFMACTestStall(VU, VUregsn->VFread1, VUregsn->VFr1xyzw);
	}
}

void _recvuAddFMACStalls(VURegs * VU, _VURegsNum *VUregsn) {
	if (VUregsn->VFwrite) {
		_recvuFMACAdd(VU, VUregsn->VFwrite, VUregsn->VFwxyzw);
	}

	if (VUregsn->VIwrite & (1 << REG_CLIP_FLAG)) {
//		SysPrintf("REG_CLIP_FLAG pipe\n");
		_recvuFMACAdd(VU, -REG_CLIP_FLAG, 0);
	}
}

void _recvuFlushFDIV(VURegs * VU) {
	int cycle;

	if (VU->fdiv.enable == 0) return;

	cycle = VU->fdiv.Cycle - (vucycle - VU->fdiv.sCycle);
//	SysPrintf("waiting FDIV pipe %d\n", cycle);
	VU->fdiv.enable = 0;
	vucycle+= cycle;
}

void _recvuFlushEFU(VURegs * VU) {
	int cycle;

	if (VU->efu.enable == 0) return;

	cycle = VU->efu.Cycle - (vucycle - VU->efu.sCycle);
//	SysPrintf("waiting FDIV pipe %d\n", cycle);
	VU->efu.enable = 0;
	vucycle+= cycle;
}

void _recvuTestFDIVStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	_recvuFlushFDIV(VU);
}

void _recvuTestEFUStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	_recvuFlushEFU(VU);
}

void _recvuAddFDIVStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	if (VUregsn->VIwrite & (1 << REG_Q)) {
		_recvuFDIVAdd(VU, VUregsn->cycles);
	}
}

void _recvuAddEFUStalls(VURegs * VU, _VURegsNum *VUregsn) {
//	_vuTestFMACStalls(VURegs * VU, _VURegsNum *VUregsn);
	if (VUregsn->VIwrite & (1 << REG_P)) {
		_recvuEFUAdd(VU, VUregsn->cycles);
	}
}

void _recvuTestUpperStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuTestFMACStalls(VU, VUregsn); break;
	}
}

void _recvuTestLowerStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuTestFMACStalls(VU, VUregsn); break;
		case VUPIPE_FDIV: _recvuTestFDIVStalls(VU, VUregsn); break;
		case VUPIPE_EFU:  _recvuTestEFUStalls(VU, VUregsn); break;
	}
}

void _recvuAddUpperStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuAddFMACStalls(VU, VUregsn); break;
	}
}

void _recvuAddLowerStalls(VURegs * VU, _VURegsNum *VUregsn) {
	switch (VUregsn->pipe) {
		case VUPIPE_FMAC: _recvuAddFMACStalls(VU, VUregsn); break;
		case VUPIPE_FDIV: _recvuAddFDIVStalls(VU, VUregsn); break;
		case VUPIPE_EFU:  _recvuAddEFUStalls(VU, VUregsn); break;
	}
}


void _vurecAnalyzeOp(VURegs *VU, _vuopinfo *info) {
	_VURegsNum lregs;
	_VURegsNum uregs;
	u32 *ptr; 

//	SysPrintf("_vurecAnalyzeOp %x; %p\n", pc, info);
	ptr = (u32*)&VU->Micro[pc]; 
	pc += 8; 

	if (ptr[1] & 0x40000000) { 
		branch = 2; 
	} 
 
	VU->code = ptr[1];
	if (VU == &VU1) {
		VU1regs_UPPER_OPCODE[VU->code & 0x3f](&uregs);
	} else {
		VU0regs_UPPER_OPCODE[VU->code & 0x3f](&uregs);
	}
	_recvuTestUpperStalls(VU, &uregs);
	switch(VU->code & 0x3f) {
		case 0x10: case 0x11: case 0x12: case 0x13:
		case 0x14: case 0x15: case 0x16: case 0x17:
		case 0x19: case 0x1b:
		case 0x2b: case 0x2f:
			break;

		case 0x3c:
			switch ((VU->code >> 6) & 0x1f) {
				case 0x4: case 0x5:
					info->statusflag|= 4;
					info->macflag|= 4;
					break;
			}
			break;
		case 0x3d:
			switch ((VU->code >> 6) & 0x1f) {
				case 0x4: case 0x5: case 0x7:
					info->statusflag|= 4;
					info->macflag|= 4;
					break;
			}
			break;
		case 0x3e:
			switch ((VU->code >> 6) & 0x1f) {
				case 0x4: case 0x5:
					info->statusflag|= 4;
					info->macflag|= 4;
					break;
			}
			break;
		case 0x3f:
			switch ((VU->code >> 6) & 0x1f) {
				case 0x4: case 0x5: case 0x7: case 0xb:
					info->statusflag|= 4;
					info->macflag|= 4;
					break;
			}
			break;

		default:
			info->statusflag|= 4;
			info->macflag|= 4;
			break;
	}

	if (uregs.VIwrite & (1 << REG_CLIP_FLAG)) {
		info->clipflag = 1;
	}

	/* check upper flags */ 
	if (ptr[1] & 0x80000000) { /* I flag */ 
		info->cycle = vucycle;
	} else {
		VU->code = ptr[0]; 
		if (VU == &VU1) {
			VU1regs_LOWER_OPCODE[VU->code >> 25](&lregs);
		} else {
			VU0regs_LOWER_OPCODE[VU->code >> 25](&lregs);
		}
		_recvuTestLowerStalls(VU, &lregs);
		info->cycle = vucycle;

		if (lregs.pipe == VUPIPE_BRANCH) {
			branch = 1;
		}

		if (lregs.VIwrite & (1 << REG_Q)) {
//			SysPrintf("write to Q\n");
			info->q = 2;
			info->cycles = lregs.cycles;
		} else 
		if (lregs.pipe == VUPIPE_FDIV) {
			info->q = 1;
		}

		if (lregs.VIwrite & (1 << REG_P)) {
//			SysPrintf("write to P\n");
			info->p = 2;
			info->cycles = lregs.cycles;
		} else 
		if (lregs.pipe == VUPIPE_EFU) {
			info->p = 1;
		}

/*		if (lregs.VIread & (1 << REG_CLIP_FLAG)) {
			info->clipflag = 1;
		}*/

		if (lregs.VIread & (1 << REG_STATUS_FLAG)) {
			info->statusflag|= 2;
		}

		if (lregs.VIread & (1 << REG_MAC_FLAG)) {
			info->macflag|= 2;
		}

		_recvuAddLowerStalls(VU, &lregs);
	}
	_recvuAddUpperStalls(VU, &uregs);

	_recvuTestPipes(VU);
	vucycle++;
}

void _vurecAnalyzeBlock(VURegs *VU, _vuopinfo *info, int count) {
	int i, j;
	int qcycle=-1;
	int pcycle=-1;

	for (i=0; i<count; i++) {
//		SysPrintf("_vurecAnalyzeBlock %x: cycle=%d\n", pc, info[i].cycle);
		if (info[i].clipflag & 2) {
			for (j=i; j<count; j++) {
				if (info[j].clipflag) break;
			}
			if (j == count) {
			} else {
			}
		}

		if (info[i].statusflag & 2) {
			for (j=i; j>0; j--) {
				if (info[j].statusflag & 4 &&
					info[j].cycle <= (info[i].cycle-3)) break;
			}
			if (j > 0) j--;
			info[j].statusflag|= 1;
		}

		if (info[i].macflag & 2) {
			for (j=i; j>0; j--) {
				if (info[j].macflag & 4 &&
					info[j].cycle <= (info[i].cycle-3)) break;
			}
			if (j > 0) j--;
			info[j].macflag|= 1;
		}

		if (qcycle != -1 &&
			info[i].q == 1) {
			qcycle = -1;
		}

		if (info[i].q & 2) {
//			SysPrintf("write at q\n");
			if (qcycle != -1) {
				info[i-1].q = 1;
			}
			qcycle = info[i].cycle + info[i].cycles;
		}

		if (qcycle != -1 &&
			info[i].cycle > qcycle) {
///			SysPrintf("flush q\n");
			qcycle = -1;
			info[i-1].q = 1;
		}

		if (pcycle != -1 &&
			info[i].p == 1) {
			pcycle = -1;
		}

		if (info[i].p & 2) {
			if (pcycle != -1) {
				info[i-1].p = 1;
			}
			pcycle = info[i].cycle + info[i].cycles;
		}

		if (pcycle != -1 &&
			info[i].cycle >= pcycle) {
			pcycle = -1;
			info[i-1].p = 1;
		}
	}

	if (qcycle != -1) {
		/* hmm, this is really tricky, we still have a fdiv
		   pipe pending, but handling this over rec is really
		   hard, let's hope this works ok, else i'll gotta
		   kick my brain really hard to fix it :P (linuz)  */
		info[i-1].q = 1;
	}
	if (pcycle != -1) {
		info[i-1].p = 1;
	}
}

/******************************/
/*        VU Flags            */
/******************************/

void recResetFlags( void ) {
}



void recVU_MAC_UPDATE(VURegs * VU, int reg, int xyzw) {
	int t0reg = _allocTempXMMreg(-1);
	int t1reg;

	if (xyzw != 0) t1reg = _allocTempXMMreg(-1);

	switch(xyzw) {
		case 0:
			MOVD32XMMtoR(EAX, reg);
			break;
		case 1:
			PSHUFDRtoR(t0reg, reg, 0x55);
			MOVD32XMMtoR(EAX, t0reg);
			break;
		case 2:
			PSHUFDRtoR(t0reg, reg, 0xaa);
			MOVD32XMMtoR(EAX, t0reg);
			break;
		case 3:
			PSHUFDRtoR(t0reg, reg, 0xff);
			MOVD32XMMtoR(EAX, t0reg);
			break;
	}

	SHR32ItoR(EAX, 31);
	SHL32ItoR(EAX, 4+(3-xyzw));
	OR32RtoR(EDX, EAX);

	if (xyzw == 0) {
		XORPSRtoR(t0reg, t0reg);
		CMPEQSSRtoR(t0reg, reg);
		MOVD32XMMtoR(EAX, t0reg);
	} else {
		XORPSRtoR(t1reg, t1reg);
		CMPEQSSRtoR(t1reg, t0reg);
		MOVD32XMMtoR(EAX, t1reg);
	}
	AND32ItoR(EAX, 0x1);
	SHL32ItoR(EAX, 0+(3-xyzw));
	OR32RtoR(EDX, EAX);

	_freeXMMreg(t0reg);
	if (xyzw != 0) _freeXMMreg(t1reg);
}

void recVU_MAC_UPDATExyzw(VURegs * VU, int reg) {
	if (_X) { recVU_MAC_UPDATE(VU, reg, 0); }
	if (_Y) { recVU_MAC_UPDATE(VU, reg, 1); }
	if (_Z) { recVU_MAC_UPDATE(VU, reg, 2); }
	if (_W) { recVU_MAC_UPDATE(VU, reg, 3); }
}

void recVUMI_STAT_REG_UPDATE(VURegs *VU, int reg) {
	XOR32RtoR(EDX, EDX);

	if (_X_Y_Z_W == 0xf) {
		 recVU_MAC_UPDATExyzw(VU, reg);
	} else {
		if (_X) { recVU_MAC_UPDATE(VU, reg, 0); }
		if (_Y) { recVU_MAC_UPDATE(VU, reg, 1); }
		if (_Z) { recVU_MAC_UPDATE(VU, reg, 2); }
		if (_W) { recVU_MAC_UPDATE(VU, reg, 3); }
	}

	MOV32RtoM((uptr)&VU->macflag, EDX);
	MOV32MtoR(ECX, (uptr)&VU->statusflag);
	AND32ItoR(ECX, 0xF);
	SHL32ItoR(ECX, 6);

	TEST32ItoR(EDX, 0x000F);
	j8Ptr[0] = JZ8(0);
	OR32ItoR(ECX, (1<<6)|(1<<0));
	x86SetJ8(j8Ptr[0]); 

	TEST32ItoR(EDX, 0x00F0);
	j8Ptr[0] = JZ8(0);
	OR32ItoR(ECX, (1<<7)|(1<<1));
	x86SetJ8(j8Ptr[0]); 

	TEST32ItoR(EDX, 0x0F00);
	j8Ptr[0] = JZ8(0);
	OR32ItoR(ECX, (1<<8)|(1<<2));
	x86SetJ8(j8Ptr[0]); 

	TEST32ItoR(EDX, 0xF000);
	j8Ptr[0] = JZ8(0);
	OR32ItoR(ECX, (1<<9)|(1<<3));
	x86SetJ8(j8Ptr[0]); 

	MOV32RtoM((uptr)&VU->statusflag, ECX);
}


/******************************/
/*   VU Upper instructions    */
/******************************/

static u32 const_abs_table[16][4] = 
{
   { 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff },
   { 0xffffffff, 0xffffffff, 0xffffffff, 0x7fffffff },
   { 0xffffffff, 0xffffffff, 0x7fffffff, 0xffffffff },
   { 0xffffffff, 0xffffffff, 0x7fffffff, 0x7fffffff },
   { 0xffffffff, 0x7fffffff, 0xffffffff, 0xffffffff },
   { 0xffffffff, 0x7fffffff, 0xffffffff, 0x7fffffff },
   { 0xffffffff, 0x7fffffff, 0x7fffffff, 0xffffffff },
   { 0xffffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff },
   { 0x7fffffff, 0xffffffff, 0xffffffff, 0xffffffff },
   { 0x7fffffff, 0xffffffff, 0xffffffff, 0x7fffffff },
   { 0x7fffffff, 0xffffffff, 0x7fffffff, 0xffffffff },
   { 0x7fffffff, 0xffffffff, 0x7fffffff, 0x7fffffff },
   { 0x7fffffff, 0x7fffffff, 0xffffffff, 0xffffffff },
   { 0x7fffffff, 0x7fffffff, 0xffffffff, 0x7fffffff },
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0xffffffff },
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff },
};


void _vurecABS(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);

	MOVAPSRtoR(t0reg, fsreg);
	ANDPSMtoR(t0reg, (u32)&const_abs_table[ _X_Y_Z_W ][ 0 ] );

	ANDPSMtoR(ftreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
	ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
	ORPSRtoR(ftreg, t0reg);

	_freeXMMreg(t0reg);
}

void _vurecABSxyzw(VURegs *VU) {
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);

	if (_Ft_ == _Fs_) {
		ANDPSMtoR(ftreg, (u32)&const_abs_table[ _X_Y_Z_W ][ 0 ] );
	} else {
		MOVAPSRtoR(ftreg, fsreg);
		ANDPSMtoR(ftreg, (u32)&const_abs_table[ _X_Y_Z_W ][ 0 ] );
	}
}

void recVUMI_ABS(VURegs *VU) {
    if ( _Ft_ == 0 ) return;

	if (_X_Y_Z_W != 0xf) {
		_vurecABS(VU);
	} else {
		_vurecABSxyzw(VU);
	}

	_clearNeededXMMregs();
}

void recVUMI_ADD(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			ADDPSRtoR(fdreg, ftreg);
		} else
		if (_Fd_ == _Ft_) {
			ADDPSRtoR(fdreg, fsreg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			ADDPSRtoR(fdreg, ftreg);
		}
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_ADD_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			ADDPSRtoR(fdreg, t0reg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			ADDPSRtoR(fdreg, t0reg);
		}
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_ADD_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		ADDPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_ADDi(VURegs *VU) { recVUMI_ADD_iq(VU, VU_REGI_ADDR); }
void recVUMI_ADDq(VURegs *VU) { recVUMI_ADD_iq(VU, VU_REGQ_ADDR); }
void recVUMI_ADDx(VURegs *VU) { recVUMI_ADD_xyzw(VU, 0); }
void recVUMI_ADDy(VURegs *VU) { recVUMI_ADD_xyzw(VU, 1); }
void recVUMI_ADDz(VURegs *VU) { recVUMI_ADD_xyzw(VU, 2); }
void recVUMI_ADDw(VURegs *VU) { recVUMI_ADD_xyzw(VU, 3); }

void recVUMI_ADDA(VURegs *VU) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVAPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, ftreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, fsreg);
		ADDPSRtoR(accreg, ftreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_ADDA_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVSSMtoR(t0reg, addr); 
		PSHUFDRtoR(t0reg, t0reg, 0x00);

		ADDPSRtoR(t0reg, fsreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVSSMtoR(accreg, addr); 
		PSHUFDRtoR(accreg, accreg, 0x00);

		ADDPSRtoR(accreg, fsreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_ADDAi(VURegs *VU) { recVUMI_ADDA_iq(VU, VU_REGI_ADDR); }
void recVUMI_ADDAq(VURegs *VU) { recVUMI_ADDA_iq(VU, VU_REGQ_ADDR); }

void recVUMI_ADDA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);
	ADDPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_ADDAx(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 0); }
void recVUMI_ADDAy(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 1); }
void recVUMI_ADDAz(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 2); }
void recVUMI_ADDAw(VURegs *VU) { recVUMI_ADDA_xyzw(VU, 3); }

void recVUMI_SUB(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		SUBPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			SUBPSRtoR(fdreg, ftreg);
		} else
		if (_Fd_ == _Ft_) {
			t0reg = _allocTempXMMreg(-1);

			MOVAPSRtoR(t0reg, fsreg);
			SUBPSRtoR(t0reg, fdreg);
			MOVAPSRtoR(fdreg, t0reg);

			_freeXMMreg(t0reg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			SUBPSRtoR(fdreg, ftreg);
		}
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);

	_clearNeededXMMregs();
}

void recVUMI_SUB_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int t1reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_SUB_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int t1reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_SUBi(VURegs *VU) { recVUMI_SUB_iq(VU, VU_REGI_ADDR); }
void recVUMI_SUBq(VURegs *VU) { recVUMI_SUB_iq(VU, VU_REGQ_ADDR); }
void recVUMI_SUBx(VURegs *VU) { recVUMI_SUB_xyzw(VU, 0); }
void recVUMI_SUBy(VURegs *VU) { recVUMI_SUB_xyzw(VU, 1); }
void recVUMI_SUBz(VURegs *VU) { recVUMI_SUB_xyzw(VU, 2); }
void recVUMI_SUBw(VURegs *VU) { recVUMI_SUB_xyzw(VU, 3); }

void recVUMI_SUBA(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int accreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		SUBPSRtoR(t0reg, ftreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVAPSRtoR(accreg, fsreg);
		SUBPSRtoR(accreg, ftreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_SUBA_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int t1reg;
	int fsreg;
	int accreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVAPSRtoR(accreg, fsreg);
		SUBPSRtoR(accreg, t0reg);
	}

	_freeXMMreg(t0reg);

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_SUBA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int t1reg;
	int fsreg;
	int ftreg;
	int accreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, fsreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVAPSRtoR(accreg, fsreg);
		SUBPSRtoR(accreg, t0reg);
	}

	_freeXMMreg(t0reg);

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_SUBAi(VURegs *VU) { recVUMI_SUBA_iq(VU, VU_REGI_ADDR); }
void recVUMI_SUBAq(VURegs *VU) { recVUMI_SUBA_iq(VU, VU_REGQ_ADDR); }
void recVUMI_SUBAx(VURegs *VU) { recVUMI_SUBA_xyzw(VU, 0); }
void recVUMI_SUBAy(VURegs *VU) { recVUMI_SUBA_xyzw(VU, 1); }
void recVUMI_SUBAz(VURegs *VU) { recVUMI_SUBA_xyzw(VU, 2); }
void recVUMI_SUBAw(VURegs *VU) { recVUMI_SUBA_xyzw(VU, 3); }

void recVUMI_MUL(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			MULPSRtoR(fdreg, ftreg);
		} else
		if (_Fd_ == _Ft_) {
			MULPSRtoR(fdreg, fsreg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			MULPSRtoR(fdreg, ftreg);
		}
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MUL_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			MULPSRtoR(fdreg, t0reg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			MULPSRtoR(fdreg, t0reg);
		}
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MUL_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		if (_Fd_ == _Fs_) {
			MULPSRtoR(fdreg, t0reg);
		} else {
			MOVAPSRtoR(fdreg, fsreg);
			MULPSRtoR(fdreg, t0reg);
		}
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MULi(VURegs *VU) { recVUMI_MUL_iq(VU, VU_REGI_ADDR); }
void recVUMI_MULq(VURegs *VU) { recVUMI_MUL_iq(VU, VU_REGQ_ADDR); }
void recVUMI_MULx(VURegs *VU) { recVUMI_MUL_xyzw(VU, 0); }
void recVUMI_MULy(VURegs *VU) { recVUMI_MUL_xyzw(VU, 1); }
void recVUMI_MULz(VURegs *VU) { recVUMI_MUL_xyzw(VU, 2); }
void recVUMI_MULw(VURegs *VU) { recVUMI_MUL_xyzw(VU, 3); }

void recVUMI_MULA( VURegs *VU ) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		MOVAPSRtoR(accreg, fsreg);
		MULPSRtoR(accreg, ftreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MULA_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		MOVSSMtoR(t0reg, addr); 
		PSHUFDRtoR(t0reg, t0reg, 0x00);

		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);

		MOVSSMtoR(accreg, addr); 
		PSHUFDRtoR(accreg, accreg, 0x00);

		MULPSRtoR(accreg, fsreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MULAi(VURegs *VU) { recVUMI_MULA_iq(VU, VU_REGI_ADDR); }
void recVUMI_MULAq(VURegs *VU) { recVUMI_MULA_iq(VU, VU_REGQ_ADDR); }

void recVUMI_MULA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);

		_unpackVF_xyzw(t0reg, ftreg, xyzw);
		MULPSRtoR(t0reg, fsreg);

		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);
	} else {
		accreg = _allocACCtoXMMreg(VU, -1, MODE_WRITE);
		_unpackVF_xyzw(accreg, ftreg, xyzw);
		MULPSRtoR(accreg, fsreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MULAx(VURegs *VU) { recVUMI_MULA_xyzw(VU, 0); }
void recVUMI_MULAy(VURegs *VU) { recVUMI_MULA_xyzw(VU, 1); }
void recVUMI_MULAz(VURegs *VU) { recVUMI_MULA_xyzw(VU, 2); }
void recVUMI_MULAw(VURegs *VU) { recVUMI_MULA_xyzw(VU, 3); }

void recVUMI_MADD(VURegs *VU) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, ftreg);
		ADDPSRtoR(fdreg, accreg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MADD_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
		ADDPSRtoR(fdreg, accreg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MADD_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MOVAPSRtoR(fdreg, fsreg);
		MULPSRtoR(fdreg, t0reg);
		ADDPSRtoR(fdreg, accreg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MADDi(VURegs *VU) { recVUMI_MADD_iq(VU, VU_REGI_ADDR); }
void recVUMI_MADDq(VURegs *VU) { recVUMI_MADD_iq(VU, VU_REGQ_ADDR); }
void recVUMI_MADDx(VURegs *VU) { recVUMI_MADD_xyzw(VU, 0); }
void recVUMI_MADDy(VURegs *VU) { recVUMI_MADD_xyzw(VU, 1); }
void recVUMI_MADDz(VURegs *VU) { recVUMI_MADD_xyzw(VU, 2); }
void recVUMI_MADDw(VURegs *VU) { recVUMI_MADD_xyzw(VU, 3); }

void recVUMI_MADDA( VURegs *VU ) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);

	MOVAPSRtoR(t0reg, fsreg);
	MULPSRtoR(t0reg, ftreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MADDA_iq( VURegs *VU, u32 addr) {
	int t0reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MADDAi( VURegs *VU ) { recVUMI_MADDA_iq( VU, VU_REGI_ADDR ); }
void recVUMI_MADDAq( VURegs *VU ) { recVUMI_MADDA_iq( VU, VU_REGQ_ADDR ); }

void recVUMI_MADDA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);
	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		ADDPSRtoR(t0reg, accreg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ORPSRtoR(accreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ADDPSRtoR(accreg, t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MADDAx( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 0); }
void recVUMI_MADDAy( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 1); }
void recVUMI_MADDAz( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 2); }
void recVUMI_MADDAw( VURegs *VU ) { recVUMI_MADDA_xyzw(VU, 3); }

void recVUMI_MSUB(VURegs *VU) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);
		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
		_freeXMMreg(t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MULPSRtoR(t0reg, ftreg);
		MOVAPSRtoR(fdreg, accreg);
		SUBPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MSUB_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t1reg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		MOVAPSRtoR(fdreg, accreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MSUB_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;
	int ftreg;
	int fdreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	if (_Fd_) _addNeededVFtoXMMreg(_Fd_);
	_addNeededACCtoXMMreg(VU);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);
		t1reg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		if (_Fd_) fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);
		else fdreg = _allocTempXMMreg(-1);

		MULPSRtoR(t0reg, fsreg);
		MOVAPSRtoR(fdreg, accreg);
		SUBPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, fdreg);
	}
	if (_Fd_ == 0) _freeXMMreg(fdreg);
	_clearNeededXMMregs();
}

void recVUMI_MSUBi(VURegs *VU) { recVUMI_MSUB_iq(VU, VU_REGI_ADDR); }
void recVUMI_MSUBq(VURegs *VU) { recVUMI_MSUB_iq(VU, VU_REGQ_ADDR); }
void recVUMI_MSUBx(VURegs *VU) { recVUMI_MSUB_xyzw(VU, 0); }
void recVUMI_MSUBy(VURegs *VU) { recVUMI_MSUB_xyzw(VU, 1); }
void recVUMI_MSUBz(VURegs *VU) { recVUMI_MSUB_xyzw(VU, 2); }
void recVUMI_MSUBw(VURegs *VU) { recVUMI_MSUB_xyzw(VU, 3); }

void recVUMI_MSUBA( VURegs *VU ) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);

	MOVAPSRtoR(t0reg, fsreg);
	MULPSRtoR(t0reg, ftreg);

	if (_X_Y_Z_W != 0xf) {
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		SUBPSRtoR(accreg, t0reg);
	}

	_freeXMMreg(t0reg);
	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MSUBA_iq( VURegs *VU, u32 addr) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		SUBPSRtoR(accreg, t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MSUBA_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int t1reg;
	int accreg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededACCtoXMMreg(VU);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);
	accreg = _allocACCtoXMMreg(VU, -1, MODE_READ | MODE_WRITE);
	t0reg = _allocTempXMMreg(-1);
	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	MULPSRtoR(t0reg, fsreg);

	if (_X_Y_Z_W != 0xf) {
		t1reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t1reg, accreg);
		SUBPSRtoR(t1reg, t0reg);

		ANDPSMtoR(accreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t1reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(accreg, t1reg);

		_freeXMMreg(t1reg);
	} else {
		SUBPSRtoR(accreg, t0reg);
	}

	if (cinfo->statusflag & 0x1 ||
		cinfo->macflag & 0x1) {
		recVUMI_STAT_REG_UPDATE(VU, accreg);
	}
	_clearNeededXMMregs();
}

void recVUMI_MSUBAi( VURegs *VU ) { recVUMI_MSUBA_iq( VU, VU_REGI_ADDR ); }
void recVUMI_MSUBAq( VURegs *VU ) { recVUMI_MSUBA_iq( VU, VU_REGQ_ADDR ); }
void recVUMI_MSUBAx( VURegs *VU ) { recVUMI_MSUBA_xyzw(VU, 0); }
void recVUMI_MSUBAy( VURegs *VU ) { recVUMI_MSUBA_xyzw(VU, 1); }
void recVUMI_MSUBAz( VURegs *VU ) { recVUMI_MSUBA_xyzw(VU, 2); }
void recVUMI_MSUBAw( VURegs *VU ) { recVUMI_MSUBA_xyzw(VU, 3); }

void recVUMI_MAX(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededVFtoXMMreg(_Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MAXPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, ftreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_MAX_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MAXPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

void recVUMI_MAX_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MAXPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MAXPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

void recVUMI_MAXi(VURegs *VU) { recVUMI_MAX_iq(VU, VU_REGI_ADDR); }
void recVUMI_MAXx(VURegs *VU) { recVUMI_MAX_xyzw(VU, 0); }
void recVUMI_MAXy(VURegs *VU) { recVUMI_MAX_xyzw(VU, 1); }
void recVUMI_MAXz(VURegs *VU) { recVUMI_MAX_xyzw(VU, 2); }
void recVUMI_MAXw(VURegs *VU) { recVUMI_MAX_xyzw(VU, 3); }

void recVUMI_MINI(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededVFtoXMMreg(_Fd_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

		MOVAPSRtoR(t0reg, fsreg);
		MINPSRtoR(t0reg, ftreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, ftreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_MINI_iq(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	MOVSSMtoR(t0reg, addr); 
	PSHUFDRtoR(t0reg, t0reg, 0x00);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MINPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

void recVUMI_MINI_xyzw(VURegs *VU, int xyzw) {
	int t0reg;
	int fsreg;
	int ftreg;
	int fdreg;

	if ( _Fd_ == 0 ) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	_addNeededVFtoXMMreg(_Fd_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	_unpackVF_xyzw(t0reg, ftreg, xyzw);

	if (_X_Y_Z_W != 0xf) {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_READ | MODE_WRITE);

		MINPSRtoR(t0reg, fsreg);

		ANDPSMtoR(fdreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(fdreg, t0reg);
	} else {
		fdreg = _allocVFtoXMMreg(VU, -1, _Fd_, MODE_WRITE);

		MOVAPSRtoR(fdreg, fsreg);
		MINPSRtoR(fdreg, t0reg);
	}

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

void recVUMI_MINIi(VURegs *VU) { recVUMI_MINI_iq(VU, VU_REGI_ADDR); }
void recVUMI_MINIx(VURegs *VU) { recVUMI_MINI_xyzw(VU, 0); }
void recVUMI_MINIy(VURegs *VU) { recVUMI_MINI_xyzw(VU, 1); }
void recVUMI_MINIz(VURegs *VU) { recVUMI_MINI_xyzw(VU, 2); }
void recVUMI_MINIw(VURegs *VU) { recVUMI_MINI_xyzw(VU, 3); }

void recVUMI_OPMULA( VURegs *VU )
{
   //VU->ACC.f.x = VU_MACx_UPDATE(VU, (double)VU->VF[_Fs_].f.y * VU->VF[_Ft_].f.z); 
   //VU->ACC.f.y = VU_MACy_UPDATE(VU, (double)VU->VF[_Fs_].f.z * VU->VF[_Ft_].f.x); 
   //VU->ACC.f.z = VU_MACz_UPDATE(VU, (double)VU->VF[_Fs_].f.x * VU->VF[_Ft_].f.y); 
	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
   MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
   SHUFPSRtoR( XMM0, XMM0, 0xC9 );           // XMM0 = WXZY
   SHUFPSRtoR( XMM1, XMM1, 0xD2 );           // XMM1 = WYXZ
   MULPSRtoR( XMM0, XMM1 );
   MOVAPSRtoM( VU_ACCx_ADDR, XMM0 );

   CPU_SSE_END;
}

void recVUMI_OPMSUB( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

	_freeXMMregs(VU);
   CPU_SSE_START;

   MOVAPSMtoR( XMM0, VU_ACCx_ADDR );
   MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Fs_ ) );
   MOVAPSMtoR( XMM2, VU_VFx_ADDR( _Ft_ ) );
   SHUFPSRtoR( XMM1, XMM1, 0xC9 );
   SHUFPSRtoR( XMM2, XMM2, 0xD2 );
   MULPSRtoR( XMM1, XMM2 );
   SUBPSRtoR( XMM0, XMM1 );
   MOVAPSRtoM( VU_VFx_ADDR( _Fd_ ), XMM0 );

   CPU_SSE_END;
}

void recVUMI_NOP( VURegs *VU ) 
{
}

void recVUMI_FTOI0(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;

	if ( _Ft_ == 0 ) return; 

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

		CVTPS2DQRtoR(t0reg, fsreg);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(ftreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
		CVTPS2DQRtoR(ftreg, fsreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_FTOIX(VURegs *VU, u32 addr) {
	int t0reg;
	int fsreg;
	int ftreg;

	if ( _Ft_ == 0 ) return; 

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

		MULPSMtoR(t0reg, addr);
		CVTPS2DQRtoR(t0reg, fsreg);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(ftreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
		MOVAPSRtoR(ftreg, fsreg);
		MULPSMtoR(ftreg, addr);
		CVTPS2DQRtoR(ftreg, ftreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_FTOI4( VURegs *VU ) {
	recVUMI_FTOIX(VU, (u32)&recMult_float_to_int4[0]);
}

void recVUMI_FTOI12( VURegs *VU ) {
	recVUMI_FTOIX(VU, (u32)&recMult_float_to_int12[0]);
}

void recVUMI_FTOI15( VURegs *VU ) {
	recVUMI_FTOIX(VU, (u32)&recMult_float_to_int15[0]);
}

void recVUMI_ITOF0( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);
   if ( _X_Y_Z_W != 0xf )
   {
      CVTDQ2PSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
      ORPSRtoR( XMM0, XMM1 );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }
   else
   {
      CVTDQ2PSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );
      MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );
   }

}

void recVUMI_ITOF4( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_ITOF12( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_ITOF15( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

static u32 const_clipAND[4] = 
   { 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff };
static u32 const_clipOR[4] = 
   { 0x80000000, 0x80000000, 0x80000000, 0x80000000 };

void recVUMI_CLIP(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

	MOV32MtoR(EAX, (uptr)&VU->clipflag);
	SHL32ItoR(EAX, 6);

	_unpackVF_xyzw(t0reg, ftreg, 3);
	ANDPSMtoR(t0reg, (u32)const_clipAND);
	CMPNLEPSRtoR(t0reg, fsreg);
//EXTRW();
	MOVD64XMMtoR(EDX, t0reg);
	OR32RtoR(EAX, EDX);
	MOVD64XMMtoR(EDX, t0reg);
	OR32RtoR(EAX, EDX);

	_unpackVF_xyzw(t0reg, ftreg, 3);
	ORPSMtoR(t0reg, (u32)const_clipOR);
	CMPLTPSRtoR(t0reg, fsreg);


	MOV32RtoM((uptr)&VU->clipflag, EAX);
	MOV32RtoM((uptr)&VU->VI[REG_CLIP_FLAG].UL, EAX);

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

/******************************/
/*   VU Lower instructions    */
/******************************/

void recVUMI_DIV(VURegs *VU) {
	int t0reg;
	int fsreg;
	int ftreg;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	t0reg = _allocTempXMMreg(-1);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);
	ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ);

/*	if (_Fsf_ == 0 && _Ftf_ == 0) {
		MOVSSRtoR(t0reg, fsreg);
		DIVSSRtoR(t0reg, ftreg);
		MOVDXMMtoM((u32)&VU->q.F, t0reg);
	} else {
	}*/
	_unpackVF_xyzw(t0reg, fsreg, _Fsf_);
//	_unpackVF_xyzw(t1reg, ftreg, _Ftf_);
//	MOVSSRtoR(t0reg, fsreg);
	DIVPSRtoR(t0reg, ftreg);
	_unpackVF_xyzw(t0reg, t0reg, _Ftf_);
	MOVDXMMtoM((u32)&VU->q.F, t0reg);

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
/*	_freeXMMregs(VU);

	SetFPUstate( );

	MOV32MtoR( EAX, (u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ] );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 0 ] = JNE8( 0 );
	MOV32MtoR( EAX, (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 1 ] = JNE8( 0 );
	recVU_UPDATE_STAT_REG_DI2( VU );
	j8Ptr[ 2 ] = JMP8( 0 );
	x86SetJ8( j8Ptr[ 1 ] );
	recVU_UPDATE_STAT_REG_DI1( VU );
	j8Ptr[ 3 ] = JMP8( 0 );
	x86SetJ8( j8Ptr[ 0 ] );
	recVU_UPDATE_STAT_REG_DI0( VU );

	FLD32( (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	FDIV32( (u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ] );
	FSTP32( (u32)&VU->q.F );
	VU->qcycles = 6;

	x86SetJ8( j8Ptr[ 2 ] );
	x86SetJ8( j8Ptr[ 3 ] );*/
}

void recVUMI_SQRT( VURegs *VU ) {
	_freeXMMregs(VU);

	FLD32((u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ]);
	FSQRT();
	FSTP32( (u32)&VU->q.F );
}

static u32 rsqrttmp;

void recVUMI_RSQRT(VURegs *VU) {
	_freeXMMregs(VU);

	FLD32((u32)&VU->VF[ _Ft_ ].F[ _Ftf_ ]);
	FSQRT( );
	FSTP32( (u32)&rsqrttmp );

	MOV32MtoR( EAX, (u32)&rsqrttmp );
	OR32RtoR( EAX, EAX );
	j8Ptr[ 0 ] = JE8( 0 );				
	FLD32( (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
	FDIV32( (u32)&rsqrttmp );
	FSTP32( (u32)&VU->q.F );
	x86SetJ8( j8Ptr[ 0 ] );
}

void _addISIMMtoIT(VURegs *VU, s16 imm) {
	if (_Ft_ == _Fs_) {
		if (imm != 0 ) {
			ADD16ItoM(VU_VI_ADDR(_Ft_ ), imm);
		}
	} else {
		if (_Fs_ == 0) {
			MOV16ItoM(VU_VI_ADDR(_Ft_), imm);
		} else {
			MOV16MtoR(EAX, VU_VI_ADDR( _Fs_ ) );
			if (imm != 0) {
				ADD16ItoR( EAX, imm );
			}
			MOV16RtoM(VU_VI_ADDR( _Ft_ ), EAX );
		}
	}
}

void recVUMI_IADDI(VURegs *VU) {
	s16 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( VU->code >> 6 ) & 0x1f;
	imm = ( imm & 0x10 ? 0xfff0 : 0) | ( imm & 0xf );
	_addISIMMtoIT(VU, imm);
}

void recVUMI_IADDIU(VURegs *VU) {
	u32 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( ( VU->code >> 10 ) & 0x7800 ) | ( VU->code & 0x7ff );
	_addISIMMtoIT(VU, imm);
}

void recVUMI_IADD( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( ( _Ft_ == 0 ) && ( _Fs_ == 0 ) )
   {
      MOV16ItoM( VU_VI_ADDR( _Fd_ ), 0 );
   } 
   else if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else if ( _Ft_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      ADD16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_IAND( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( ( _Fs_ == 0 ) || ( _Ft_ == 0 ) )
   {
      XOR16RtoR( EAX, EAX );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      if ( _Fs_ == _Ft_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else if ( _Fd_ == _Fs_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
         AND16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else if ( _Fd_ == _Ft_ )
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         AND16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
      else
      {
         MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
         AND16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
         MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      }
   }
}

void recVUMI_IOR( VURegs *VU )
{
   if ( _Fd_ == 0 ) return;

   if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      if ( _Ft_ != 0 )
      {
         OR16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      }
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_ISUB( VURegs *VU ) {
   if ( _Fd_ == 0 ) return;

   if ( _Fs_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      NEG16R( EAX );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
      return;
   }
   else if ( _Ft_ == 0 )
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
   else
   {
      MOV16MtoR( EAX, VU_VI_ADDR( _Fs_ ) );
      SUB16MtoR( EAX, VU_VI_ADDR( _Ft_ ) );
      MOV16RtoM( VU_VI_ADDR( _Fd_ ), EAX );
   }
}

void recVUMI_ISUBIU( VURegs *VU ) {
	u32 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( ( VU->code >> 10 ) & 0x7800 ) | ( VU->code & 0x7ff );
	if (_Ft_ == _Fs_) {
		if (imm != 0 ) {
			SUB16ItoM(VU_VI_ADDR(_Ft_ ), imm);
		}
	} else {
		if (_Fs_ == 0) {
			MOV16ItoM(VU_VI_ADDR(_Ft_), -imm);
		} else {
			MOV16MtoR(EAX, VU_VI_ADDR( _Fs_ ) );
			if (imm != 0) {
				SUB16ItoR( EAX, imm );
			}
			MOV16RtoM(VU_VI_ADDR( _Ft_ ), EAX );
		}
	}
}

void recVUMI_MOVE( VURegs *VU ) {
	int t0reg;
	int fsreg;
	int ftreg;

	if (_Ft_ == 0) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);

		MOVAPSRtoR(t0reg, fsreg);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(ftreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);

		MOVAPSRtoR(ftreg, fsreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_MFIR( VURegs *VU ) {
	int t0reg;
	int ftreg;

	if ( _Ft_ == 0 ) return;

	_addNeededVFtoXMMreg(_Ft_);
	t0reg = _allocTempXMMreg(-1);

	MOVSX32M16toR(EAX, VU_VI_ADDR(_Fs_));
	MOVD32RtoXMM(t0reg, EAX);
	UNPCKLPSRtoR(t0reg, t0reg);
	MOVLHPSRtoR(t0reg, t0reg);

    if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);

		ORPSRtoR(ftreg, t0reg);
    } else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
		MOVAPSRtoR(ftreg, t0reg);
	}

	_freeXMMreg(t0reg);
	_clearNeededXMMregs();
}

void recVUMI_MTIR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, (u32)&VU->VF[ _Fs_ ].F[ _Fsf_ ] );
   MOV16RtoM( VU_VI_ADDR( _Ft_ ), EAX );
} 

void recVUMI_MR32( VURegs *VU ) {
	int t0reg;
	int fsreg;
	int ftreg;

	if (_Ft_ == 0) return;

	_addNeededVFtoXMMreg(_Fs_);
	_addNeededVFtoXMMreg(_Ft_);
	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

	if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);

		PSHUFDRtoR(t0reg, fsreg, 0x39);

		ANDPSMtoR(ftreg, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(ftreg, t0reg);

		_freeXMMreg(t0reg);
	} else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);

		PSHUFDRtoR(ftreg, fsreg, 0x39);
	}

	_clearNeededXMMregs();



/*   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOVAPSMtoR( XMM0, VU_VFx_ADDR( _Fs_ ) );

   if ( _X_Y_Z_W != 0xf )
   {
      PSHUFDRtoR(XMM0, XMM0, ((0) << 6) | ((3) << 4) | ((2) << 2) | (1));

      MOVAPSMtoR( XMM1, VU_VFx_ADDR( _Ft_ ) );
      ANDPSMtoR( XMM0, (u32)&SSEmovMask[ _X_Y_Z_W ][ 0 ] );
      ANDPSMtoR( XMM1, (u32)&SSEmovMask[ 15 - _X_Y_Z_W ][ 0 ] );
		ORPSRtoR(XMM0, XMM1);
   } else {
      PSHUFDMtoR(XMM0, VU_VFx_ADDR( _Ft_ ), ((0) << 6) | ((3) << 4) | ((2) << 2) | (1));
   }

   MOVAPSRtoM( VU_VFx_ADDR( _Ft_ ), XMM0 );*/
}

void _loadEAX(VURegs *VU) {
	int t0reg;
	int ftreg;

    if (_X_Y_Z_W != 0xf) {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_READ | MODE_WRITE);
		t0reg = _allocTempXMMreg(-1);

	    MOVAPSRmtoR(t0reg, EAX);
		ANDPSMtoR(ftreg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[_X_Y_Z_W][0]);

		ORPSRtoR(ftreg, t0reg);
		_freeXMMreg(t0reg);
    } else {
		ftreg = _allocVFtoXMMreg(VU, -1, _Ft_, MODE_WRITE);
	    MOVAPSRmtoR(ftreg, EAX);
	}

	_clearNeededXMMregs();
}

void recVUMI_LQ(VURegs *VU) {
	s16 imm;

	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem + ((imm * 16) & 0x3fff) );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_loadEAX(VU);
}

void recVUMI_LQD( VURegs *VU ) {
	if ( _Fs_ != 0 ) {
		DEC16M( VU_VI_ADDR( _Fs_ ) );
	}

	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_loadEAX(VU);
}

void recVUMI_LQI(VURegs *VU) {
	if ( _Ft_ == 0 ) {
		INC32M( VU_VI_ADDR( _Fs_ ) );
		return;
	}

    if (_Fs_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
    } else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
		INC32M( VU_VI_ADDR( _Fs_ ) );
    }

	_loadEAX(VU);
}

void _saveEAX(VURegs *VU) {
	int t0reg;
	int fsreg;

	fsreg = _allocVFtoXMMreg(VU, -1, _Fs_, MODE_READ);

    if (_X_Y_Z_W != 0xf) {
		t0reg = _allocTempXMMreg(-1);

	    MOVAPSRmtoR(t0reg, EAX);
		ANDPSMtoR(fsreg, (u32)&SSEmovMask[_X_Y_Z_W][0]);
		ANDPSMtoR(t0reg, (u32)&SSEmovMask[15 - _X_Y_Z_W][0]);

		ORPSRtoR(t0reg, fsreg);
		MOVAPSRtoRm(EAX, t0reg);

		_freeXMMreg(t0reg);
    } else {
		MOVAPSRtoRm(EAX, fsreg);
	}

	_clearNeededXMMregs();
}

void recVUMI_SQ(VURegs *VU) {
	s16 imm;
 
	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Ft_ == 0 ) {
      MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_saveEAX(VU);
}

void recVUMI_SQD(VURegs *VU) {
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		SUB32ItoM( VU_VI_ADDR( _Ft_ ), 1 );
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	_saveEAX(VU);
}

void recVUMI_SQI(VURegs *VU) {
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Ft_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
		ADD32ItoM( VU_VI_ADDR( _Ft_ ), 1 );
	}

	_saveEAX(VU);
}

void recVUMI_ILW(VURegs *VU) {
	s16 imm;
 
	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if ( _Ft_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV32RmStoR( ECX, EAX, EDX, 2 );
	MOV16RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], ECX );
}

void recVUMI_ISW( VURegs *VU ) {
	s16 imm;
 
	if ( _Ft_ == 0 ) return;

	imm = ( VU->code & 0x400) ? ( VU->code & 0x3ff) | 0xfc00 : ( VU->code & 0x3ff); 
	if (_Ft_ == 0) {
		MOV32ItoR( EAX, (u32)VU->Mem + imm * 16 );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		ADD32ItoR( EAX, imm * 16 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	MOV32MtoR( ECX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV16RtoRmS( ECX, EAX, EDX, 2 );
}

void recVUMI_ILWR( VURegs *VU ) {
	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV32RmStoR( ECX, EAX, EDX, 2 );
	MOV16RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], ECX );
}

void recVUMI_ISWR( VURegs *VU ) {
	if ( _Ft_ == 0 ) return;

	if ( _Fs_ == 0 ) {
		MOV32ItoR( EAX, (u32)VU->Mem );
	} else {
		MOVZX32M16toR( EAX, VU_VI_ADDR( _Fs_ ) );
		SHL32ItoR( EAX, 4 );
		AND32ItoR( EAX, 0x3fff);
		ADD32ItoR( EAX, (u32)VU->Mem );
	}

	MOV32MtoR( ECX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	if (_X) XOR32RtoR(EDX, EDX);
	else if (_Y) MOV32ItoR(EDX, 1);
	else if (_Z) MOV32ItoR(EDX, 2);
	else if (_W) MOV32ItoR(EDX, 3);
	MOV16RtoRmS( ECX, EAX, EDX, 2 );
}

void recVUMI_RINIT(VURegs *VU) {
	_freeXMMregs(VU);
	MOV32MtoR( EAX, VU_VFx_ADDR( _Fs_ ) + 4 * _Fsf_ );
	AND32ItoR( EAX, 0x7fffff );
	OR32ItoR( EAX, 0x7f << 23 );
	MOV32RtoM( VU_REGR_ADDR, EAX );
}

void recVUMI_RGET(VURegs *VU) {
   if ( _Ft_ == 0 ) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_REGR_ADDR );
   if ( _X )   MOV32RtoM( VU_VFx_ADDR( _Ft_ ), EAX );
   if ( _Y )   MOV32RtoM( VU_VFy_ADDR( _Ft_ ), EAX );
   if ( _Z )   MOV32RtoM( VU_VFz_ADDR( _Ft_ ), EAX );
   if ( _W )   MOV32RtoM( VU_VFw_ADDR( _Ft_ ), EAX );
}

void recVUMI_RNEXT( VURegs *VU )
{
   if ( _Ft_ == 0) return;

	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_REGR_ADDR );
   MOV32ItoR( ECX, 0x15a4e35 );
   IMUL32R( ECX );
   INC32R( EAX );
   SHR32ItoR( EAX, 9 );
   MOV32RtoM( VU_REGR_ADDR, EAX );
   OR32ItoR( EAX, 0x7f << 23 );
   if ( _X )   MOV32RtoM( VU_VFx_ADDR( _Ft_ ), EAX );
   if ( _Y )   MOV32RtoM( VU_VFy_ADDR( _Ft_ ), EAX );
   if ( _Z )   MOV32RtoM( VU_VFz_ADDR( _Ft_ ), EAX );
   if ( _W )   MOV32RtoM( VU_VFw_ADDR( _Ft_ ), EAX );
}

void recVUMI_RXOR( VURegs *VU )
{
	_freeXMMregs(VU);

   MOV32MtoR( EAX, VU_VFx_ADDR( _Fs_ ) + 4 * _Fsf_ );
   XOR32MtoR( EAX, VU_REGR_ADDR );
   AND32ItoR( EAX, 0x7fffff );
   OR32ItoR ( EAX, 0x7f << 23 );
   MOV32RtoM( VU_REGR_ADDR, EAX );
} 

void recVUMI_WAITQ( VURegs *VU )
{
}

void recVUMI_FSAND( VURegs *VU ) {
	u16 imm;
	imm = (((VU->code >> 21 ) & 0x1) << 11) | (VU->code & 0x7ff);
	if(_Ft_ == 0) return; 

	MOV32MtoR(EAX, (u32)&VU->VI[REG_STATUS_FLAG].US[0]);
	AND32ItoR(EAX, imm);
	MOV32RtoM((u32)&VU->VI[_Ft_].US[0], EAX);
}

void recVUMI_FSEQ( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FSOR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FSSET(VURegs *VU) {
	u16 imm = 0;
	imm = (((VU->code >> 21 ) & 0x1) << 0x5) | ((VU->code >> 6) & 0x1F);
	imm = (imm & 0x3f) << 0x6;

	if (imm == 0) {
		AND32ItoM((u32)&VU->statusflag, 0x3F);
	} else {
		MOV32MtoR(EAX, (u32)&VU->statusflag);
		AND32ItoR(EAX, 0x3F);
		OR32ItoR(EAX, imm);
		MOV32RtoM((u32)&VU->statusflag, EAX);
	}
}

void recVUMI_FMAND( VURegs *VU ) {
   if ( _Ft_ == 0 ) return;

   MOV32MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
   AND32MtoR( EAX, (u32)&VU->VI[ REG_MAC_FLAG ].US[ 0 ] );
   AND32ItoR( EAX, 0xFFFF );
   MOV32RtoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], EAX );
}

void recVUMI_FMEQ( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FMOR( VURegs *VU )
{
   if ( _Ft_ == 0 ) return;
}

void recVUMI_FCAND( VURegs *VU ) {
   MOV32MtoR( EAX, (u32)&VU->VI[ REG_CLIP_FLAG ].UL );
   AND32ItoR( EAX, VU->code & 0xFFFFFF );
   NEG32R( EAX );
   SBB32RtoR( EAX, EAX );
   NEG32R( EAX );
   MOV32RtoM( (u32)&VU->VI[ 1 ].UL, EAX );
}

void recVUMI_FCEQ( VURegs *VU )
{
}

void recVUMI_FCOR( VURegs *VU )
{
}

void recVUMI_FCSET( VURegs *VU )
{
}

void recVUMI_FCGET( VURegs *VU )
{
}

s32 _recbranchAddr(VURegs * VU) {
	s32 bpc = pc + _Imm11_ * 8; 
	if (bpc < 0) {
		bpc = pc + _UImm11_ * 8; 
	}
	if (VU == &VU1) {
		bpc&= 0x3fff;
	} else {
		bpc&= 0x0fff;
	}

	return bpc;
}

void recVUMI_IBEQ(VURegs *VU) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16MtoR( EAX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	CMP16MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );

	j8Ptr[ 0 ] = JNE8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );

	branch = 1;
}

void recVUMI_IBGEZ( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

	j8Ptr[ 0 ] = JL8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );

	branch = 1;
}

void recVUMI_IBGTZ( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

	j8Ptr[ 0 ] = JLE8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );

	branch = 1;
}

void recVUMI_IBLEZ( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

	j8Ptr[ 0 ] = JG8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );

	branch = 1;
}

void recVUMI_IBLTZ( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	CMP16ItoM( (u32)&VU->VI[ _Fs_ ].US[ 0 ], 0x0 );

	j8Ptr[ 0 ] = JGE8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );
    
	branch = 1;
}

void recVUMI_IBNE( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV16ItoM( (u32)&VU->VI[ REG_TPC ].UL, (u32)pc );
	MOV16MtoR( EAX, (u32)&VU->VI[ _Ft_ ].US[ 0 ] );
	CMP16MtoR( EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );

	j8Ptr[ 0 ] = JE8( 0 );
	MOV32ItoM((uptr)&VU->branch, 2);
	MOV32ItoM((uptr)&VU->branchpc, bpc);
	x86SetJ8( j8Ptr[ 0 ] );
    
	branch = 1;
}

void recVUMI_B(VURegs *VU) {
	u32 bpc = _recbranchAddr(VU); 

	MOV32ItoM((u32)&VU->VI[ REG_TPC ].UL, bpc);

	branch = 3;
}

void recVUMI_BAL( VURegs *VU ) {
	u32 bpc = _recbranchAddr(VU); 

	MOV32ItoM((u32)&VU->VI[ REG_TPC ].UL, bpc);

	if ( _Ft_ ) {
		MOV16ItoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], (pc + 8)/8 );
	}

	branch = 3;
}

void recVUMI_JR(VURegs *VU) {
	MOV32MtoR(EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
	SHL32ItoR(EAX, 3);
	MOV32RtoM((u32)&VU->VI[ REG_TPC ].UL, EAX);

	branch = 3;
}

void recVUMI_JALR(VURegs *VU) {
	MOV32MtoR(EAX, (u32)&VU->VI[ _Fs_ ].US[ 0 ] );
	SHL32ItoR(EAX, 3);
	MOV32RtoM((u32)&VU->VI[ REG_TPC ].UL, EAX);

	if ( _Ft_ ) {
		MOV16ItoM( (u32)&VU->VI[ _Ft_ ].US[ 0 ], (pc + 8)/8 );
	}

	branch = 3;
}

void recVUMI_MFP(VURegs *VU) {
   if ( _Ft_ == 0 ) return;
}

void recVUMI_WAITP(VURegs *VU) {
}

void recVUMI_ESADD( VURegs *VU )
{
}

void recVUMI_ERSADD( VURegs *VU )
{
}

void recVUMI_ELENG( VURegs *VU )
{
}

void recVUMI_ERLENG( VURegs *VU )
{
}

void recVUMI_EATANxy( VURegs *VU )
{
}

void recVUMI_EATANxz( VURegs *VU )
{
}

void recVUMI_ESUM( VURegs *VU )
{
}

void recVUMI_ERCPR( VURegs *VU )
{
}

void recVUMI_ESQRT( VURegs *VU )
{
}

void recVUMI_ERSQRT( VURegs *VU )
{
}

void recVUMI_ESIN( VURegs *VU )
{
}

void recVUMI_EATAN( VURegs *VU )
{
}

void recVUMI_EEXP( VURegs *VU )
{
}

void recVUMI_XITOP( VURegs *VU )
{
}

void recVUMI_XTOP( VURegs *VU ) {
	if ( _Ft_ == 0 ) return;

	MOV32MtoR( EAX, (u32)&VU->vifRegs->top );
	MOV16RtoM( VU_VI_ADDR( _Ft_ ), EAX );
}

void TT() {
	GSgifTransfer1(0, 0);
}

void recVUMI_XGKICK( VURegs *VU ) {
#ifdef __x86_64__
	MOV32MtoR(RSI, (uptr)&VU->VI[_Fs_].US[0]);
	SHL32ItoR(RSI, 4);
	AND32ItoR(RSI, 0x3fff);

	MOV64MtoR(RDI, (uptr)VU->Mem );
	MOV64MtoR(RDX, (uptr)GSgifTransfer1);
	JMP64R(RDX);
#else
	MOV32MtoR(EAX, (uptr)&VU->VI[_Fs_].US[0]);
	SHL32ItoR(EAX, 4);
	AND32ItoR(EAX, 0x3fff);

	PUSH32R(EAX);
	PUSH32I((u32)VU->Mem);
	CALLFunc((u32)GSgifTransfer1);
#endif
}

