
#include "ix86.h"

/********************/
/* SSE instructions */
/********************/

#define SSEMtoR( code, overb ) \
	if (to > 7) Rex(1, to >> 3, 0, 0); \
	write16( code ); \
	ModRM( 0, to, DISP32 ); \
	write32( MEMADDR(from, 4 + overb) ); \

#define SSERtoM( code, overb ) \
	if (from > 7) Rex(1, from >> 3, 0, 0); \
	write16( code ); \
	ModRM( 0, from, DISP32 ); \
	write32( MEMADDR(to, 4 + overb) ); \

#define SSE_SS_MtoR( code, overb ) \
	write8( 0xf3 ); \
	if (to > 7) Rex(1, to >> 3, 0, 0); \
	write16( code ); \
	ModRM( 0, to, DISP32 ); \
	write32( MEMADDR(from, 4 + overb) ); \

#define SSE_SS_RtoM( code, overb ) \
	write8( 0xf3 ); \
	if (from > 7) Rex(1, from >> 3, 0, 0); \
	write16( code ); \
	ModRM( 0, from, DISP32 ); \
	write32( MEMADDR(to, 4 + overb) ); \

#define SSERtoR( code ) \
	if (to > 7 || from > 7) Rex(1, to >> 3, 0, from >> 3); \
	write16( code ); \
	ModRM( 3, to, from );

#define SSEMtoR66( code ) \
	write8( 0x66 ); \
	SSEMtoR( code, 0 );

#define SSERtoM66( code ) \
	write8( 0x66 ); \
	SSERtoM( code, 0 );

#define SSERtoR66( code ) \
	write8( 0x66 ); \
	SSERtoR( code );

#define _SSERtoR66( code ) \
	write8( 0x66 ); \
	if (to > 7 || from > 7) Rex(1, from >> 3, 0, to >> 3); \
	write16( code ); \
	ModRM( 3, from, to );

#define SSE_SS_RtoR( code ) \
	write8( 0xf3 ); \
	if (to > 7 || from > 7) Rex(1, to >> 3, 0, from >> 3); \
	write16( code ); \
	ModRM( 3, to, from );

#define CMPPSMtoR( op ) \
   SSEMtoR( 0xc20f, 1 ); \
   write8( op );

#define CMPPSRtoR( op ) \
   SSERtoR( 0xc20f ); \
   write8( op );

#define CMPSSMtoR( op ) \
   SSE_SS_MtoR( 0xc20f, 1 ); \
   write8( op );

#define CMPSSRtoR( op ) \
   SSE_SS_RtoR( 0xc20f ); \
   write8( op );



void MOVAPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x280f, 0 ); }
void MOVAPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x290f, 0 ); }
void MOVAPSRtoR( x86SSERegType to, x86SSERegType from )     { SSERtoR( 0x280f ); }

/* movups [r32][r32*scale] to xmm1 */
void MOVUPSRmStoR( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	if (to > 7) Rex(1, to >> 3, 0, 0);
	write16( 0x100f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movups xmm1 to [r32][r32*scale] */
void MOVUPSRtoRmS( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	if (to > 7) Rex(1, to >> 3, 0, 0);
	write16( 0x110f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movups [r32] to r32 */
void MOVUPSRmtoR( x86IntRegType to, x86IntRegType from ) 
{
	if (to > 7) Rex(1, to >> 3, 0, 0);
	write16( 0x100f );
	ModRM( 0, to, from );
}

/* movups r32 to [r32] */
void MOVUPSRtoRm( x86IntRegType to, x86IntRegType from ) 
{
	write16( 0x110f );
	ModRM( 0, from, to );
}

/* movaps [r32][r32*scale] to xmm1 */
void MOVAPSRmStoR( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x100f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movaps xmm1 to [r32][r32*scale] */
void MOVAPSRtoRmS( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x110f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movaps [r32] to r32 */
void MOVAPSRmtoR( x86IntRegType to, x86IntRegType from ) 
{
	if (to > 7) Rex(1, to >> 3, 0, 0);
	write16( 0x280f );
	ModRM( 0, to, from );
}

/* movaps r32 to [r32] */
void MOVAPSRtoRm( x86IntRegType to, x86IntRegType from ) 
{
	if (from > 7) Rex(1, from >> 3, 0, 0);
	write16( 0x290f );
	ModRM( 0, from, to );
}

void MOVSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x100f, 0 ); }
void MOVSSRtoM( u32 to, x86SSERegType from )           { SSE_SS_RtoM( 0x110f, 0 ); }
void MOVSSRtoR( x86SSERegType to, x86SSERegType from )      { SSE_SS_RtoR( 0x100f ); }

void MOVLPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x120f, 0 ); }
void MOVLPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x130f, 0 ); }

void MOVHPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x160f, 0 ); }
void MOVHPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x170f, 0 ); }

void MOVLHPSRtoR( x86SSERegType to, x86SSERegType from )    { SSERtoR( 0x160f ); }
void MOVHLPSRtoR( x86SSERegType to, x86SSERegType from )    { SSERtoR( 0x120f ); }


void ORPSMtoR( x86SSERegType to, u32 from )            { SSEMtoR( 0x560f, 0 ); }
void ORPSRtoR( x86SSERegType to, x86SSERegType from )  { SSERtoR( 0x560f ); }
void ORSSMtoR( x86SSERegType to, u32 from )            { SSE_SS_MtoR( 0x560f, 0 ); }
void ORSSRtoR( x86SSERegType to, x86SSERegType from )  { SSE_SS_RtoR( 0x560f ); }

void XORPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x570f, 0 ); }
void XORPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x570f ); }
void XORSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x570f, 0 ); }
void XORSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x570f ); }

void ANDPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x540f, 0 ); }
void ANDPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x540f ); }
void ANDSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x540f, 0 ); }
void ANDSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x540f ); }

void ANDNPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x550f, 0 ); }
void ANDNPSRtoR( x86SSERegType to, x86SSERegType from ){ SSERtoR( 0x550f ); }
void ANDNSSMtoR( x86SSERegType to, u32 from )          { SSE_SS_MtoR( 0x550f, 0 ); }
void ANDNSSRtoR( x86SSERegType to, x86SSERegType from ){ SSE_SS_RtoR( 0x550f ); }

void ADDPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x580f, 0 ); }
void ADDPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x580f ); }
void ADDSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x580f, 0 ); }
void ADDSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x580f ); }

void SUBPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5c0f, 0 ); }
void SUBPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x5c0f ); }
void SUBSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x5c0f, 0 ); }
void SUBSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x5c0f ); }

void MULPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x590f, 0 ); }
void MULPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x590f ); }
void MULSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x590f, 0 ); }
void MULSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x590f ); }

void RCPPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x530f ); }
void RCPPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x530f, 0 ); }

void CMPEQPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 0 ); }
void CMPEQPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 0 ); }
void CMPLTPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 1 ); }
void CMPLTPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 1 ); }
void CMPLEPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 2 ); }
void CMPLEPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 2 ); }
void CMPUNORDPSMtoR( x86SSERegType to, u32 from )      { CMPPSMtoR( 3 ); }
void CMPUNORDPSRtoR( x86SSERegType to, x86SSERegType from ) { CMPPSRtoR( 3 ); }
void CMPNEPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 4 ); }
void CMPNEPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 4 ); }
void CMPNLTPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 5 ); }
void CMPNLTPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 5 ); }
void CMPNLEPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 6 ); }
void CMPNLEPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 6 ); }
void CMPORDPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 7 ); }
void CMPORDPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 7 ); }

void CVTPI2PSMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x2a0f, 0 ); }
void CVTPI2PSRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x2a0f ); }
void CVTPS2PIMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x2d0f, 0 ); }
void CVTPS2PIRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x2d0f ); }

void CVTDQ2PSMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x5b0f, 0 ); }
void CVTDQ2PSRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x5b0f ); }
void CVTPS2DQMtoR( x86SSERegType to, u32 from )        { SSEMtoR66( 0x5b0f ); }
void CVTPS2DQRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR66( 0x5b0f ); }

void MAXPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5f0f, 0 ); }
void MAXPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x5f0f ); }
void MAXSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x5f0f, 0 ); }
void MAXSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x5f0f ); }

void MINPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5d0f, 0 ); }
void MINPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x5d0f ); }
void MINSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x5d0f, 0 ); }
void MINSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x5d0f ); }

void RSQRTPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x520f, 0 ); }
void RSQRTPSRtoR( x86SSERegType to, x86SSERegType from ){ SSERtoR( 0x520f ); }
void RSQRTSSMtoR( x86SSERegType to, u32 from )          { SSE_SS_MtoR( 0x520f, 0 ); }
void RSQRTSSRtoR( x86SSERegType to, x86SSERegType from ){ SSE_SS_RtoR( 0x520f ); }

void SQRTPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x510f, 0 ); }
void SQRTPSRtoR( x86SSERegType to, x86SSERegType from ){ SSERtoR( 0x510f ); }
void SQRTSSMtoR( x86SSERegType to, u32 from )          { SSE_SS_MtoR( 0x510f, 0 ); }
void SQRTSSRtoR( x86SSERegType to, x86SSERegType from ){ SSE_SS_RtoR( 0x510f ); }

void PMAXSWRtoR( x86SSERegType to, x86SSERegType from ){ SSERtoR( 0xEE0F ); }
void PMINSWRtoR( x86SSERegType to, x86SSERegType from ){ SSERtoR( 0xEA0F ); }

void UNPCKLPSRtoR( x86SSERegType to, x86SSERegType from )	{ SSERtoR( 0x140F ); }
void UNPCKHPSRtoR( x86SSERegType to, x86SSERegType from )	{ SSERtoR( 0x150F ); }

void SHUFPSRtoR( x86SSERegType to, x86SSERegType from, u8 imm8 )	{ SSERtoR( 0xC60F ); write8( imm8 ); }
void SHUFPSMtoR( x86SSERegType to, u32 from, u8 imm8 )		{ SSEMtoR( 0xC60F, 1 ); write8( imm8 ); }

void MOVDMtoXMM( x86SSERegType to, u32 from ) { SSEMtoR66(0x6E0F); }
void MOVDXMMtoM( u32 to, x86SSERegType from ) { SSERtoM66(0x7E0F); }
void MOVD32RtoXMM( x86SSERegType to, x86IntRegType from ) { _SSERtoR66(0x6E0F); }
void MOVD64RtoXMM( x86SSERegType to, x86IntRegType from ) { _SSERtoR66(0x6E0F); }
void MOVD32XMMtoR( x86IntRegType to, x86SSERegType from ) { _SSERtoR66(0x7E0F); }
void MOVD64XMMtoR( x86IntRegType to, x86SSERegType from ) { _SSERtoR66(0x7E0F); }

void PSHUFDRtoR( x86SSERegType to, x86SSERegType from, u8 imm8 )	{ SSERtoR66( 0x700F ); write8( imm8 ); }
void PSHUFDMtoR( x86SSERegType to, u32 from, u8 imm8 )	{ SSEMtoR66( 0x700F ); write8( imm8 ); }

void DIVPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5e0F, 0 ); }
void DIVPSRtoR( x86SSERegType to, x86SSERegType from ) { SSERtoR( 0x5e0F ); }
void DIVSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x5e0F, 0 ); }
void DIVSSRtoR( x86SSERegType to, x86SSERegType from ) { SSE_SS_RtoR( 0x5e0F ); }

void CMPEQSSMtoR( x86SSERegType to, u32 from )         { CMPSSMtoR( 0 ); }
void CMPEQSSRtoR( x86SSERegType to, x86SSERegType from )    { CMPSSRtoR( 0 ); }
void CMPLTSSMtoR( x86SSERegType to, u32 from )         { CMPSSMtoR( 1 ); }
void CMPLTSSRtoR( x86SSERegType to, x86SSERegType from )    { CMPSSRtoR( 1 ); }
void CMPLESSMtoR( x86SSERegType to, u32 from )         { CMPSSMtoR( 2 ); }
void CMPLESSRtoR( x86SSERegType to, x86SSERegType from )    { CMPSSRtoR( 2 ); }
void CMPUNORDSSMtoR( x86SSERegType to, u32 from )      { CMPSSMtoR( 3 ); }
void CMPUNORDSSRtoR( x86SSERegType to, x86SSERegType from ) { CMPSSRtoR( 3 ); }
void CMPNESSMtoR( x86SSERegType to, u32 from )         { CMPSSMtoR( 4 ); }
void CMPNESSRtoR( x86SSERegType to, x86SSERegType from )    { CMPSSRtoR( 4 ); }
void CMPNLTSSMtoR( x86SSERegType to, u32 from )        { CMPSSMtoR( 5 ); }
void CMPNLTSSRtoR( x86SSERegType to, x86SSERegType from )   { CMPSSRtoR( 5 ); }
void CMPNLESSMtoR( x86SSERegType to, u32 from )        { CMPSSMtoR( 6 ); }
void CMPNLESSRtoR( x86SSERegType to, x86SSERegType from )   { CMPSSRtoR( 6 ); }
void CMPORDSSMtoR( x86SSERegType to, u32 from )        { CMPSSMtoR( 7 ); }
void CMPORDSSRtoR( x86SSERegType to, x86SSERegType from )   { CMPSSRtoR( 7 ); }

void STMXCSR( u32 from ) {
	write16( 0xAE0F );
	ModRM( 0, 0x3, DISP32 );
	write32( MEMADDR(from, 4) );
}

void LDMXCSR( u32 from ) {
	write16( 0xAE0F );
	ModRM( 0, 0x2, DISP32 );
	write32( MEMADDR(from, 4) );
}

